/*
 * Aligner.cpp
 *
 */
#include "LoadStat.hpp"
#include "EScript.hpp"
#include "Alpha.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
#include "ScoreParam.hpp"
#include "DPPath.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
//#include "DPDeckPHMM.hpp"
#include "DPDeckPSCFG.hpp"
// #include "DPDeckSCFG.hpp"
// #include "McCaskill.hpp"
#include "Alifold.hpp"
#include "Aligner.hpp"
#include "ProbCons.hpp"

/**
 * getValue() should be an increasing function.
 * if the returned value hi >= i_beg
 * then (hi = argmax (getValue(hi1) <= target) is satisfied.
 * else (target < getValue(x))is true for all x = i_beg..i_last
 */
template <typename GetValue>
static int
bisection(int i_beg, int i_last, float target, GetValue getValue)
{

#if false
#ifdef DEBUG
  cout << "---\nbisection:\n";
  cout << setprecision(3);
  cout << "  - { range: [ " << i_beg << ", " << i_last << " ], ";
  cout << "target: " << target << " }\n";
#endif
#endif

  int lo = i_beg;
  int hi = i_last;
  while (lo <= hi) {
    int mid = (lo + hi) / 2;
    float value = getValue(mid);
    if (target < value) {
      hi = mid - 1;
    } else {
      lo = mid + 1;
    }

#if false
#ifdef DEBUG
    cout << "  - { range: [ " << lo << ", " << hi << " ], ";
    cout << "value: " << value << " }\n";
#endif
#endif
    
  }
  return hi;
}

Aligner&
Aligner::
create(const string& alignerType)
{
  static const struct {string name; int type;} table[] = {
    {"PHMM",        STRIP_PHMM},
    {"PSCFG",       STRIP_PSCFG},
    {"CHUNK_PHMM",  CHUNK_PHMM},
    {"CHUNK_PSCFG", CHUNK_PSCFG},
    {"SCFG",        SCFG},
    {"MCCASKILL",   MCCASKILL},
    {"ALIFOLD",     ALIFOLD},
    {"PROBCONS",    PROBCONS},
  };
  static const int nType = (sizeof(table) / sizeof(table[0]));

  int type = -1;
  for (int i = 0; i < nType; i++) {
    if (alignerType == table[i].name) {
      type = table[i].type;
      break;
    }
  }

  switch (type) {
  case STRIP_PHMM: 
#ifdef RELEASE0
    Die(" STRIP_PHMM is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    return *new AlignerStrip(STRIP_PHMM, *new DPDeckPHMM);
#endif
  case STRIP_PSCFG: 
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
  case SCFG:
#ifdef RELEASE0
    Die("SCFG is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    return *new AlignerFold(SCFG, *new DPDeckSCFG);
#endif
  case MCCASKILL:
#ifdef RELEASE0
  Die("MCCASKILL is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    return *new AlignerFold(MCCASKILL, *new McCaskill);
#endif
  case ALIFOLD:
    return *new AlignerFold(ALIFOLD, *new Alifold);
  case PROBCONS:
    return *new AlignerStrip(PROBCONS, *new ProbCons);
  case CHUNK_PHMM: 
#ifdef RELEASE0
  Die("CHUNK_PHMM is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    return *new AlignerChunk(CHUNK_PHMM,
			     *new AlignerStrip(STRIP_PHMM, *new DPDeckPHMM));
#endif
  case CHUNK_PSCFG:
#ifdef RELEASE0
    Die("CHUNK_PSCFG is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    return *new AlignerChunk(CHUNK_PSCFG,
			     *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG));
#endif
  default:
#ifdef RELEASE0
    Die("CHUNK_PSCFG is not supported");
    return *new AlignerStrip(STRIP_PSCFG, *new DPDeckPSCFG);
#else
    Die("aligner type %s is not supported.", alignerType.c_str());
    return *new AlignerStrip(STRIP_PHMM, *new DPDeckPHMM);
#endif
  }
}
const DPPath&
Aligner::
getDPPath() const
{
  errorAbort(); return *new DPPath;
}


AlignerStrip::
AlignerStrip(int type, DPDeck& dpDeck)
  : Super(type),
    dpDeck(dpDeck),
    dpRegion(*new DPRegion),
    dpPath(*new DPPath),
    maxStripWidth(-1),
    initialDPPathIsSet(false),
    stripWidth(0),
    skipSize(1),
    memRequested((float)DEFAULT_MEM_SIZE_IN_MB),
    keepMinBandWidth(false),
    maxLWRatio(-1.0),
    minLWRatio(-1.0),
    bubbleSize(-1),
    bubbleOverlapRatio(0.3),
    shrinkByAlignProb(false),
    pAlignProb(NULL),
    alignProbCutoff(0.001),
    timeFactor(1.0),
    maxTime(10000),
    skipSizeExponent(2)
{}

AlignerStrip::
~AlignerStrip()
{
  delete &dpDeck;
  delete &dpRegion;
  delete &dpPath;
}
int
AlignerStrip::
getLengthX() const
{
  return (pMSAX ? pMSAX->getUngappedLength() : (-1));
}
int
AlignerStrip::
getLengthY() const
{
  return (pMSAY ? pMSAY->getUngappedLength() : (-1));
}
int
AlignerStrip::
getMinBandWidth() const
{
  return (abs(getLengthX() - getLengthY()) + 1);
}
int
AlignerStrip::
getMaxBandWidth() const
{
  return ((getLengthX() + 1) + (getLengthY() + 1) - 1);
}
float
AlignerStrip::
getScore() const
{
  return dpDeck.getScore();
}
const DPPath&
AlignerStrip::
getDPPath() const
{
  return dpPath;
}
Vector<float>
AlignerStrip::
computeFeatureVector() const
{
  return dpDeck.computeFeatureVector();
}
Vector<float>
AlignerStrip::
getParamVector() const
{
  const Vector<float>& pv = dpDeck.getParamVector();
  return pv;
}
Vector<float>
AlignerStrip::
getParamVectorFromCountVector(const Vector<float>& countVector) const
{
  return dpDeck.getParamVectorFromCountVector(countVector);
}
int
AlignerStrip::
getParamVectorSize() const
{
  return dpDeck.getParamVectorSize();
}
const char* 
AlignerStrip::
getParamNameFromOffset(int offset) const
{
  return dpDeck.getParamNameFromOffset(offset);
}
void
AlignerStrip::
setParamVector(const Vector<float>& pv)
{
  dpDeck.setParamVector(pv);
}
string 
AlignerStrip::
featureVect2String(const Vector<float>& fv, const int& indentLevel) const
{
  return dpDeck.featureVect2String(fv, indentLevel);
}
void
AlignerStrip::
setDPPath(const DPPath& dpPath1)
{
  dpPath = dpPath1;
  dpDeck.setDPPath(dpPath);
  initialDPPathIsSet = true;
}
void
AlignerStrip::
setSkipSize(int skipSize0)
{
  skipSize = skipSize0;
  if (skipSize > 0) {
    dpDeck.setSkipSize(skipSize0);
  }
}
int
AlignerStrip::
getSkipSize() const
{
  return dpDeck.getSkipSize();
}
void
AlignerStrip::
setPurpose(const string& purpose0)
{
  dpDeck.setPurpose(purpose0);
}
void 
AlignerStrip::
setCYKMode(const string& cyk_mode)
{
  dpDeck.setCYKMode(cyk_mode);
}
void
AlignerStrip::
setScoreMode(const string& modeName)
{
  dpDeck.setScoreMode(modeName);
}
void
AlignerStrip::
setDoRefine(bool doRefine0)
{// this is reset for each aligner.run
  dpDeck.setDoRefine(doRefine0);
}
void
AlignerStrip::
setRefinementStrategy(const string& strategy)
{
  dpDeck.setRefinementStrategy(strategy);
}
void
AlignerStrip::
setSequenceSimilarity(const float& z)
{
  dpDeck.setSequenceSimilarity(z);
}
void
AlignerStrip::
setDPAlongPath(bool dpAlongPath0)
{// this is reset for each aligner.run
  if (dpAlongPath0) {
    setMaxStripWidth(1);
  }
  dpDeck.setDPAlongPath(dpAlongPath0);
}

void
AlignerStrip::
setMemRequested(float memRequested0)
{
  memRequested = memRequested0;
  dpDeck.setMaxMemory(memRequested);
}
void
AlignerStrip::
setMaxStripWidth(int maxStripWidth0)
{
  maxStripWidth = maxStripWidth0;
}
void
AlignerStrip::
setKeepMinBandWidth(bool keepMinBandWidth0)
{
  keepMinBandWidth = keepMinBandWidth0;
}
void
AlignerStrip::
setMaxLWRatio(float maxLWRatio0)
{
  maxLWRatio = maxLWRatio0;
}
void
AlignerStrip::
setMinLWRatio(float minLWRatio0)
{
  minLWRatio = minLWRatio0;
}
void
AlignerStrip::
setMSAs(const MSA& msaX, const MSA& msaY)
{
  pMSAX = &msaX;
  pMSAY = &msaY;
  dpDeck.setMSAs(msaX, msaY);
}
void
AlignerStrip::
setMSAs(const MSA& msaX, const MSA& msaY,
	const UpTriMtx<float>& pairProbX,
	const UpTriMtx<float>& pairProbY)
{
  pMSAX = &msaX;
  pMSAY = &msaY;
  dpDeck.setMSAs(msaX, msaY, pairProbX, pairProbY);
}
void
AlignerStrip::
setScoreParam(ScoreParam& scoreParam)
{
  dpDeck.setScoreParam(scoreParam);
}
void
AlignerStrip::
setScoreParamList(Vector<ScoreParam>& scoreParamList)
{
  dpDeck.setScoreParamList(scoreParamList);
}
void
AlignerStrip::
setGuideTree(GuideTree& guideTree)
{
  dpDeck.setGuideTree(guideTree);
}
void
AlignerStrip::
setWeight(Weight& weights)
{
  dpDeck.setWeight(weights);
}
void
AlignerStrip::
setPairProbData(const PairProbData& pairProbData)
{
  dpDeck.setPairProbData(pairProbData);
}
void
AlignerStrip::
setAlignProb(const NMatrix<float>& alignProb)
{
  dpDeck.setAlignProb(alignProb);
  pAlignProb = &alignProb;
}
void
AlignerStrip::
setAlignProbData(const AlignProbData& alignProbData)
{
  dpDeck.setAlignProbData(alignProbData);
}
void 
AlignerStrip::
getAlignProb(NMatrix<float>& alignProb) const
{
  dpDeck.getAlignProb(alignProb);
}
void 
AlignerStrip::
getPairProb(UpTriMtx<float>& pairProb) const
{
  dpDeck.getPairProb(pairProb);
}
float
AlignerStrip::
getDEDistance() const
{
  return dpDeck.getDEDistance();
}
int
AlignerStrip::
getMinStripWidth() const 
{
  int minWidth = 1;
  if (minLWRatio >= 0.0) {
    int minWidth1 =  (int)(minLWRatio * (float)(getLengthX() + getLengthY()));
    minWidth = max(minWidth1, minWidth);
  }
  if (keepMinBandWidth) {
    int minBandWidth = getMinBandWidth();
    minWidth = max(minWidth, 2 * minBandWidth);
  }
  if (maxStripWidth > 0) {
    minWidth = min(minWidth, maxStripWidth);
  }
  return minWidth;
}
int
AlignerStrip::
getMaxStripWidth() const
{
  int minWidth0 = 1;
  float maxArea = (getLengthX() + 1) * (getLengthY() + 1);
  int maxStripWidth0 = 2 * (getLengthX() + getLengthY());
  int maxWidth = bisection(minWidth0, maxStripWidth0,
			   maxArea - 1, GetStripArea(*this));
  if (maxLWRatio >= 0.0) {
    int maxWidth1 =  (int)(maxLWRatio * maxWidth);
    maxWidth = max(minWidth0, min(maxWidth1, maxWidth));
  }
  if (maxStripWidth > 0) {
    maxWidth = max(minWidth0, min(maxStripWidth, maxWidth));
  }
  return maxWidth;
}
int
AlignerStrip::
findOptimalStripWidth()
{
  int minWidth = getMinStripWidth();
  int maxWidth = getMaxStripWidth();
  if (!((minWidth > 0) && (maxWidth > 0) && (minWidth <= maxWidth))) {
    Die("strip width constraint could not be satisfied");
  }
  int width = bisection(minWidth, maxWidth, memRequested,
			GetMemSize(*this));
#if true
  if (minWidth <= width) {
    const Rect& bbox = dpPath.getRect0();
    width = bisection(minWidth, maxWidth, maxTime,
		      GetTimeEstimateBBox(*this, bbox));
  }
#endif

  if (width < minWidth) {
    return (-1);
  } else {
    return width;
  }
}
int
AlignerStrip::
findOptimalStripWidthBBox(Rect bbox)
{
  int minWidth = getMinStripWidth();
  int maxWidth = getMaxStripWidth();
  if (!((minWidth > 0) && (maxWidth > 0) && (minWidth <= maxWidth))) {
    Die("strip width constraint could not be satisfied");
  }
  int width = bisection(minWidth, maxWidth, memRequested,
			GetMemSizeBBox(*this, bbox));
#if true
  if (minWidth <= width) {
    width = bisection(minWidth, maxWidth, maxTime,
		      GetTimeEstimateBBox(*this, bbox));
  }
#endif

  if (width < minWidth) {
    return (-1);
  } else {
    return width;
  }
}
void
AlignerStrip::
findOptimalSkipSizeStripWidthBBox(int& skipSize1, int& stripWidth1, Rect bbox)
{
  int minWidth = getMinStripWidth();
  int maxWidth = getMaxStripWidth();
  if (!((minWidth > 0) && (maxWidth > 0) && (minWidth <= maxWidth))) {
    Die("strip width constraint could not be satisfied");
  }
  if (skipSize1 > 0) {
    dpDeck.setSkipSize(skipSize1);
    stripWidth1 = findOptimalStripWidthBBox(bbox);
  } else {
    enum {
      MAX_SKIP_SIZE = 10
    };
    float z = 0.0;
    int best_skip_size = 0;
    int best_strip_width = 0;
    for (int ss = 1; ss <= MAX_SKIP_SIZE; ss++) {
      dpDeck.setSkipSize(ss);
      int width = bisection(minWidth, maxWidth, memRequested,
			    GetMemSizeBBox(*this, bbox));
      //cout << "mem_request: " << memRequested << endl;
      //cout << "width: " << width << endl;
      if (minWidth <= width) {
	width = bisection(minWidth, width, maxTime,
			  GetTimeEstimateBBox(*this, bbox));
      }
      if (minWidth <= width) {
	setStripDPRegion(dpRegion, dpPath, bbox, width);
	float z1 = (dpRegion.getArea() / pow((float)ss, skipSizeExponent));//objective function
#if 0
	cout << "[ ss: " << ss << ", z1: " << z1 << " ]\n";
	cout << "width: " << width << endl;
#if 0
	map<string, float> h = dpDeck.computeDPIterationCount();
	if (!h.empty()) {
	  cout << "\ndp_iteration_count:\n";
	  for (map<string, float>::const_iterator i = h.begin(); 
	       i != h.end(); i++) {
	    cout << "  " << i->first << ": " << i->second << "\n";
	  }
	}
#endif
#endif
	if (z < z1) {
	  z = z1;
	  best_skip_size = ss;
	  best_strip_width = width;
	}
      }
    }
    if (best_skip_size > 0) {
      skipSize1 = best_skip_size;
      stripWidth1 = best_strip_width;
      dpDeck.setSkipSize(skipSize1);
      setStripDPRegion(dpRegion, dpPath, bbox, stripWidth1);
      //dpDeck.setDPRegion(dpRegion);
    } else {
      skipSize1 = (-1);
      stripWidth1 = (-1);
      dpDeck.setSkipSize(1);
    }
  }
}
void
AlignerStrip::
setUpDPRegion()
{
  if (!initialDPPathIsSet) Die("Initial DPPath is not set");

#if false
  stripWidth = findOptimalStripWidth();
  //dpPath.getStripDPRegion(stripWidth, dpRegion);
  const Rect& bbox = dpPath.getRect0();
  setStripDPRegion(dpRegion, dpPath, bbox, stripWidth);
#else
  int skipSize1 = skipSize;
  int stripWidth1 = stripWidth;
  const Rect& bbox = dpPath.getRect0();
  findOptimalSkipSizeStripWidthBBox(skipSize1, stripWidth1, bbox);
  if (stripWidth1 <= 0) {
    Die("could not find DPRegion");
  } else {
    dpDeck.setSkipSize(skipSize1);
    setStripDPRegion(dpRegion, dpPath, bbox, stripWidth1);
  }
#endif
}
float
AlignerStrip::
computeWorkspaceSizeInMB()
{
  setUpDPRegion();
  return dpDeck.computeWorkspaceSizeInMB();
}
float
AlignerStrip::
computeDPIterationCountInUnitsOfMegaTimes()
{
  setUpDPRegion();
  return dpDeck.computeWorkspaceSizeInMB();
}
void
AlignerStrip::
printWorkspaceSize(ostream& fo)
{
  setUpDPRegion();
  dpDeck.printWorkspaceSize(fo);
}
void
AlignerStrip::
printDPIterationCounts(ostream& fo)
{
  setUpDPRegion();
  dpDeck.printDPIterationCounts(fo);
}
void
AlignerStrip::
run()
{
  Assert(dpPath.getLettLenX() == getLengthX());
  Assert(dpPath.getLettLenY() == getLengthY());
  if (!initialDPPathIsSet) Die("Initial DPPath is not set");

#if false
  if (skipSize <= 0) {
    int len = max(getLengthX(), getLengthY());
    CArray<int, 2> table[] = {
      {{150, 2}},
      {{200, 3}},
      {{250, 4}},
      {{300, 5}},
      {{350, 6}},
      {{400, 7}},
      {{500, 8}},
      {{600, 9}},
      {{700,10}}
    };
    int n = (sizeof(table) / sizeof(table[0]));
    int skipSize1 = 1;
    for (int i = 0; i < n; i++) {
      if (table[i][0] < len) {
	skipSize1 = table[i][1];
      }
    }
    dpDeck.setSkipSize(skipSize1);
  } else {
    dpDeck.setSkipSize(skipSize);
  }
#endif

  const Rect& bbox = dpPath.getRect0();
  bool onBubbles = false;
#if false
  stripWidth = findOptimalStripWidth();// memory constraints min, maxlwRatio are considered.
#else
  int skipSize1 = skipSize;
  int stripWidth1 = stripWidth;
  findOptimalSkipSizeStripWidthBBox(skipSize1, stripWidth1, bbox);
  stripWidth = stripWidth1;
  // cout << "stripWidth: " << stripWidth << endl;
#endif
  if (dpDeck.getPurpose() != DPDeck::CYK) {
    Check(stripWidth > 0, 
	  "could not satisfy strip width constraint or specified memory size too small");
  } else {
    if (stripWidth <= 0) {
      onBubbles = true;
      Warn("runOnBubbles do not supported");
      if (bubbleSize <= 0) {
	Warn("could not satisfy strip width constraint "
	     "or specified memory size too small. "
	     "switching to anchored alignment");
      }
    } else {
      if (0 < bubbleSize && bubbleSize < getLengthX()) {// only x direction will be accounted
	onBubbles = true;
	cout << "bubbleSize: " << bubbleSize << endl;
	Warn("runOnBubbles do not supported");
      }
    }
  }

  if (onBubbles) {
#if false
    runOnBubbles();
#else
    Warn("runOnBubbles do not supported");
    int skipSize1 = 1;
    stripWidth = 1;
    dpDeck.setSkipSize(skipSize1);
    setStripDPRegion(dpRegion, dpPath, bbox, stripWidth);
    dpDeck.setDPRegion(dpRegion);
    runDPRegion(dpRegion);
    dpPath = dpDeck.getDPPath();
#endif
  } else {
    //dpPath.getStripDPRegion(stripWidth, dpRegion);
#if true
    dpDeck.setSkipSize(skipSize1);
#endif
    setStripDPRegion(dpRegion, dpPath, bbox, stripWidth);
    dpDeck.setDPRegion(dpRegion);

#if true
    if (getType() == STRIP_PSCFG && maxStripWidth != 1) {
      ostringstream oss;
      oss << "length_x: " << getLengthX() << "\n";
      oss << "length_y: " << getLengthY() << "\n";
      oss << "skip_size: " << dpDeck.getSkipSize() << "\n";
      oss << "strip_width: " << (stripWidth / 2) << "\n";
      float t = getTimeEstimate();
      int sec = (int)(t * 60.0 + 1.0);//round up
      string str = sec2str(sec);
      oss << "estimated_time: " << "{ sec: " << sec << ", string: " << str << " }\n";
      oss << "workspace_size_MB: " << (int) dpDeck.computeWorkspaceSizeInMB() << "\n";
      // oss << "dp_area: " << dpRegion.getArea() << "\n";
      cout << oss.str();
      // LoadStat::os << oss.str();
    }
#endif

    runDPRegion(dpRegion);
    dpPath = dpDeck.getDPPath();
  }

  initialDPPathIsSet = false;
  maxStripWidth = (-1); // clear the stripWidth restriction
  dpDeck.setDPAlongPath(false); // clear the pairing constraint
  dpDeck.setDoRefine(false);// clear the pairing constraint
}
void
AlignerStrip::
runOnBubbles()
{
  const Rect& bbox0 = dpPath.getRect0();
  Rect bbox = bbox0;
  while (bbox.bx < bbox0.ex) {
    bbox = findOptimalBBox(bbox.bx, dpPath);
    Check(bbox.isNormal(),
	  "no bounding box that satisfies stripWidth constraint");
#if false
    stripWidth = findOptimalStripWidthBBox(bbox);
    //dpPath.getStripDPRegionBBox(stripWidth, bbox, dpRegion);
    setStripDPRegion(dpRegion, dpPath, bbox, stripWidth);
#else
    int skipSize1 = skipSize;
    int stripWidth1 = stripWidth;
    findOptimalSkipSizeStripWidthBBox(skipSize1, stripWidth1, bbox);
    stripWidth = stripWidth1;
    Check(stripWidth1 > 0);
    dpDeck.setSkipSize(skipSize1);
    setStripDPRegion(dpRegion, dpPath, bbox, stripWidth1);
#endif

#ifdef DEBUG
    cout << "bubble region: bbox " << bbox.toString() << " sw " << stripWidth << "\n";
    cout << dpRegion.toString();
    dpRegion.draw();
    cout << flush;
#endif
    runDPRegion(dpRegion);
    dpPath = dpDeck.getDPPath();
    //bbox.bx = max((bbox.ex - stripWidth), (((bbox.bx + 1) + bbox.ex) / 2));
    if (bbox.ex < bbox0.ex) {
      bbox.bx += max(1, (int)(bbox.lengthX() * bubbleOverlapRatio));
    } else {
      bbox.bx = bbox0.ex;
    }
  }
}
Rect
AlignerStrip::
findOptimalBBox(int x_beg, DPPath& dpPath1)
{
  DPPath::Iterator it(dpPath1);
  Rect bbox0 = dpPath1.getRect0();
  Rect bbox(Rect::null());
  Check(x_beg < bbox0.ex);
  Vector<int> y_begs;
  Vector<int> y_lasts;
  dpPath1.getYRange(y_begs, y_lasts);
  bbox.bx = x_beg;
  bbox.by = y_begs[x_beg];
  for (int x = (bbox.bx + 1); x <= bbox0.ex; x++) {
    if (0 < bubbleSize && bbox.isNormal() && bubbleSize < bbox.lengthX()) break;

    Rect bbox1 = bbox;
    bbox1.ex = x;
    bbox1.ey = y_lasts[x];
#if false
    int sw = findOptimalStripWidthBBox(bbox1);
#else
    int skipSize1 = skipSize;
    int stripWidth1 = stripWidth;
    findOptimalSkipSizeStripWidthBBox(skipSize1, stripWidth1, bbox);
    int sw = stripWidth1;
#endif
    if (sw > 0) {
      bbox.e = bbox1.e;
    }
  }
  if (bbox.isNormal()) {
    return bbox;
  } else {
    bbox.e = bbox0.e;
    int sw = findMaximalStripWidthBBox(bbox);
    if (sw > 0) {
      return bbox;
    } else {
      return Rect::null();
    }
  }
}
int
AlignerStrip::
findMaximalStripWidthBBox(Rect bbox)
{
  int minWidth = 1;
  int maxWidth = getMaxStripWidth();
  int bboxWidth = (bbox.lengthX() + bbox.lengthY());
  if (bboxWidth < maxWidth) {
    maxWidth = bboxWidth;
  }
  if (!((minWidth > 0) && (maxWidth > 0) && (minWidth <= maxWidth))) {
    Die("strip width constraint could not be satisfied");
  }

  // skipSize should be set
  int width = bisection(minWidth, maxWidth, memRequested,
			GetMemSizeBBox(*this, bbox));
#if 1
  if (minWidth <= width) {
    width = bisection(minWidth, width, maxTime,
		      GetTimeEstimateBBox(*this, bbox));
  }
#endif

  if (width < minWidth) {
    return (-1);
  } else {
    return width;
  }
}
void
AlignerStrip::
runDPRegion(const DPRegion& dpRegion1)
{
//#if 1
//  dpRegion1.draw();
//#endif
  dpDeck.setDPRegion(dpRegion1);
  float workspaceSize = dpDeck.computeWorkspaceSizeInMB();
  float dpIterationCount = 
    dpDeck.computeDPIterationCountInUnitsOfMegaTimes();
  LoadStat::addLengths(getLengthX(), getLengthY());
  LoadStat::addStripWidth(stripWidth);
  LoadStat::incrementDPCount();
  LoadStat::addWorkspaceSize(workspaceSize);
  LoadStat::addDPIterationCount(dpIterationCount);

#ifdef DEBUG
  cout << "---\n";
  cout << "alignerStrip:\n";
  cout << setprecision(3);
  cout << "  stripWidth: " << stripWidth << "\n";
  //cout << "  workspaceSize: " << workspaceSize << "MB" << endl;
  //cout << "  dpIterationCount: " << dpIterationCount << "MT" << endl;
  Assert( workspaceSize <= memRequested);
#endif

  dpDeck.run();

#ifdef DEBUG
  cout << "score=" << dpDeck.getScore() << endl;
#endif
}
void
AlignerStrip::
setStripDPRegion(DPRegion& dpRegion1, 
		 const DPPath& dpPath1, const Rect& bbox1, const int stripWidth1)
{
  dpPath.getStripDPRegionBBox(stripWidth1, bbox1, dpRegion1);
  if (shrinkByAlignProb) {
    shrinkDPRegion(dpRegion1, dpPath1);
  }
}
void
AlignerStrip::
shrinkDPRegion(DPRegion& dpRegion1, const DPPath& dpPath1)
{
  Vector<int> y_begs0;
  Vector<int> y_lasts0;
  Vector<int> y_begs;
  Vector<int> y_lasts;
  dpPath1.getYRange(y_begs0, y_lasts0);
  dpRegion1.getYRange(y_begs, y_lasts);
  const Rect& rect0 = dpPath1.getRect0();

  Assert(y_begs0.size() == y_begs.size());
  Assert((int)y_begs0.size() == rect0.lengthX());
  Check(pAlignProb);
  Check(pAlignProb->getRowLength() == rect0.lengthX() - 1);
  Check(pAlignProb->getColLength() == rect0.lengthY() - 1);
  const NMatrix<float>& mtx = (*pAlignProb);
  const float& cutoff = log(alignProbCutoff);

  int y_last0 = rect0.by;
  for (int i = 0; i < (int)y_lasts.size(); i++) {
    int x = (rect0.bx + i);
    Assert(y_lasts0[i] <= y_lasts[i]);
    y_last0 = max(y_last0, y_lasts0[i]);
    bool match_found = false;
    int y = y_lasts[i];
    for (;y >= y_last0; y--) {
      Assert(i == 0 || y_lasts[i - 1] <= y);
      if (x < rect0.ex && y < rect0.ey && cutoff <= mtx.ref(x, y)) {
	match_found = true;
	break;
      }
    }
    y_lasts[i] = max(y_last0, y);
    y_last0 = (match_found ? (y + 1) : max(y_last0, y));
  }
  int y_beg0 = rect0.ey;
  for (int i = ((int)y_begs.size() - 1); i > 0; i--) {
    int x = (rect0.bx + i);
    Assert(y_begs[i] <= y_begs0[i]);
    y_beg0 = min(y_beg0, y_begs0[i]);
    bool match_found = false;
    int y = y_begs[i];
    for (; y <= y_beg0; y++) {
      Assert((i == ((int)y_begs.size() - 1)) || y <= y_begs[i + 1]);
      if (rect0.bx < x && rect0.by < y && cutoff <= mtx.ref(x - 1, y - 1)) {
	match_found = true;
	break;
      }
    }
    y_begs[i] = min(y_beg0, y);
    y_beg0 = (match_found ? (y - 1) : min(y_beg0, y));
  }
  dpRegion1.setFromYRange(rect0, y_begs, y_lasts);
  //#ifdef DEBUG
  //dpRegion1.draw();
  //#endif
}
float
AlignerStrip::
getTimeEstimate() const
{
  return timeFactor * dpDeck.getTimeEstimate();
}

AlignerFold::
AlignerFold(int type, DPDeck& dpDeck)
  : Super(type),
    dpDeck(dpDeck),
    dpPath(*new DPPath),
    dpRegion(*new DPRegion),
    initialDPPathIsSet(false)
{
}
AlignerFold::
~AlignerFold()
{
  delete &dpPath;
  delete &dpRegion;
}
float 
AlignerFold::
getScore() const
{
  return dpDeck.getScore();
}
const DPPath&
AlignerFold::
getDPPath() const
{
  return dpDeck.getDPPath();
}
Vector<float>
AlignerFold::
computeFeatureVector() const
{
  return dpDeck.computeFeatureVector();
}
Vector<float>
AlignerFold::
getParamVector() const
{
  return dpDeck.getParamVector();
}
Vector<float>
AlignerFold::
getParamVectorFromCountVector(const Vector<float>& countVector) const
{
  return dpDeck.getParamVectorFromCountVector(countVector);
}
int
AlignerFold::
getParamVectorSize() const
{
  return dpDeck.getParamVectorSize();
}
const char* 
AlignerFold::
getParamNameFromOffset(int offset) const
{
  return dpDeck.getParamNameFromOffset(offset);
}
void
AlignerFold::
setParamVector(const Vector<float>& pv)
{
  dpDeck.setParamVector(pv);
}
string 
AlignerFold::
featureVect2String(const Vector<float>& fv, const int& indentLevel) const
{
  return dpDeck.featureVect2String(fv, indentLevel);
}
void
AlignerFold::
setDPPath(const DPPath& dpPath1)
{
  dpPath = dpPath1;
  dpDeck.setDPPath(dpPath);
  initialDPPathIsSet = true;
}
void
AlignerFold::
setSkipSize(int skipSize0)
{
  dpDeck.setSkipSize(skipSize0);
}
void
AlignerFold::
setPurpose(const string& purpose0)
{
  dpDeck.setPurpose(purpose0);
}
void 
AlignerFold::
setCYKMode(const string& cyk_mode)
{
  dpDeck.setCYKMode(cyk_mode);
}
void
AlignerFold::
setScoreMode(const string& modeName)
{
  dpDeck.setScoreMode(modeName);
}
void
AlignerFold::
setDoRefine(bool doRefine0)
{// this is reset for each aligner.run
  dpDeck.setDoRefine(doRefine0);
}
void
AlignerFold::
setRefinementStrategy(const string& strategy)
{
  dpDeck.setRefinementStrategy(strategy);
}
void
AlignerFold::
setSequenceSimilarity(const float& z)
{
  dpDeck.setSequenceSimilarity(z);
}
int
AlignerFold::
getSkipSize() const
{
  return dpDeck.getSkipSize();
}
void
AlignerFold::
setMemRequested(float memRequested0)
{
  dpDeck.setMaxMemory(memRequested0);
}
void
AlignerFold::
setDPAlongPath(bool dpAlongPath0)
{
  dpDeck.setDPAlongPath(dpAlongPath0);
}
void
AlignerFold::
setScoreParam(ScoreParam& scoreParam)
{
  dpDeck.setScoreParam(scoreParam);
}
void
AlignerFold::
setScoreParamList(Vector<ScoreParam>& scoreParamList)
{
  dpDeck.setScoreParamList(scoreParamList);
}
void
AlignerFold::
setGuideTree(GuideTree& guideTree)
{
  dpDeck.setGuideTree(guideTree);
}
void
AlignerFold::
setWeight(Weight& weights)
{
  dpDeck.setWeight(weights);
}
void
AlignerFold::
setPairProbData(const PairProbData& pairProbData)
{
  dpDeck.setPairProbData(pairProbData);
}
void
AlignerFold::
setAlignProbData(const AlignProbData& alignProbData)
{
  dpDeck.setAlignProbData(alignProbData);
}
//void 
//AlignerFold::
//getAlignProb(NMatrix<float>& alignProb) const
//{
//  dpDeck.getAlignProb(alignProb);
//}

void 
AlignerFold::
getPairProb(UpTriMtx<float>& pairProb) const
{
  dpDeck.getPairProb(pairProb);
}
void
AlignerFold::
setMSAs(const MSA& msaX, const MSA& msaY)
{
  dpDeck.setMSAs(msaX, msaY);
}
void
AlignerFold::
setMSAs(const MSA& msaX, const MSA& msaY,
	const UpTriMtx<float>& pairProbX,
	const UpTriMtx<float>& pairProbY)
{
  dpDeck.setMSAs(msaX, msaY, pairProbX, pairProbY);
}
void
AlignerFold::
setUpDPRegion()
{
  if (!initialDPPathIsSet) Die("Initial DPPath is not set");

  const int stripWidth = 1;
  dpPath.getStripDPRegion(stripWidth, dpRegion);
  dpDeck.setDPRegion(dpRegion);
}
void
AlignerFold::
run()
{
  setUpDPRegion();
#if defined(DEBUG) && 0
  float workspaceSize = dpDeck.computeWorkspaceSizeInMB();
  float dpIterationCount = 
    dpDeck.computeDPIterationCountInUnitsOfMegaTimes();
#endif
  //LoadStat::incrementDPCount();
  //LoadStat::addWorkspaceSize(workspaceSize);
  //LoadStat::addDPIterationCount(dpIterationCount);

#if defined(DEBUG) && 0
  cout << "---\n";
  cout << "alignerFold:\n";
  cout << setprecision(3);
  cout << "  workspaceSize: " << workspaceSize << "MB" << endl;
  cout << "  dpIterationCount: " << dpIterationCount << "MT" << endl;
#endif

  dpDeck.run();

#if defined(DEBUG) && 0
  cout << "score=" << dpDeck.getScore() << endl;
#endif

  dpPath = dpDeck.getDPPath();
  initialDPPathIsSet = false;
  dpDeck.setDPAlongPath(false); // clear the pairing constraint
  dpDeck.setDoRefine(false);// clear the pairing constraint
}
float
AlignerFold::
computeWorkspaceSizeInMB()
{
  setUpDPRegion();
  return dpDeck.computeWorkspaceSizeInMB();
}
float
AlignerFold::
computeDPIterationCountInUnitsOfMegaTimes()
{
  setUpDPRegion();
  return dpDeck.computeDPIterationCountInUnitsOfMegaTimes();
}
void
AlignerFold::
printWorkspaceSize(ostream& fo)
{
  setUpDPRegion();
  return dpDeck.printWorkspaceSize(fo);
}
void
AlignerFold::
printDPIterationCounts(ostream& fo)
{
  setUpDPRegion();
  return dpDeck.printDPIterationCounts(fo);
}

AlignerChunk::
AlignerChunk(int type, AlignerStrip& aligner)
  : Super(type),
    chunkSize(1),
    skipSize(1),
    nTailX(0), nTailY(0),
    dpPath(),
    aligner(aligner)
{
}
AlignerChunk::
~AlignerChunk()
{
  delete &aligner;
}
float
AlignerChunk::
getScore() const
{
  return aligner.getScore();
}
const DPPath&
AlignerChunk::
getDPPath() const
{
  return dpPath;
}
void
AlignerChunk::
setPairProbData(const PairProbData& pairProbData)
{
  aligner.setPairProbData(pairProbData);
}
void
AlignerChunk::
setAlignProbData(const AlignProbData& alignProbData)
{
  aligner.setAlignProbData(alignProbData);
}
void
AlignerChunk::
setRefinementStrategy(const string& strategy)
{
  aligner.setRefinementStrategy(strategy);
}
void 
AlignerChunk::
setCYKMode(const string& cyk_mode)
{
  aligner.setCYKMode(cyk_mode);
}
void 
AlignerChunk::
getAlignProb(NMatrix<float>& alignProb) const
{
  aligner.getAlignProb(alignProb);
}
void 
AlignerChunk::
getPairProb(UpTriMtx<float>& pairProb) const
{
  aligner.getPairProb(pairProb);
}
Vector<float>
AlignerChunk::
computeFeatureVector() const
{
  return aligner.computeFeatureVector();
}
Vector<float>
AlignerChunk::
getParamVector() const
{
  return aligner.getParamVector();
}
int
AlignerChunk::
getParamVectorSize() const
{
  return aligner.getParamVectorSize();
}
const char* 
AlignerChunk::
getParamNameFromOffset(int offset) const
{
  return aligner.getParamNameFromOffset(offset);
}
void
AlignerChunk::
setParamVector(const Vector<float>& pv)
{
  aligner.setParamVector(pv);
}
float
AlignerChunk::
computeWorkspaceSizeInMB()
{
  setUpAligner();
  return aligner.computeWorkspaceSizeInMB();
}
float
AlignerChunk::
computeDPIterationCountInUnitsOfMegaTimes()
{
  setUpAligner();
  return aligner.computeDPIterationCountInUnitsOfMegaTimes();
}
void
AlignerChunk::
printWorkspaceSize(ostream& fo)
{
  setUpAligner();
  aligner.printWorkspaceSize(fo);
}
void
AlignerChunk::
printDPIterationCounts(ostream& fo)
{
  setUpAligner();
  aligner.printDPIterationCounts(fo);
}
//void
//AlignerChunk::
//setLockFlag(bool lockFlag0)
//{
//  aligner.setLockFlag(lockFlag0);
//}
void
AlignerChunk::
setPurpose(const string& purpose0)
{
  aligner.setPurpose(purpose0);
}
void
AlignerChunk::
setDPAlongPath(bool dpAlongPath0)
{
  aligner.setDPAlongPath(dpAlongPath0);
}
void
AlignerChunk::
setScoreParam(ScoreParam& scoreParam)
{
  aligner.setScoreParam(scoreParam);
}
void
AlignerChunk::
setGuideTree(GuideTree& guideTree)
{
  aligner.setGuideTree(guideTree);
}
void
AlignerChunk::
setWeight(Weight& weights)
{
  aligner.setWeight(weights);
}

void
AlignerChunk::
setUpAligner()
{
  //profileX = *pProfileX;
  //profileY = *pProfileY;
  //nTailX = computeNTail(profileX.length(), chunkSize);
  //nTailY = computeNTail(profileY.length(), chunkSize);
  //profileX.chunk(chunkSize);
  //profileY.chunk(chunkSize);
  //aligner.setProfiles(profileX, profileY);

  if (pMSAY->size() > 0) {
    dpPath.setDiagonalPath(pMSAX->getUngappedLength(),
			   pMSAY->getUngappedLength());
  } else {
    dpPath.setDiagonalPath(pMSAX->getUngappedLength(),
			   pMSAX->getUngappedLength());
  }

  aligner.setDPPath(dpPath);
  aligner.setSkipSize(getSkipSizeChunk());
  aligner.setMaxLWRatio(getMaxLWRatioChunk());
}
void
AlignerChunk::
run()
{
  setUpAligner();

  bool accountForWidth = LoadStat::accountForWidth;//dirty method
  LoadStat::setAccountForWidth(false);
  aligner.run();
  LoadStat::setAccountForWidth(accountForWidth);//dirty method

  dpPath = aligner.getDPPath();
  dpPath.testConsistency();
  dpPath.expand(chunkSize, nTailX, nTailY);

  dpPath.testConsistency();
  Assert(dpPath.getLettLenX() == getLengthX());
  Assert(dpPath.getLettLenY() == getLengthY());
}



  
    


  
  
  

