/*
 *
 * Alpha.hpp
 *
 */
#ifndef ALPHA_HPP
#define ALPHA_HPP
#include "Util.hpp"

struct Alpha {
  enum {
    NBASE = 4,// A, C, G, U
    NBASEG = (NBASE + 1),   // A, C, G, U, Gap
    NBASENG = (NBASE + 2),    // A, C, G, U, N, Gap
    NPBASE = (NBASE * NBASE),
    NPBASEG = (NBASEG * NBASEG),
    NPBASENG = (NBASENG * NBASENG),
    NCODE = 7,
    BASEBIT = 2,
    BASEMSK = ~(~0 << BASEBIT)
  };
  enum BaseCode {
    A_CODE       = 0,
    C_CODE       = 1,
    G_CODE       = 2,
    U_CODE       = 3,
    N_CODE       = 4,
    GAP_CODE     = 5,
    INVALID_CODE = 16
  };
  enum PairCode {
    AA_CODE =  0, AC_CODE =  1, AG_CODE =  2, AU_CODE =  3,
    CA_CODE =  4, CC_CODE =  5, CG_CODE =  6, CU_CODE =  7,
    GA_CODE =  8, GC_CODE =  9, GG_CODE = 10, GU_CODE = 11,
    UA_CODE = 12, UC_CODE = 13, UG_CODE = 14, UU_CODE = 15,
    NN_CODE = 16, GP_CODE = 17, INVALID_PAIR_CODE = 16
  };
  enum ReducedPairCode {
    REDUCED_AU_CODE = 0,
    REDUCED_CG_CODE = 1,
    REDUCED_GC_CODE = 2,
    REDUCED_GU_CODE = 3,
    REDUCED_UA_CODE = 4,
    REDUCED_UG_CODE = 5,
    REDUCED_MM_CODE = 6,
    REDUCED_NN_CODE = 7,
    REDUCED_N_CAN_BASE = 6, //canonical bases
    REDUCED_NPBASE  = 7,
    REDUCED_NPBASEN = 8,
    REDUCED_INVALID_PAIR_CODE = 16
  };

  static const int N_CANONICAL = 6;
  static const int canonicalPairs[N_CANONICAL];
  static const int N_NON_CANONICAL = 10;
  static const int nonCanonicalPairs[N_NON_CANONICAL];
  static const int i2nt[NCODE];
  static const int nt2i[256];
  static const bool isCanonicalBasePair[NPBASE];

  static bool IsValidCode(const int& c) {return A_CODE <= c && c <= GAP_CODE;}
  static bool IsBaseCode(const int& c) {return (A_CODE <= c && c <= U_CODE);}
  static bool IsBasePairCode(const int& c){return (AA_CODE <= c && c <= UU_CODE);}
  static bool IsReducedBasePairCode(const int& c) {
    return (REDUCED_AU_CODE <= c && c <= REDUCED_MM_CODE);
  }
  static bool IsAmbiguousCode(const int& c) {return c == N_CODE;}
  static bool IsGapCode(const int& c) {return c == GAP_CODE;}
  static bool IsValidChar(const unsigned char& c) {
    return IsValidCode(nt2i[c]);
  }
  static bool IsBaseChar(const int& c) {return IsBaseCode(nt2i[c]);}
  static bool IsAmbiguousChar(const int& c) {return IsAmbiguousCode(nt2i[c]);}
  static bool IsGapChar(const int& c) {return IsGapCode(nt2i[c]);}
  static int nt2code(const int& nt) {
    if (0 <= nt && nt < 256) {
      return nt2i[nt];
    } else {
      return INVALID_CODE;
    }
  }
  static int getPairCode1(const int& c, const int& c1) {
    return (IsBaseCode(c) && IsBaseCode(c1) 
	    ? ((c << BASEBIT) | c1)
	    : INVALID_PAIR_CODE);
  }
  static int getPairCode(const int& c, const int& c1) {
    typedef CArray<CArray<PairCode,NBASENG>, NBASENG> Table;
    static const Table table = {{
      {{ AA_CODE,  AC_CODE,  AG_CODE,  AU_CODE,  NN_CODE, GP_CODE}},
      {{ CA_CODE,  CC_CODE,  CG_CODE,  CU_CODE,  NN_CODE, GP_CODE}},
      {{ GA_CODE,  GC_CODE,  GG_CODE,  GU_CODE,  NN_CODE, GP_CODE}},
      {{ UA_CODE,  UC_CODE,  UG_CODE,  UU_CODE,  NN_CODE, GP_CODE}},
      {{ NN_CODE,  NN_CODE,  NN_CODE,  NN_CODE,  NN_CODE, GP_CODE}},
      {{ GP_CODE,  GP_CODE,  GP_CODE,  GP_CODE,  GP_CODE, GP_CODE}}
    }};
    return table[c][c1];
  }
  static bool isValidPairCode(const int& pairCode) {
    return (0 <= pairCode && pairCode < NPBASE);
  }
  static bool isCanonicalPairCode(const int& pairCode) {
    int c = reducePairCode(pairCode);
    Check(c != REDUCED_INVALID_PAIR_CODE);
    return (c != REDUCED_MM_CODE);
  }
  static void pair2Bases(const int& pairCode, int& c, int& c1) {
    Assert(IsBasePairCode(pairCode));
    c1 = pairCode & BASEMSK;
    c = (pairCode >> BASEBIT) & BASEMSK;
  }
  static int getReducedPairCode(const int& c, const int& c1) {
    typedef CArray<CArray<ReducedPairCode,NBASENG>,NBASENG> Table;
    static const Table table = {{
      {{
	REDUCED_MM_CODE,//AA
	REDUCED_MM_CODE,//AC
	REDUCED_MM_CODE,//AG
	REDUCED_AU_CODE,//AU
	REDUCED_MM_CODE,//AN
	REDUCED_NN_CODE //A-
      }},
      {{
	REDUCED_MM_CODE,//CA
	REDUCED_MM_CODE,//CC
	REDUCED_CG_CODE,//CG
	REDUCED_MM_CODE,//CU
	REDUCED_MM_CODE,//CN
	REDUCED_NN_CODE //C-
      }},
      {{
	REDUCED_MM_CODE,//GA
	REDUCED_GC_CODE,//GC
	REDUCED_MM_CODE,//GG
	REDUCED_GU_CODE,//GU
	REDUCED_MM_CODE,//GN
	REDUCED_NN_CODE //G-
      }},
      {{
	REDUCED_UA_CODE,//UA
	REDUCED_MM_CODE,//UC
	REDUCED_UG_CODE,//UG
	REDUCED_MM_CODE,//UU
	REDUCED_MM_CODE,//UN
	REDUCED_NN_CODE //U-
      }},
      {{
	REDUCED_MM_CODE,//NA
	REDUCED_MM_CODE,//NC
	REDUCED_MM_CODE,//NG
	REDUCED_MM_CODE,//NU
	REDUCED_MM_CODE,//NN
	REDUCED_NN_CODE //N-
      }},
      {{
	REDUCED_NN_CODE,//-A
	REDUCED_NN_CODE,//-C
	REDUCED_NN_CODE,//-G
	REDUCED_NN_CODE,//-U
	REDUCED_NN_CODE,//-N
	REDUCED_NN_CODE //--
      }}
    }};   
    return table[c][c1];
  }
  static int reducePairCode(const int& pairCode) {
    typedef CArray<ReducedPairCode,NPBASE> Table;
    static const Table table = {{
      REDUCED_MM_CODE, REDUCED_MM_CODE, REDUCED_MM_CODE, REDUCED_AU_CODE,
      REDUCED_MM_CODE, REDUCED_MM_CODE, REDUCED_CG_CODE, REDUCED_MM_CODE,
      REDUCED_MM_CODE, REDUCED_GC_CODE, REDUCED_MM_CODE, REDUCED_GU_CODE,
      REDUCED_UA_CODE, REDUCED_MM_CODE, REDUCED_UG_CODE, REDUCED_MM_CODE,
    }};
    return (IsBasePairCode(pairCode)
	    ? table[pairCode]
	    : REDUCED_INVALID_PAIR_CODE);
  }
  static bool isValidReducedPairCode(const int& pairCode) {
    return (0 <= pairCode && pairCode < REDUCED_NPBASEN);
  }
  static bool isCanonicalReducedPairCode(const int& pairCode) {
    return (REDUCED_AU_CODE <= pairCode
	    && pairCode <= REDUCED_UG_CODE);
  }
  static int flipPairCode(const int& pairCode) {
    if (!IsBasePairCode(pairCode)) return pairCode;
    int c, c1;
    pair2Bases(pairCode, c, c1);
    return getPairCode(c1, c);
  }
  static int flipReducedPairCode(const int& reducedPairCode) {
    static const int table[REDUCED_NPBASE + 1] = {
      REDUCED_UA_CODE,
      REDUCED_GC_CODE,
      REDUCED_CG_CODE,
      REDUCED_UG_CODE,
      REDUCED_AU_CODE,
      REDUCED_GU_CODE,
      REDUCED_MM_CODE,
      REDUCED_INVALID_PAIR_CODE
    };
    return (IsReducedBasePairCode(reducedPairCode)
	    ? table[reducedPairCode] : table[REDUCED_NPBASE]);
  }
  static void seq2i(char* s, const char* t, int len) {
    const char* const s_end = s + len;
    while (s < s_end) *s++ = nt2i[(unsigned) (*t++)];
  }
  static void i2seq(char* s, const char* t, int len) {
    const char* const s_end = s + len;
    while (s < s_end) *s++ = i2nt[(unsigned) (*t++)];
  }
  static void i2seq(ostream& fo, const char* t, int len) {
    const char* const t_end = t + len;
    while (t < t_end) fo << (char) i2nt[(unsigned) (*t++)];
  }
  static char* wd2str(unsigned int wdSize, unsigned int wd) {
    const unsigned int MAX_WD_SIZE = (sizeof(unsigned int) * 8 / BASEBIT);
    static char buf[MAX_WD_SIZE + 1] = {};
    Assert(wdSize <= MAX_WD_SIZE);

    char* s = buf + wdSize;
    *s = '\0';
    do {
      *(--s) = Alpha::i2nt[wd & BASEMSK];
      wd >>= BASEBIT;
    } while (s > buf);
    return buf;
  }
  static void printWd(ostream& fo, unsigned int wdSize, unsigned int wd) {
    fo << wd2str(wdSize, wd);
  }
  static const char* code2str(const int& code) {
    static const char table[NBASENG+1][2] = {
      "A", "C", "G", "U", "N", ".", "?"
    };
    return ((A_CODE <= code && code <= GAP_CODE)
	    ? table[code] : table[NBASENG]);
  }
  static const char* pairCode2str(const int& pairCode) {
    static const char table[NPBASE+1][3] = {
      "AA", "AC", "AG", "AU",
      "CA", "CC", "CG", "CU",
      "GA", "GC", "GG", "GU",
      "UA", "UC", "UG", "UU",
      "??"
    };
    return (IsBasePairCode(pairCode) ? table[pairCode] : table[NPBASE]);
  }
  static const char* reducedPairCode2str(const int& reducedPairCode) {
    static const char table[REDUCED_NPBASE+1][3] = {
      "AU", "CG", "GC", "GU", "UA", "UG", "MM", "??"
    };
    return (IsReducedBasePairCode(reducedPairCode)
	    ? table[reducedPairCode] : table[REDUCED_NPBASE]);
  }
  static char nt2Code(const char& c){
    if (!IsValidChar(c)) Die("character %c is not valid", c);

    return (char) nt2i[(int) c];
  }
  static char code2char(const int& c) {
     static const char table[NBASENG+1] = {
      'A', 'C', 'G', 'U', 'N', '.', '?'
    };
    return ((A_CODE <= c && c <= GAP_CODE)
	    ? table[(int) c] : table[NBASENG]);
  }
  static string generateRandomRNA(const int& len) {
    static const char nt[5] = "ACGU";
    string rna(len, '\0');
    for (int i = 0; i < len; i++) {
      rna[i] = nt[Rand(4)];
    }
    return rna;
  }
  static void addBaseCodeCount(CArray<int, NBASENG>& counts, const char* s, int len) {
    // counts.fill(0); add count
    while (len-- > 0) {
      counts[*s++]++;
    }
  }
  inline static int getNStack() {
    return (Alpha::REDUCED_NPBASE * (Alpha::REDUCED_NPBASE + 1)) / 2;
  }
  struct GetStackIndex {
    enum{N = Alpha::REDUCED_NPBASE};
    typedef CMatrix<int, N, N> Table;
    Table table;
    inline GetStackIndex() {
      table.fill(-1);
      int k = 0;
      for (int i = 0; i < N; i++) {
	for (int j = 0; j < N; j++) {
	  if (table.ref(i, j) == (-1)) {
	    int i1 = flipReducedPairCode(j);
	    int j1 = flipReducedPairCode(i);
	    table.ref(i, j) = table.ref(i1, j1) = k++;
	  }
	}
      }
      Assert(k == getNStack());
    }
  };
  inline static int getStackIndex(int i, int j) {
    static const GetStackIndex g;
    return g.table.ref(i, j);
  }
};

#endif
