/*
 * Cluster.cpp
 *
 */
#include "Dmtx.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"

Cluster::
Cluster()
  : pDmtx(NULL), distType(WAVG_DIST), guideTree(),
    nexts(), prevs(), nnbrs(), ndidx(), minDists(), wgts(), hgts(),
    use_similarity(false)
{
}
Cluster::
~Cluster()
{
}
void
Cluster::
setDmtx(Dmtx& dmtx)
{
  pDmtx = &dmtx;
  guideTree.setSeqIds(dmtx.getSeqIds());
  int nLeaf = guideTree.getNLeaf();

  nexts.resize(nLeaf);
  prevs.resize(nLeaf);
  nnbrs.resize(nLeaf);
  ndidx.resize(nLeaf);
  minDists.resize(nLeaf);
  wgts.resize(nLeaf);
  hgts.resize(nLeaf);
}
void
Cluster::
setDistType(const string& distTypeName)
{
  static const struct {string name; DistType type;} table[] = {
    {"MIN_DIST", MIN_DIST},
    {"AVG_DIST", AVG_DIST},
    {"MAX_DIST", MAX_DIST},
    {"WAVG_DIST", WAVG_DIST},
    {"AVG2_DIST", AVG2_DIST},// for dij > 1 root distance may overflow
  };
  static const int N = sizeof(table) / sizeof(table[0]);
  for (int i = 0; i < N; i++) {
    if (distTypeName == table[i].name) {
      distType = table[i].type;
      return;
    }
  }
  Die("bad distTypeName %s", distTypeName.c_str());
}
const string&
Cluster::
getDistTypeName() const
{
  static const string nullName = "??";
  static const struct {string name; DistType type;} table[] = {
    {"MIN_DIST", MIN_DIST},
    {"AVG_DIST", AVG_DIST},
    {"MAX_DIST", MAX_DIST},
    {"WAVG_DIST", WAVG_DIST},
    {"AVG2_DIST", AVG2_DIST}
  };
  static const int N = sizeof(table) / sizeof(table[0]);
  for (int i = 0; i < N; i++) {
    if (distType == table[i].type) {
      return table[i].name;
    }
  }
  return nullName;
}
void
Cluster::
run()
{
  buildTreeInit();
  buildTreeRun();
}
bool
Cluster::
isLess(float a, float b) const
{
  if (use_similarity) {
    return (a > b);
  } else {
    return (a < b);
  }
}
void
Cluster::
buildTreeInit()
{
  Assert(pDmtx);
  const float INF = (use_similarity ? (-HUGE_VAL) : HUGE_VAL);
  const Dmtx& dmtx = *pDmtx;
  const int nLeaf = guideTree.getNLeaf();

  for (int i = 0; i < nLeaf; i++) {
    ndidx[i] = i;
    nexts[i] = ((i == nLeaf - 1) ? NULL_INDEX : (i + 1));
    prevs[i] = ((i == 0) ? NULL_INDEX : (i - 1));
    wgts[i] = 1.0;
    hgts[i] = 0.0;
  }
  for (int j = 0; j < nLeaf; j++) {
    float dmin = INF;
    int nnbr = NULL_INDEX;
    for (int i = 0; i < j; i++) {
      const float& z = dmtx.ref(i, j);
      if (isLess(z, dmin)) {
	dmin = z;
	nnbr = i;
      }
      if (isLess(z, minDists[i])) {
	minDists[i] = z;
	nnbrs[i] = j;
      }
    }
    minDists[j] = dmin;
    nnbrs[j] = nnbr;
  }
}

void
Cluster::
buildTreeRun()
{
  Assert(pDmtx);
  const float INF = (use_similarity ? (-HUGE_VAL) : HUGE_VAL);
  Dmtx& dmtx = *pDmtx;
  const int nLeaf = guideTree.getNLeaf();
  const int nNode = guideTree.getNNode();

  float dmin0 = INF;
  for (int i = 0; i < nLeaf; i++) {
    for (int j = (i + 1); j < nLeaf; j++) {
      if (isLess(dmtx.ref(i, j), dmin0)) {
	dmin0 = dmtx.ref(i, j);
      }
    }
  }

  for(int n = nLeaf; n < nNode; n++){
    // find mindist (i0,j0)
    int i0 = NULL_INDEX;
    float dmin = INF;
    for (int i = 0; i != NULL_INDEX; i = nexts[i]) {
      if (isLess(minDists[i], dmin)) {
	dmin = minDists[i];
	i0 = i;
      }
    }
    Assert(0 <= i0 && i0 < nLeaf);

    int j0 = nnbrs[i0];
    if (i0 > j0) {int i = i0; i0 = j0; j0 = i;}// swap s.t. i0<j0
  
    //fill row i0
    minDists[i0] = INF;
    nnbrs[i0] = NULL_INDEX;
    for (int j = 0; j != NULL_INDEX; j = nexts[j]) {
      if (j == i0 || j == j0) continue;

      const float& z = 
	calcDist(dmtx.ref(i0, j), dmtx.ref(j, j0), dmin, wgts[i0], wgts[j0]);
      dmtx.ref(i0, j) = z;

      if (isLess(z, minDists[i0])) {
	minDists[i0] = z;
	nnbrs[i0] = j;
      }
      if (distType == (use_similarity ? MIN_DIST : MAX_DIST)) {	
	if(nnbrs[j] == j0)
	  nnbrs[j] = i0;
      } else {
	if (nnbrs[j] == i0 || nnbrs[j] == j0) {
	  float dmin1 = INF;
	  int nnbr1 = NULL_INDEX;
	  for (int k = 0; k != NULL_INDEX; k = nexts[k]) {
	    if(k == j || k == j0) continue;

	    const float& z = dmtx.ref(j, k);
	    if(isLess(z, dmin1)){
	      dmin1 = z;
	      nnbr1 = k;
	    }
	  }
	  minDists[j] = dmin1;
	  nnbrs[j] = nnbr1;
	}
      }
    }
    float hgt0 = 0.0;
    float eleni0 = 0.0;
    float elenj0 = 0.0;
    if (use_similarity) {
      // use_similarity case, edge lengths are not meaningful
      hgt0 = 0.5 * (1.0 + (dmin0 - dmin)) / (1.0 + abs(dmin0));
      eleni0 = abs(hgt0 - hgts[i0]);
      elenj0 = abs(hgt0 - hgts[j0]);
    } else {
      hgt0 = dmin / 2.0;
      eleni0 = (hgt0 - hgts[i0]);
      elenj0 = (hgt0 - hgts[j0]);
      Assert(eleni0 >= 0.0);
      Assert(elenj0 >= 0.0);
    }
    guideTree.joinNodes(n, ndidx[i0], ndidx[j0], eleni0, elenj0);
    
    //if (prevs[j0] != NULL_INDEX) this is unnecessary
    nexts[prevs[j0]] = nexts[j0];
    if (nexts[j0] != NULL_INDEX)
      prevs[nexts[j0]] = prevs[j0];
    //nexts[j0] = NULL_INDEX;
    //prevs[j0] = NULL_INDEX;

    ndidx[i0] = n;
    //ndidx[j0] = NULL_INDEX;
    wgts[i0] = (wgts[i0] + wgts[j0]);
    //wgts[j0] = 0;
    hgts[i0] = hgt0;
    //hgts[j0] = 0;
    //print(stdout);
  }
}
string
Cluster::
toString() const
{
  if (!pDmtx) return "";

  const Dmtx& dmtx = *pDmtx;
  const int nNode = guideTree.getNNode();
  const int nLeaf = guideTree.getNLeaf();
  ostringstream oss;
  oss << setprecision(2);
  oss << "distType: " << getDistTypeName() << ' ';
  oss << "nLeaf: " << nLeaf << " nNode: " << nNode << '\n';

  for (int i = 0; i != NULL_INDEX; i = nexts[i]) {
    oss << i << " nd:" << ndidx[i] << " w:" << wgts[i];
    oss << " h:" << hgts[i] << " nbr:" << nnbrs[i];
    oss << " dmin:" << minDists[i] << ";\t";
    for (int j = 0; j != NULL_INDEX; j = nexts[j]) {
      if (i == j) {
	oss << setw(4) << 'x'  << ' ';
      } else {
	oss << setw(4) << dmtx.ref(i, j) << ' ';
      }
    }
    oss << '\n';
  }
  return oss.str();
}
void
Cluster::
print() const
{
  cout << toString() << flush;
}
float
Cluster::
calcDist(float d1, float d2, float d12, float wgt1, float wgt2)
{
  switch (distType) {
  case MIN_DIST: return min(d1, d2);
  case AVG_DIST: return (d1 + d2) / 2.0;
  case MAX_DIST: return max(d1, d2);
  case WAVG_DIST: return (d1 * wgt1 + d2 * wgt2) / (wgt1 + wgt2);
  case AVG2_DIST: return (d1 + d2) * d12 / 2.0;
  default: Die("bad dist type"); return IMPOSSIBLE;
  }
}
Cluster_NJ::
Cluster_NJ()
  : pDmtx(NULL), guideTree()
{
}
Cluster_NJ::
~Cluster_NJ()
{
}
void
Cluster_NJ::
setDmtx(Dmtx& dmtx)
{
  pDmtx = &dmtx;
  guideTree.setSeqIds(dmtx.getSeqIds());
}
void
Cluster_NJ::
run()
{
  Array<bool> is_active(guideTree.getNNode());
  is_active.fill(false);
  for (int i = 0; i < (int)guideTree.getNLeaf(); i++) {
    is_active[i] = true;
  }
  Matrix<float> dmtx;
  dmtx.setLengths(guideTree.getNNode(), guideTree.getNNode());
  dmtx.fill(0.0);
  for (int i = 0; i < pDmtx->length(); i++) {
    for (int j = (i + 1); j < pDmtx->length(); j++) {
      dmtx.ref(i, j) = dmtx.ref(j, i) = pDmtx->ref(i, j);
    }
  }
  Array<float> r(guideTree.getNNode());
  r.fill(0.0);
  
  int n = guideTree.getNNode();
  int m = guideTree.getNLeaf();
  for (int k = m; k < (n - 1); k++) {

    for (int i = 0; i < k; i++) {
      if (!is_active[i]) continue;

      r[i] = 0.0;
      for (int j = 0; j < k; j++) {
	if (!is_active[j]) continue;

	r[i] += dmtx.ref(i, j);// diagonal part (i == j) of dmtx is zero
      }
      r[i] /= (float)(m - 2);
    }
    float dmin = 100000.0;
    int imin = (-1);
    int jmin = (-1);
    for (int i = 0; i < k; i++) {
      if (!is_active[i]) continue;

      for (int j = (i + 1); j < k; j++) {
	if (!is_active[j]) continue;

	float dij = dmtx.ref(i, j) - (r[i] + r[j]);
	if (dij < dmin) {
	  dmin = dij;
	  imin = i;
	  jmin = j;
	}
      }
    }
    Assert(imin >= 0 && jmin >= 0);

    float dik = 0.5 * (dmin + r[imin] - r[jmin]);
    float djk = (dmin - dik);
    if (dik < 0.0) {
      djk += abs(dik);
      dik = 0.0;
    }
    if (djk < 0.0) {
      dik += abs(djk);
      djk = 0.0;
    }
    Check(dik >= 0.0 && djk >= 0.0);
    guideTree.joinNodes(k, imin, jmin, dik, djk);
    
    for (int i = 0; i < k; i++) {
      if (!is_active[i]) continue;

      float d = 0.5 * (dmtx.ref(imin, i) + dmtx.ref(jmin, i) - dmin);
      dmtx.ref(i, k) = dmtx.ref(k, i) = d;
    }
    is_active[imin] = false;
    is_active[jmin] = false;
    is_active[k] = true;
    m--;
  }

  float dmin = 100000.0;
  int imin = (-1);
  int jmin = (-1);
  for (int i = 0; i < (n - 1); i++) {
    if (!is_active[i]) continue;

    for (int j = (i + 1); j < (n - 1); j++) {
      if (!is_active[j]) continue;

      float dij = dmtx.ref(i, j);
      if (dij < dmin) {
	Assert(imin == (-1) && jmin == (-1));
	dmin = dij;
	imin = i;
	jmin = j;
      }
    }
  }
  Assert(imin >= 0 && jmin >= 0);
  float dik = (0.5 * dmin);
  float djk = (0.5 * dmin);
  Check(dik >= 0.0 && djk >= 0.0);
  guideTree.joinNodes((n - 1), imin, jmin, dik, djk);
}
string
Cluster_NJ::
toString() const
{
  const int nNode = guideTree.getNNode();
  const int nLeaf = guideTree.getNLeaf();
  ostringstream oss;
  oss << "nLeaf: " << nLeaf << " nNode: " << nNode << '\n';
  return oss.str();
}
void
Cluster_NJ::
print() const
{
  cout << toString() << flush;
}
