/*
 * DPScore.hpp
 *
 */
#ifndef DP_SCORE_HPP
#define DP_SCORE_HPP
#include "Util.hpp"

//#define DPSCORE_INT_TYPE
//#define DPSCORE_FLOAT_TYPE

class DPScore {
public:
#ifdef DPSCORE_INT_TYPE
  typedef int DPScoreT;
#else
  typedef float DPScoreT;
  //typedef double DPScoreT;
#endif

  template <typename T> struct LogF;
  template <typename T> struct LogF_Simple;
  class FVecSimple;
  class FVecLogF;

  inline static const DPScoreT& getNegInfDPScore() {
#ifdef DPSCORE_INT_TYPE
    static const DPScoreT NegInfDPScore_ = 
      (DPScoreT)((float)INT_MIN / (float)10.0);
#else
    static const DPScoreT NegInfDPScore_ = IMPOSSIBLE;
#endif
    return NegInfDPScore_;
  }
  inline static DPScoreT getDPScoreThreshold() {
    static const DPScoreT thr = (DPScoreT)(0.9 * (float) getNegInfDPScore());
    return thr;
  }
  inline float getFloatThreshold() const {
    return (float)getDPScoreThreshold();
  }
  inline static const DPScoreT& getZeroDPScore() {
    static const DPScoreT ZeroDPScore_ = 0;
    return ZeroDPScore_;
  }
  inline static bool IsFiniteDPScore(const DPScoreT& z) {
    return (z > getDPScoreThreshold());
  }
  class FuncFloat2DPScore {
  public:
    FuncFloat2DPScore(const DPScore& dps) : dps(dps) {}
    FuncFloat2DPScore(const FuncFloat2DPScore& func) : dps(func.dps) {}
    ~FuncFloat2DPScore() {}
    DPScoreT operator()(const float& z) const {return dps.float2DPScore(z);}
  private:
    const DPScore& dps;
    FuncFloat2DPScore();
    FuncFloat2DPScore& operator=(const FuncFloat2DPScore& func);
  };

  DPScore() : dpScoreScale(getDefaultDPScoreScale()) {}
  float getDPScoreScale() const {return dpScoreScale;}
  void setDPScoreScale(const float& z) {dpScoreScale = z;}

  inline DPScoreT float2DPScore(float z) const {
    // cannot write 'const float&' to use bind1st
    const float& z1 = (z * dpScoreScale);
    if (z1 <= getFloatThreshold()) {
      return getNegInfDPScore();
    } else {
      return (DPScoreT)z1;
    }
  }
  // tentative
  inline DPScoreT float2DPScore_SQRT(float z) const {
    // cannot write 'const float&' to use bind1st
    const float& z1 = (z * sqrt(dpScoreScale));
    if (z1 <= getFloatThreshold()) {
      return getNegInfDPScore();
    } else {
      return (DPScoreT)z1;
    }
  }
  inline float dpScore2Float(const DPScoreT& z) const {
    return ((float)(z) / dpScoreScale);
  }

#ifdef DPSCORE_INT_TYPE
  inline void logAdd(DPScoreT& a, const DPScoreT& b) const {
    const float& T = dpScoreScale;
    if (a >= b) {
      a += (DPScoreT)(T * log(1.0 + exp((float)(b - a)/ T)));
    } else {
      a = b + (DPScoreT)(T * log(exp((float)(a - b) / T) + 1.0));
    }
  }
#else
  // assuming dpScoreScale = 1.0 and DPScoreT = float
  static void logAdd(DPScoreT& a, const DPScoreT& b) {
    logAddBasic(a, b);
    //logAdd_ProbCons(a, b);
  }
#endif

  FuncFloat2DPScore getFuncFloat2DPScore() const {
    return FuncFloat2DPScore(*this);
  }
private:
  float dpScoreScale;
  static float getDefaultDPScoreScale() {
#ifdef DEBUG
#ifdef DPSCORE_INT_TYPE
    return 1000.0;
#else
    //return 2.0;
    return 1.0;
#endif
#else
#ifdef DPSCORE_INT_TYPE
    return 500.0;
#else
    return 1.0;
#endif
#endif
  }
};

#if 0
template <typename T>
struct 
DPScore::
LogF 
{
  T z_p;
  T z_n;

  inline static const LogF& null() {
    static const LogF nullVal = {getNegInf(), getNegInf()};
    return nullVal;
  }
  inline static T getNegInf() {
    return (-FLT_MAX * 0.01);
  }
  inline static T getMinExpVal() {
    return 1.0e-15;
  }
  inline LogF& setExpVal(const T& z) {
    if (z > getMinExpVal()) {
      z_p = log(z);
      z_n = getNegInf();
    } else if (z < -getMinExpVal()) {
      z_p = getNegInf();
      z_n = log(-z);
    } else {
      *this = null();
    }
    return *this;
  }
  inline LogF& setLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      z_p = z;
      z_n = getNegInf();
    } else {
      z_p = getNegInf();
      z_n = z;
    }
    return *this;
  }
  inline LogF& setVal(const LogF& logF) {
    z_p = logF.z_p;
    z_n = logF.z_n;
    return *this;
  }
  inline T getExpVal() const {
    if (z_p >= z_n) {
      return exp(z_p) * (1 - exp(z_n - z_p));
    } else {
      return (- exp(z_n) * (1 - exp(z_p - z_n)));
    }
  }
  inline void getLogVal(T& z, bool& sgn) const {
    if (z_p >= z_n) {
      z = (z_p + log(1 - exp(z_n - z_p)));
      sgn = true;
    } else {
      z = (z_n + log(1 - exp(z_p - z_n)));
      sgn = false;
    }
  }
  inline LogF& addExpVal(const T& z) {
    if (z > getMinExpVal()) {
      logAddBasic(z_p, log(z));
    } else if (z < -getMinExpVal()) {
      logAddBasic(z_n, log(-z));
    }
    return *this;
  }
  inline LogF& addLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      logAddBasic(z_p, z);
    } else {
      logAddBasic(z_n, z);
    }
    return *this;
  }
  inline LogF& add(const LogF& logF) {
    logAddBasic(z_p, logF.z_p);
    logAddBasic(z_n, logF.z_n);
    return *this;
  }
  inline LogF& subtract(const LogF& logF) {
    logAddBasic(z_p, logF.z_n);
    logAddBasic(z_n, logF.z_p);
    return *this;
  }
  inline LogF& multiplyExpVal(const T& z) {
    if (z > getMinExpVal()) {
      z_p += log(z);
      z_n += log(z);
    } else if (z < -getMinExpVal()) {
      std::swap(z_p, z_n);
      z_p += log(-z);
      z_n += log(-z);
    } else {
      *this = null();
    }
    return *this;
  }
  inline LogF& multiplyLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      z_p += z;
      z_n += z;
    } else {
      std::swap(z_p, z_n);
      z_p += z;
      z_n += z;
    }
    return *this;
  }
  inline LogF& multiply(const LogF& logF) {
    const T& z_p1 = z_p + logF.z_p;
    const T& z_p2 = z_n + logF.z_n;
    const T& z_n1 = z_p + logF.z_n;
    const T& z_n2 = z_n + logF.z_p;
    z_p = z_p1;
    z_n = z_n1;
    logAddBasic(z_p, z_p2);
    logAddBasic(z_n, z_n2);
    return *this;
  }
  inline LogF& operator=(const T& z1) {
    return setExpVal(z1);
  }
  inline LogF& operator+=(const LogF& z1) {
    return add(z1);
  }
  inline LogF& operator-=(const LogF& z1) {
    return addLogVal(z1.z, !(z1.sgn));
  }
  inline LogF& operator*=(const LogF& z1) {
    return multiply(z1);
  }
  inline LogF& operator/=(const LogF& z1) {
    return multiplyLogVal(-z1.z, z1.sgn);
  }
  inline LogF operator+(const LogF& z1) const {
    LogF z2(*this);
    return (z2 += z1);
  }
  inline LogF operator-(const LogF& z1) const {
    LogF z2(*this);
    return (z2 -= z1);
  }
  inline LogF operator*(const LogF& z1) const {
    LogF z2(*this);
    return (z2 *= z1);
  }
  inline LogF operator/(const LogF& z1) const {
    LogF z2(*this);
    return (z2 /= z1);
  }
  //inline operator T() const {
  //  return getExpVal();
  //}
};

#elif 1
template <typename T>
struct 
DPScore::
LogF 
{
  T w;
  inline static const LogF& null() {
    static const LogF nullVal = {getNegInf()};
    return nullVal;
  }
  inline static T getNegInf() {
    return (0.0);
  }
  inline static T getMinExpVal() {
    return 0.0;
  }
  inline LogF& setExpVal(const T& z) {
    w = z;
    return *this;
  }
  inline LogF& setLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w = exp(z);
    } else {
      w = -exp(z);
    }
    return *this;
  }
  inline LogF& setVal(const LogF& logF) {
    w = logF.w;
    return *this;
  }
  inline T getExpVal() const {
    return w;
  }
  inline void getLogVal(T& z, bool& sgn) const {
    if (w >= 0) {
      z = log(w);
      sgn = true;
    } else {
      z = log(-w);
      sgn = false;
    }
  }
  inline LogF& addExpVal(const T& z) {
    w += z;
    return *this;
  }
  inline LogF& addLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w += exp(z);
    } else {
      w += -exp(z);
    }
    return *this;
  }
  inline LogF& add(const LogF& logF) {
    w += logF.w;
    return *this;
  }
  inline LogF& subtract(const LogF& logF) {
    w -= logF.w;
    return *this;
  }
  inline LogF& multiplyExpVal(const T& z) {
    w *= z;
    return *this;
  }
  inline LogF& multiplyLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w *= exp(z);
    } else {
      w *= -exp(z);
    }
    return *this;
  }
  inline LogF& multiply(const LogF& logF) {
    w *= logF.w;
    return *this;
  }
  inline LogF& operator=(const T& z1) {
    return setExpVal(z1);
  }
  inline LogF& operator+=(const LogF& z1) {
    return add(z1);
  }
  inline LogF& operator-=(const LogF& z1) {
    return subtract(z1);
  }
  inline LogF& operator*=(const LogF& z1) {
    return multiply(z1);
  }
  inline LogF& operator/=(const LogF& z1) {
    w /= z1.w;
    return *this;
  }
  inline LogF operator+(const LogF& z1) const {
    LogF z2(*this);
    return (z2 += z1);
  }
  inline LogF operator-(const LogF& z1) const {
    LogF z2(*this);
    return (z2 -= z1);
  }
  inline LogF operator*(const LogF& z1) const {
    LogF z2(*this);
    return (z2 *= z1);
  }
  inline LogF operator/(const LogF& z1) const {
    LogF z2(*this);
    return (z2 /= z1);
  }
  //inline operator T() const {
  //  return getExpVal();
  //}
};
#else

template <typename T>
struct 
DPScore::
LogF 
{
  T z;
  bool sgn;

  inline static T getNegInf() {
    return (-FLT_MAX * 0.01);
  }
  inline static const LogF& null() {
    static const LogF nullVal = {getNegInf(), true};
    return nullVal;
  }
  inline LogF& setNull() {
    z = getNegInf(); 
    sgn = true;
    return (*this);
  }
  inline LogF& setExpVal(const T& z1) {
    if (z1 == 0.0) {
      setNull();
    } else if (z1 > 0.0) {
      z = log(z1);
      sgn = true;
    } else {
      z = log(-z1);
      sgn = false;
    }
    return *this;
  }
  inline LogF& setLogVal(const T& z1, const bool& sgn1 = true) {
    z = z1;
    sgn = sgn1;
    return *this;
  }
  inline LogF& setVal(const LogF& logF) {
    z = logF.z;
    sgn = logF.sgn;
    return *this;
  }
  inline T getExpVal() const {
    return (sgn ? exp(z) : (-exp(z)));
  }
  inline void getLogVal(T& z1, bool& sgn1) const {
    z1 = z;
    sgn1 = sgn;
  }
  inline LogF& addExpVal(const T& z1) {
    LogF z2;
    z2.setExpVal(z1);
    return add(z2);
  }
  inline LogF& addLogVal(const T& z1, const bool& sgn1 = true) {
    LogF z2;
    z2.setLogVal(z1, sgn1);
    return add(z2);
  }
  inline LogF& add(const LogF& logF) {
    static const T min_diff = 1.0e-15;
    if (sgn == logF.sgn) {
      logAddBasic(z, logF.z);
    } else {
      const T& diff = (z - logF.z);
      if (diff > min_diff) {
	z += log(1.0 - exp(-diff));
      } else if (diff < -min_diff) {
	z = (logF.z + log(1.0 - exp(diff)));
	sgn = logF.sgn;
      } else {
	setNull();
      }
    }
    return *this;
  }
  inline LogF& subtract(const LogF& logF) {
    static const T min_diff = 1.0e-15;
    if (sgn == (!logF.sgn)) {
      logAddBasic(z, logF.z);
    } else {
      const T& diff = (z - logF.z);
      if (diff > min_diff) {
	z += log(1.0 - exp(-diff));
      } else if (diff < -min_diff) {
	z = (logF.z + log(1.0 - exp(diff)));
	sgn = (!logF.sgn);
      } else {
	setNull();
      }
    }
    return *this;
  }
  inline LogF& multiplyExpVal(const T& z1) {
    LogF z2;
    z2.setExpVal(z1);
    multiply(z2);
    return *this;
  }
  inline LogF& multiplyLogVal(const T& z1, const bool& sgn1 = true) {
    LogF z2;
    z2.setLogVal(z1, sgn1);
    multiply(z2);
    return *this;
  }
  inline LogF& multiply(const LogF& logF) {
    z += logF.z;
    sgn = (sgn == logF.sgn);
    return *this;
  }
  inline LogF& operator=(const T& z1) {
    return setExpVal(z1);
  }
  inline LogF& operator+=(const LogF& z1) {
    return add(z1);
  }
  inline LogF& operator-=(const LogF& z1) {
    return addLogVal(z1.z, !(z1.sgn));
  }
  inline LogF& operator*=(const LogF& z1) {
    return multiply(z1);
  }
  inline LogF& operator/=(const LogF& z1) {
    return multiplyLogVal(-z1.z, z1.sgn);
  }
  inline LogF operator+(const LogF& z1) const {
    LogF z2(*this);
    return (z2 += z1);
  }
  inline LogF operator-(const LogF& z1) const {
    LogF z2(*this);
    return (z2 -= z1);
  }
  inline LogF operator*(const LogF& z1) const {
    LogF z2(*this);
    return (z2 *= z1);
  }
  inline LogF operator/(const LogF& z1) const {
    LogF z2(*this);
    return (z2 /= z1);
  }
  //inline operator T() const {
  //  return getExpVal();
  //}
};
#endif

class
DPScore::
FVecSimple 
{
public:
  typedef Array<float> Data;
  Data data;

  FVecSimple(int size, const float val)
    : data(size, val) {}
  FVecSimple(const FVecSimple& fv)
    : data(fv.data) {}
  ~FVecSimple() {}
  void multiplyLogVal(int i, float z, bool sgn = true) {
    data[i] *= (sgn ? exp(z) : -exp(z));
  }
  void addLogVal(int i, float z, bool sgn = true) {
    data[i] += (sgn ? exp(z) : -exp(z));
  }
  void fillLogVal(float z, bool sgn = true) {
    data.fill(sgn ? exp(z) : -exp(z));
  }
  void fillLogValRange(int i_beg, int i_end,
		       float z, bool sgn = true) {
    fill(data.begin() + i_beg, data.begin() + i_end, 
	 (sgn ? exp(z) : -exp(z)));
  }
  void multiplyExpVal(int i, float z) {
    data[i] *= z;
  }
  void addExpVal(int i, float z) {
    data[i] += z;
  }
  void fillExpVal(float z) {
    data.fill(z);
  }
  void fillLogValRange(int i_beg, int i_end, float z) {
    fill(data.begin() + i_beg, data.begin() + i_end, z);
  }
  Array<float> getExpValArray() const {
    return data;
  }
};


template <typename T>
struct 
DPScore::
LogF_Simple
{
  T w;
  inline static const LogF_Simple& null() {
    static const LogF_Simple nullVal = {getNegInf()};
    return nullVal;
  }
  inline static T getNegInf() {
    return (0.0);
  }
  inline static T getMinExpVal() {
    return 0.0;
  }
  inline LogF_Simple& setExpVal(const T& z) {
    w = z;
    return *this;
  }
  inline LogF_Simple& setLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w = exp(z);
    } else {
      w = -exp(z);
    }
    return *this;
  }
  inline LogF_Simple& setVal(const LogF_Simple& logF) {
    w = logF.w;
    return *this;
  }
  inline T getExpVal() const {
    return w;
  }
  inline void getLogVal(T& z, bool& sgn) const {
    if (w >= 0) {
      z = log(w);
      sgn = true;
    } else {
      z = log(-w);
      sgn = false;
    }
  }
  inline LogF_Simple& addExpVal(const T& z) {
    w += z;
    return *this;
  }
  inline LogF_Simple& addLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w += exp(z);
    } else {
      w += -exp(z);
    }
    return *this;
  }
  inline LogF_Simple& add(const LogF_Simple& logF) {
    w += logF.w;
    return *this;
  }
  inline LogF_Simple& subtract(const LogF_Simple& logF) {
    w -= logF.w;
    return *this;
  }
  inline LogF_Simple& multiplyExpVal(const T& z) {
    w *= z;
    return *this;
  }
  inline LogF_Simple& multiplyLogVal(const T& z, const bool& sgn = true) {
    if (sgn) {
      w *= exp(z);
    } else {
      w *= -exp(z);
    }
    return *this;
  }
  inline LogF_Simple& multiply(const LogF_Simple& logF) {
    w *= logF.w;
    return *this;
  }
  inline LogF_Simple& operator=(const T& z1) {
    return setExpVal(z1);
  }
  inline LogF_Simple& operator+=(const LogF_Simple& z1) {
    return add(z1);
  }
  inline LogF_Simple& operator-=(const LogF_Simple& z1) {
    return subtract(z1);
  }
  inline LogF_Simple& operator*=(const LogF_Simple& z1) {
    return multiply(z1);
  }
  inline LogF_Simple& operator/=(const LogF_Simple& z1) {
    w /= z1.w;
    return *this;
  }
  inline LogF_Simple operator+(const LogF_Simple& z1) const {
    LogF_Simple z2(*this);
    return (z2 += z1);
  }
  inline LogF_Simple operator-(const LogF_Simple& z1) const {
    LogF_Simple z2(*this);
    return (z2 -= z1);
  }
  inline LogF_Simple operator*(const LogF_Simple& z1) const {
    LogF_Simple z2(*this);
    return (z2 *= z1);
  }
  inline LogF_Simple operator/(const LogF_Simple& z1) const {
    LogF_Simple z2(*this);
    return (z2 /= z1);
  }
  //inline operator T() const {
  //  return getExpVal();
  //}
};

class
DPScore::
FVecLogF
{
public:
  typedef DPScore::LogF<double> Datum;
  typedef Array<Datum> Data;
  Data data;

  FVecLogF(int size, float val)
    : data(size) {
    fillExpVal(val);
  }
  FVecLogF(const FVecLogF& fv)
    : data(fv.data) {}
  ~FVecLogF() {}
  void multiplyLogVal(int i, float z, bool sgn = true) {
    data[i].multiplyLogVal(z, sgn);
  }
  void addLogVal(int i, float z, bool sgn = true) {
    data[i].addLogVal(z, sgn);
  }
  void fillLogVal(float z, bool sgn = true) {
    Datum datum;
    datum.setLogVal(z, sgn);
    data.fill(datum);
  }
  void fillLogValRange(int i_beg, int i_end,
		       float z, bool sgn = true) {
    Datum datum;
    datum.setLogVal(z, sgn);
    fill(data.begin() + i_beg, data.begin() + i_end, datum);
  }
  void multiplyExpVal(int i, float z) {
    data[i].multiplyExpVal(z);
  }
  void addExpVal(int i, float z) {
    data[i].addExpVal(z);
  }
  void fillExpVal(float z) {
    Datum datum;
    datum.setExpVal(z);
    data.fill(datum);
  }
  void fillLogValRange(int i_beg, int i_end, float z) {
    Datum datum;
    datum.setExpVal(z);
    fill(data.begin() + i_beg, data.begin() + i_end, datum);
  }
  Array<float> getExpValArray() const {
    Array<float> v(data.size());
    Array<float>::iterator k = v.begin();
    for (Data::const_iterator i = data.begin(); i != data.end(); i++) {
      (*k++) = i->getExpVal();
    }
    return v;
  }
};

#endif
