/*
 * IntPairs.hpp
 *
 */
#ifndef INT_PAIRS_HPP
#define INT_PAIRS_HPP

class IntPairs0 {
public:
  IntPairs0(const Vector<int>& vals)
    : idx2val(vals.begin(), vals.end()),
      val_beg(getValBeg(vals.begin(), vals.end())),
      val_end(getValEnd(vals.begin(), vals.end())),
      val2idx(getVal2IdxSize()) {
    setVal2Idx();
  }
  ~IntPairs0() {}
  bool empty() const {return idx2val.empty();}
  int size() const {return idx2val.size();}
  bool hasVal(int val) const {
    return (!empty()
	    && (val_beg <= val)
	    && (val < val_end)
	    && (val2idx[val - val_beg] != nullVal()));
  }
  const int& indexOf(int val) const {
    Assert(hasVal(val));
    const int& v = (val - val_beg);
    const int& idx = val2idx[v];
    Assert(idx != nullVal());
    return idx;
  }

  //type definitions
  typedef int*             iterator;
  typedef const int*       const_iterator;
  typedef int&             reference;
  typedef const int&       const_reference;
  //iterator support
  //iterator begin() {return idx2val.begin();}
  const_iterator begin() const {return idx2val.begin();}
  //iterator end() {return idx2val.end();}
  const_iterator end() const {return idx2val.end();}
  //reference front() {return idx2val.front();}
  const_reference front() const {return idx2val.front();}
  //reference back() {return idx2val.back();}
  const_reference back() const {return idx2val.back();}
  
  int getValBeg() const {
    Assert(val_beg != nullVal());
    return val_beg;
  }
  int getValLast() const {
    Assert(val_end != nullVal());
    return (val_end - 1);
  }
  const int& operator[](int i) const {return idx2val[i];}

private:
  Array<int> idx2val;
  const int val_beg;
  const int val_end;
  Array<int> val2idx;

  static int nullIdx() {return INT_MIN;}
  static int nullVal() {return INT_MIN;}
  template <typename InpIt>
  static int getValBeg(InpIt first, InpIt last) {
    InpIt i = min_element(first, last);
    return (i == last ? nullVal() : (*i));
  }
  template <typename InpIt>
  static int getValEnd(InpIt first, InpIt last) {
    InpIt i = max_element(first, last);
    return (i == last ? nullVal() : ((*i) + 1));
  }
  int getVal2IdxSize() const {
    return (val_end - val_beg);
  }
  void setVal2Idx() {
    val2idx.fill(nullVal());
    for (int i = 0; i < (int)idx2val.size(); i++) {
      const int& v = (idx2val[i] - val_beg);
      val2idx[v] = i;
    }
  }
};

class IntPairs {
public:
  IntPairs() 
    : idx2val(),
      val_beg(nullVal()),
      val_end(nullVal()),
      val2idx() {
  }
  IntPairs(const Vector<int>& vals)
    : idx2val(vals) {
    checkUniqueness();
    resetVal2Idx();
  }
  IntPairs(const IntPairs& pairs)
    : idx2val(pairs.idx2val),
      val_beg(pairs.val_beg),
      val_end(pairs.val_end),
      val2idx(pairs.val2idx) {
  }
  template <typename InpIt>
  IntPairs(InpIt first, InpIt last)
    : idx2val(first, last),
      val_beg(nullVal()),
      val_end(nullVal()),
      val2idx() {
    checkUniqueness();
    resetVal2Idx();
  }
  ~IntPairs() {}
  IntPairs& operator=(const IntPairs& pairs) {
    idx2val = pairs.idx2val;
    val_beg = pairs.val_beg;
    val_end = pairs.val_end;
    val2idx = pairs.val2idx;
    return *this;
  }
  bool operator==(const IntPairs& pairs) const {
    if (idx2val.size() != pairs.idx2val.size()) return false;

    return equal(idx2val.begin(),idx2val.end(),pairs.idx2val.begin());
  }
  bool operator<(const IntPairs& pairs) const {
    return lexicographical_compare(idx2val.begin(),idx2val.end(),
				   pairs.idx2val.begin(),pairs.idx2val.end());
  }

  //type definitions
  typedef Vector<int>::iterator iterator;
  typedef Vector<int>::const_iterator const_iterator;
  typedef Vector<int>::reference reference ;
  typedef Vector<int>::const_reference const_reference ;
  //iterator support
  //iterator begin() {return idx2val.begin();}
  const_iterator begin() const {return idx2val.begin();}
  //iterator end() {return idx2val.end();}
  const_iterator end() const {return idx2val.end();}
  //reference front() {return idx2val.front();}
  const_reference front() const {return idx2val.front();}
  //reference back() {return idx2val.back();}
  const_reference back() const {return idx2val.back();}

  bool empty() const {return idx2val.empty();}
  int size() const {return idx2val.size();}
  void clear() {
    idx2val.clear();
    val_beg = nullVal();
    val_end = nullVal();
    val2idx.clear();
  }
  void push_back(int val) {
    idx2val.push_back(val);
    checkUniqueness();
    resetVal2Idx();
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {
    idx2val.assign(first, last);
    checkUniqueness();
    resetVal2Idx();
  }
  void assign(const Vector<int>& vals) {
    idx2val = vals;
    checkUniqueness();
    resetVal2Idx();
  }
  template <typename InpIt>
  void append(InpIt first, InpIt last) {
    idx2val.insert(idx2val.end(), first, last);
    checkUniqueness();
    resetVal2Idx();
  }
  void append(const IntPairs& pairs) {
    append(pairs.idx2val);
  }
  void append(const Vector<int>& vals) {
    idx2val.insert(idx2val.end(), vals.begin(), vals.end());
    checkUniqueness();
    resetVal2Idx();
  }
  bool hasVal(int val) const {
    return (!empty()
	    && (val_beg <= val)
	    && (val < val_end)
	    && (val2idx[val - val_beg] != nullVal()));
  }
  int indexOf(int val) const {
    Assert(hasVal(val));
    const int& v = (val - val_beg);
    const int& idx = val2idx[v];
    Assert(idx != nullVal());
    return idx;
  }
  const int& operator[](int i) const {return idx2val[i];}
  
  int getValBeg() const {
    Assert(val_beg != nullVal());
    return val_beg;
  }
  int getValLast() const {
    Assert(val_end != nullVal());
    return (val_end - 1);
  }

private:
  Vector<int> idx2val;

  int val_beg;
  int val_end;
  Array<int> val2idx;

  static int nullIdx() {return INT_MIN;}
  static int nullVal() {return INT_MIN;}
  template <typename InpIt>
  static int getValBeg(InpIt first, InpIt last) {
    InpIt i = min_element(first, last);
    return (i == last ? nullVal() : (*i));
  }
  template <typename InpIt>
  static int getValEnd(InpIt first, InpIt last) {
    InpIt i = max_element(first, last);
    return (i == last ? nullVal() : ((*i) + 1));
  }
  int getVal2IdxSize() const {
    return (val_end - val_beg);
  }
  void resetVal2Idx() {
    val_beg = getValBeg(idx2val.begin(), idx2val.end());
    val_end = getValEnd(idx2val.begin(), idx2val.end());
    if (val_beg == nullVal() || val_end == nullVal()) {
      clear();
      return;
    } else {
      val2idx.resize(val_end - val_beg);
      val2idx.fill(nullVal());
      for (int i = 0; i < (int)idx2val.size(); i++) {
	const int& val = idx2val[i];
	Assert(val != nullVal());
	const int& v = (val - val_beg);
	val2idx[v] = i;
      }
    }
    testConsistency();
  }

  void checkUniqueness() const {
    Vector<int> v(idx2val);
    std::sort(v.begin(), v.end());
    Vector<int>::iterator i = std::unique(v.begin(), v.end());
    Check(i == v.end(), "v.end() - i = %d", v.end() - i);
  }

  void testConsistency() const {
#ifdef DEBUG
    if (idx2val.empty()) {
      Assert(val_beg == nullVal());
      Assert(val_end == nullVal());
      Assert(val2idx.empty());
    } else {
      Assert(val_beg != nullVal());
      Assert(val_end != nullVal());
      Assert(val_beg < val_end);
      Assert(idx2val.size() <= val2idx.size());

      int val_min = nullVal();
      int val_max = nullVal();
      for (int i = 0; i < (int)idx2val.size(); i++) {
	const int& val = idx2val[i];
	Assert(val != nullVal());
	Assert(i == val2idx[val - val_beg]);
	if (val_min == nullVal()) {
	  val_min = val;
	  val_max = val;
	} else {
	  val_min = min(val_min, val);
	  val_max = max(val_max, val);
	}
      }
      Assert(val_min != nullVal());
      Assert(val_max != nullVal());
      Assert(val_min == val_beg);
      Assert((val_max + 1) == val_end);
      for (int i = 0; i < (int)val2idx.size(); i++) {
	const int& j = val2idx[i];
	if (j == nullVal()) continue;

	Assert(0 <= j && j < (int)idx2val.size());
	const int& val = (val_beg + i);
	Assert(idx2val[j] == val);
      }
    }
#endif
  }
};

template <typename T>
class ITable {
public:
  typedef Vector<T> Data;
  typedef T value_type;
  typedef typename Data::iterator iterator;
  typedef typename Data::const_iterator const_iterator;
  typedef typename Data::reference reference;
  typedef typename Data::const_reference const_reference;

  ITable() : pairs(), data() {}
  ITable(const Vector<int>& keys, const T& val = T())
    : pairs(keys), data(keys.size(), val) {}
  ITable(const ITable& iTable) : pairs(iTable.pairs), data(iTable.data){}
  ~ITable() {};
  ITable& operator=(const ITable& iTable) {
    pairs = iTable.pairs;
    data = data;
    return *this;
  }
  
  bool empty() const {return pairs.empty();}
  int size() const {return pairs.size();}
  void clear() {
    pairs.clear();
    data.clear();
  }
  void push_back(const int& key, const T& val) {
    pairs.push_back(key);
    data.push_back(val);
  }
  void assign(const Vector<int>& keys, const Vector<T>& vals) {
    Check(keys.size() == vals.size());
    pairs.assign(keys.begin(), keys.end());
    data.assign(vals.begin(), vals.end());
  }
  void append(const Vector<int>& keys, const Vector<T>& vals) {
    Check(keys.size() == vals.size());
    pairs.append(keys.begin(), keys.end());
    data.insert(data.end(), vals.begin(), vals.end());
  }

  bool hasKey(int key) const {return pairs.hasVal(key);}
  reference ref(int key) {return data[pairs.indexOf(key)];}
  const_reference ref(int key) const {return data[pairs.indexOf(key)];}
  reference operator[](int i) {return data[i];}
  const_reference operator[](int i) const {return data[i];}
  iterator begin() {return data.begin();}
  const_iterator begin() const {return data.begin();}
  iterator end() {return data.end();}
  const_iterator end() const {return data.end();}
  reference front() {return data.front();}
  const_reference front() const {return data.front();}
  reference back() {return data.back();}
  const_reference back() const {return data.back();}
  
private:
  IntPairs pairs;
  Data data;
};
#endif
