/*
 * Job.cpp
 *
 */
#include "Util.hpp"
#include "Rect.hpp"
#include "EScript.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
#include "Weight.hpp"
#include "DPPath.hpp"
#include "GuideTree.hpp"
#include "Job.hpp"
#include "SparseProb.hpp"

Job::Info&
Job::Info::
operator=(const Info& info)
{
  command = info.command;
  sendType = info.sendType;
  copy(info.nChars, info.nChars + ARRSIZE, nChars);
  copy(info.nInts, info.nInts + ARRSIZE, nInts);
  copy(info.nFloats, info.nFloats + ARRSIZE, nFloats);
  return *this;
}
void
Job::Info::
clean()
{
  command = CMD_EMPTY;
  sendType = SEND_NON;
  fill(nChars, nChars + ARRSIZE, 0);
  fill(nInts, nInts + ARRSIZE, 0);
  fill(nFloats, nFloats + ARRSIZE, 0);
}
int
Job::Info::
getCharDataSize() const
{
  return accumulate(nChars, nChars + ARRSIZE, 0);
}
int
Job::Info::
getIntDataSize() const
{
  return accumulate(nInts, nInts + ARRSIZE, 0);
}
int
Job::Info::
getFloatDataSize() const
{
  return accumulate(nFloats, nFloats + ARRSIZE, 0);
}
string
Job::Info::
toString() const
{
  ostringstream oss;
  oss << "--- \n";
  oss << "command: " << command2str(command) << "\n";
  oss << "sendType: " << sendType2str(sendType) << "\n";
  oss << "nChars: [ ";
  for (int i = 0; i < ARRSIZE; i++) {
    if (nChars[i] == 0) break;

    oss << (i == 0 ? "" : ", ") << nChars[i];
  }
  oss << " ]\n";
  oss << "nInts: [ ";
  for (int i = 0; i < ARRSIZE; i++) {
    if (nInts[i] == 0) break;

    oss << (i == 0 ? "" : ", ") << nInts[i];
  }
  oss << " ]\n";
    oss << "nFloats: [ ";
  for (int i = 0; i < ARRSIZE; i++) {
    if (nFloats[i] == 0) break;

    oss << (i == 0 ? "" : ", ") << nFloats[i];
  }
  oss << " ]\n";
  return oss.str();
}
const char*
Job::
command2str(const int& command)
{
  static const struct {int code; string str;} pairs[] = {
    {CMD_EMPTY,         "CMD_EMPTY"},
    {CMD_STRING,        "CMD_STRING"},
    {CMD_WEIGHT,        "CMD_WEIGHT"},
    {CMD_GUIDE_TREE,    "CMD_GUIDE_TREE"},
    {CMD_ALIGN,         "CMD_ALIGN"},
    {CMD_REFINE,        "CMD_REFINE"},
    {CMD_MERGE,         "CMD_MERGE"},
    {CMD_CALC_LL,       "CMD_CALC_LL"},
    {CMD_CALC_SPS,      "CMD_CALC_SPS"},
    {CMD_EVAL_SCORES,   "CMD_EVAL_SCORES"},
    {CMD_CALC_FDF,      "CMD_CALC_FDF"},
    {CMD_FOLD,          "CMD_FOLD"},
    {CMD_ALIGN_RESULT,  "CMD_ALIGN_RESULT"},
    {CMD_TEST_PARAM,    "CMD_TEST_PARAM"},
    {CMD_CALC_PAIR_PROB,"CMD_CALC_PAIR_PROB"},
    {CMD_BCAST_PAIR_PROB,"CMD_BCAST_PAIR_PROB"},
    {CMD_CALC_ALIGN_PROB,"CMD_CALC_ALIGN_PROB"},
    {CMD_BCAST_ALIGN_PROB,"CMD_BCAST_ALIGN_PROB"},
    {CMD_SMOOTH_ALIGN_PROB, "CMD_SMOOTH_ALIGN_PROB"},
    {CMD_CALC_ML_DISTANCE, "CMD_CALC_ML_DISTANCE"}
  };
  static const string bad_command = "??";
  static const int n = sizeof(pairs) / sizeof(pairs[0]);
  for (int i = 0; i < n; i++) {
    if (command == pairs[i].code) {
      return pairs[i].str.c_str();
    }
  }
  return bad_command.c_str();
}
const char*
Job::
sendType2str(const int& sendType)
{
  static const struct {SendType sendType; string str;} pairs[] = {
    {SEND_ALL, "SEND_ALL"},
    {SEND_ANY, "SEND_ANY"},
    {SEND_NON, "SEND_NON"}
  };
  static const string bad_sendType = "??";
  static const int n = sizeof(pairs) / sizeof(pairs[0]);
  for (int i = 0; i < n; i++) {
    if (sendType == pairs[i].sendType) {
      return pairs[i].str.c_str();
    }
  }
  return bad_sendType.c_str();
}
Job::
Job()
  : info(), charData(), intData(), floatData()
{
  info.clean();
}
Job::
Job(const Job& job)
{
  *this = job;
}
Job::
~Job()
{
}
Job& 
Job::
operator=(const Job& job)
{
  info = job.info;
  charData = job.charData;
  intData = job.intData;
  floatData = job.floatData;
  return *this;
}
bool
Job::
empty() const
{
 return (info.command == CMD_EMPTY);
}
void
Job::
clear()
{
  info.clean();
  charData.clear();
  intData.clear();
  floatData.clear();
}
string
Job::
toString() const
{
  ostringstream oss;
  oss << "--- \n";
  oss << "info: \n";
  oss << info.toString();
  oss << "charData: \n";
  for (int i = 0; i < (int)charData.size(); i++) {
    oss << "  - " << charData[i] << "\n";
  }
  oss << "intData: \n";
  for (int i = 0; i < (int)intData.size(); i++) {
    oss << "  - " << intData[i] << "\n";
  }
  oss << "floatData: \n";
  for (int i = 0; i < (int)floatData.size(); i++) {
    oss << "  - " << floatData[i] << "\n";
  }
  return oss.str();
}
void
Job::
print() const
{
  cout << toString() << flush;
}
void
Job::
set_Terminate()
{
  clear();
  info.command = CMD_EMPTY;
  info.sendType = SEND_ALL;
}
void
Job::
set_String(const string& str)
{
  clear();
  info.command = CMD_STRING;
  info.sendType = SEND_ALL;
  info.nChars[0] = str.size();
  charData.assign(str.begin(), str.end());
}
void
Job::
get_String(string& str) const
{
  Assert(info.command == CMD_STRING);
  Assert(info.sendType == SEND_ALL);
  Assert(info.nChars[0] == (int)charData.size());
  str.assign(charData.begin(), charData.end());
}
void 
Job::
set_Weight(const SeqIds& seqIds, const Weight& weights)
{
  clear();
  info.command    = CMD_WEIGHT;
  info.sendType   = SEND_ALL;
  info.nInts[0]   = seqIds.size();
  info.nFloats[0] = seqIds.size();
  for (int i = 0; i < seqIds.size(); i++) {
    int seqId = seqIds[i];
    float wgt = weights.ref(seqId);
    intData.push_back(seqId);
    floatData.push_back(wgt);
  }
}
void
Job::
get_Weight(Weight& weights) const {
  Assert(info.command == CMD_WEIGHT);
  Assert(info.sendType == SEND_ALL);
  Assert(info.nInts[0] == info.nFloats[0]);
  Assert(info.nInts[0] == (int)intData.size());
  Assert(info.nFloats[0] == (int)floatData.size());
  SeqIds seqIds(intData.begin(), intData.end());
  weights.assign(seqIds, floatData);
}
void
Job::
set_GuideTree(const GuideTree& guideTree)
{
  clear();
  info.command = CMD_GUIDE_TREE;
  info.sendType = SEND_ALL;
  charData = guideTree.getPackedData();
  info.nChars[0] = charData.size();
}
void
Job::
get_GuideTree(GuideTree& guideTree) const
{
  Assert(info.command == CMD_GUIDE_TREE);
  Assert(info.sendType == SEND_ALL);
  Assert(info.nChars[0] == (int)charData.size());
  guideTree.setFromPackedData(charData);
}
void
Job::
pack(const Vector<int>& intData1, const Vector<float>& floatData1)
{
  info.nInts[0] = intData1.size();
  info.nFloats[0] = floatData1.size();
  intData = intData1;
  floatData = floatData1;
}
void
Job::
unpack(Vector<int>& intData1, Vector<float>& floatData1) const
{
  Assert(info.nInts[0] == (int)intData.size());
  Assert(info.nFloats[0] == (int)floatData.size());
  intData1 = intData;
  floatData1 = floatData;
}
void
Job::
pack(const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
     const float& score, const Vector<float>& params)
{
  info.nInts[0]   = msaX.size();
  info.nInts[1]   = msaY.size();
  info.nInts[2]   = dpPath.getPackedDataSize();
  info.nFloats[0] = 1;
  info.nFloats[1] = params.size();

  copy(msaX.begin(), msaX.end(), back_inserter(intData));
  copy(msaY.begin(), msaY.end(), back_inserter(intData));
  const Vector<int>& v = dpPath.getPackedData();
  copy(v.begin(), v.end(), back_inserter(intData));

  floatData.push_back(score);
  copy(params.begin(), params.end(), back_inserter(floatData));
}
void
Job::
unpack(MSA& msaX, MSA& msaY, DPPath& dpPath,
       float& score, Vector<float>& params) const
{
  Vector<int>::const_iterator bi = intData.begin();
  Vector<int>::const_iterator ei = bi + info.nInts[0];
  msaX.assign(bi, ei);

  bi = ei;
  ei = bi + info.nInts[1];
  msaY.assign(bi, ei);

  bi = ei;
  ei = bi + info.nInts[2];
  Vector<int> v(bi, ei);
  dpPath.setFromPackedData(v);
 
  Vector<float>::const_iterator bfi = floatData.begin();
  Vector<float>::const_iterator efi = bfi + info.nFloats[0];
  score = *bfi;

  bfi = efi;
  efi = bfi + info.nFloats[1];
  params.assign(bfi, efi);
}
void
Job::
pack(const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
     const float& score, const Vector<float>& params, const string& message)
{
  info.nInts[0]   = msaX.size();
  info.nInts[1]   = msaY.size();
  info.nInts[2]   = dpPath.getPackedDataSize();
  info.nFloats[0] = 1;
  info.nFloats[1] = params.size();
  info.nChars[0] = message.size();

  copy(msaX.begin(), msaX.end(), back_inserter(intData));
  copy(msaY.begin(), msaY.end(), back_inserter(intData));
  const Vector<int>& v = dpPath.getPackedData();
  copy(v.begin(), v.end(), back_inserter(intData));

  floatData.push_back(score);
  copy(params.begin(), params.end(), back_inserter(floatData));

  copy(message.begin(), message.end(), back_inserter(charData));
}
void
Job::
unpack(MSA& msaX, MSA& msaY, DPPath& dpPath,
       float& score, Vector<float>& params, string& message) const
{
  Vector<int>::const_iterator bi = intData.begin();
  Vector<int>::const_iterator ei = bi + info.nInts[0];
  msaX.assign(bi, ei);

  bi = ei;
  ei = bi + info.nInts[1];
  msaY.assign(bi, ei);

  bi = ei;
  ei = bi + info.nInts[2];
  Vector<int> v(bi, ei);
  dpPath.setFromPackedData(v);
 
  Vector<float>::const_iterator bfi = floatData.begin();
  Vector<float>::const_iterator efi = bfi + info.nFloats[0];
  score = *bfi;

  bfi = efi;
  efi = bfi + info.nFloats[1];
  params.assign(bfi, efi);

  message.assign(charData.begin(), charData.begin() + info.nChars[0]);
}
void
Job::
pack(const GuideTree& guideTree, const Weight& weights,
     const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
     const float& score, const Vector<float>& params)
{
  pack(msaX, msaY, dpPath, score, params);
  const Vector<char>& packedData0 = guideTree.getPackedData();
  const Vector<char>& packedData1 = weights.getPackedData();
  Assert(info.nChars[0] == 0 && info.nChars[1] == 0 && charData.empty());
  info.nChars[0] = packedData0.size();
  info.nChars[1] = packedData1.size();
  charData.insert(charData.end(), packedData0.begin(),packedData0.end());
  charData.insert(charData.end(), packedData1.begin(),packedData1.end());
}
void
Job::
unpack(GuideTree& guideTree, Weight& weights,
       MSA& msaX, MSA& msaY, DPPath& dpPath,
       float& score, Vector<float>& params) const
{
  Assert((int)charData.size() == (info.nChars[0] + info.nChars[1]));
  Vector<char>::const_iterator bi = charData.begin();
  Vector<char>::const_iterator ei = (bi + info.nChars[0]);
  guideTree.setFromPackedData(Vector<char>(bi, ei));
  bi = ei;
  ei = (bi + info.nChars[1]);
  weights.setFromPackedData(Vector<char>(bi, ei));
  unpack(msaX, msaY, dpPath, score, params);
}
void
Job::
set_Align(const GuideTree& guideTree, const Weight& weights,
	  const MSA& msaX, const MSA& msaY)
{
  clear();
  info.command    = CMD_ALIGN;
  info.sendType   = SEND_ANY;
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Align(GuideTree& guideTree, Weight& weights,
	  MSA& msaX, MSA& msaY) const
{
  Assert(info.command == CMD_ALIGN);
  Assert(info.sendType == SEND_ANY);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Align1(const GuideTree& guideTree, const Weight& weights,
	  const MSA& msaX, const MSA& msaY)
{
  clear();
  info.command    = CMD_ALIGN1;
  info.sendType   = SEND_ANY;
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Align1(GuideTree& guideTree, Weight& weights,
	  MSA& msaX, MSA& msaY) const
{
  Assert(info.command == CMD_ALIGN1);
  Assert(info.sendType == SEND_ANY);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Align2(const GuideTree& guideTree, const Weight& weights,
	  const MSA& msaX, const MSA& msaY)
{
  clear();
  info.command    = CMD_ALIGN2;
  info.sendType   = SEND_ANY;
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Align2(GuideTree& guideTree, Weight& weights,
	  MSA& msaX, MSA& msaY) const
{
  Assert(info.command == CMD_ALIGN2);
  Assert(info.sendType == SEND_ANY);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Align3(const GuideTree& guideTree, const Weight& weights,
	  const MSA& msaX, const MSA& msaY)
{
  clear();
  info.command    = CMD_ALIGN3;
  info.sendType   = SEND_ANY;
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Align3(GuideTree& guideTree, Weight& weights,
	  MSA& msaX, MSA& msaY) const
{
  Assert(info.command == CMD_ALIGN3);
  Assert(info.sendType == SEND_ANY);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Refine(const GuideTree& guideTree, const Weight& weights,
	   const MSA& msaX, const MSA& msaY, const DPPath& dpPath)
{
  clear();
  info.command    = CMD_REFINE;
  info.sendType   = SEND_ANY;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Refine(GuideTree& guideTree, Weight& weights,
	   MSA& msaX, MSA& msaY, DPPath& dpPath) const
{
  Assert(info.command == CMD_REFINE);
  Assert(info.sendType == SEND_ANY);
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Refine2(const GuideTree& guideTree, const Weight& weights,
	   const MSA& msaX, const MSA& msaY, const DPPath& dpPath)
{
  clear();
  info.command    = CMD_REFINE2;
  info.sendType   = SEND_ANY;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Refine2(GuideTree& guideTree, Weight& weights,
	   MSA& msaX, MSA& msaY, DPPath& dpPath) const
{
  Assert(info.command == CMD_REFINE2);
  Assert(info.sendType == SEND_ANY);
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_CalcLL(const GuideTree& guideTree, const Weight& weights,
	   const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	   const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_CALC_LL;
  info.sendType   = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_CalcLL(GuideTree& guideTree, Weight& weights,
	   MSA& msaX, MSA& msaY, DPPath& dpPath,
	   float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_CALC_LL);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_CalcSPS(const GuideTree& guideTree, const Weight& weights,
	    const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_CALC_SPS;
  info.sendType   = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_CalcSPS(GuideTree& guideTree, Weight& weights,
	    MSA& msaX, MSA& msaY, DPPath& dpPath,
	    float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_CALC_SPS);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_EvalScores(const GuideTree& guideTree, const Weight& weights,
	       const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	       const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_EVAL_SCORES;
  info.sendType   = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_EvalScores(GuideTree& guideTree, Weight& weights,
	       MSA& msaX, MSA& msaY, DPPath& dpPath,
	       float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_EVAL_SCORES);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_CalcFDF(const GuideTree& guideTree, const Weight& weights,
	    const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_CALC_FDF;
  info.sendType   = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_CalcFDF(GuideTree& guideTree, Weight& weights,
	    MSA& msaX, MSA& msaY, DPPath& dpPath,
	    float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_CALC_FDF);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_TestParam(const GuideTree& guideTree, const Weight& weights,
	    const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_TEST_PARAM;
  info.sendType   = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_TestParam(GuideTree& guideTree, Weight& weights,
	    MSA& msaX, MSA& msaY, DPPath& dpPath,
	    float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_TEST_PARAM);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Merge(const MSA& msaX, const MSA& msaY, const DPPath& dpPath)
{
  clear();
  info.command    = CMD_MERGE;
  info.sendType   = SEND_ALL;
  float score = 0.0;
  Vector<float> params;
  pack(msaX, msaY, dpPath, score, params);
}
void
Job::
get_Merge(MSA& msaX, MSA& msaY, DPPath& dpPath) const
{
  Assert(info.command == CMD_MERGE);
  Assert(info.sendType == SEND_ALL);
  float score = 0.0;
  Vector<float> params;
  unpack(msaX, msaY, dpPath, score, params);
}
void
Job::
set_Fold(const GuideTree& guideTree, const Weight& weights,
	 const MSA& msaX, const DPPath& dpPath)
{
  clear();
  info.command    = CMD_FOLD;
  info.sendType   = SEND_ANY;
  MSA msaY(msaX);
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Fold(GuideTree& guideTree, Weight& weights,
	 MSA& msaX, DPPath& dpPath) const
{
  Assert(info.command == CMD_FOLD);
  Assert(info.sendType == SEND_ANY);
  MSA msaY(msaX);
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_Fold2Step(const GuideTree& guideTree, const Weight& weights,
	 const MSA& msaX, const DPPath& dpPath)
{
  clear();
  info.command    = CMD_FOLD_2STEP;
  info.sendType   = SEND_ANY;
  MSA msaY(msaX);
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
get_Fold2Step(GuideTree& guideTree, Weight& weights,
	      MSA& msaX, DPPath& dpPath) const
{
  Assert(info.command == CMD_FOLD_2STEP);
  Assert(info.sendType == SEND_ANY);
  MSA msaY(msaX);
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void
Job::
set_AlignResult(const MSA& msaX, const MSA& msaY,
		const DPPath& dpPath,
		const float& score, const Vector<float>& params)
{
  clear();
  info.command    = CMD_ALIGN_RESULT;
  info.sendType   = SEND_ANY;
  string message = "";
  pack(msaX, msaY, dpPath, score, params, message);
}
void
Job::
get_AlignResult(MSA& msaX, MSA& msaY, DPPath& dpPath,
		float& score, Vector<float>& params) const
{
  Assert(info.command == CMD_ALIGN_RESULT);
  Assert(info.sendType == SEND_ANY);
  string message;
  unpack(msaX, msaY, dpPath, score, params, message);
}
void
Job::
set_AlignResult(const MSA& msaX, const MSA& msaY,
		const DPPath& dpPath,
		const float& score, const Vector<float>& params,
		const string& message)
{
  clear();
  info.command    = CMD_ALIGN_RESULT;
  info.sendType   = SEND_ANY;
  pack(msaX, msaY, dpPath, score, params, message);
}
void
Job::
get_AlignResult(MSA& msaX, MSA& msaY, DPPath& dpPath,
		float& score, Vector<float>& params,
		string& message) const
{
  Assert(info.command == CMD_ALIGN_RESULT);
  Assert(info.sendType == SEND_ANY);
  unpack(msaX, msaY, dpPath, score, params, message);
}
void 
Job::
set_CalcPairProb(const GuideTree& guideTree, const Weight& weights,
		 const MSA& msaX)
{
  clear();
  info.command  = CMD_CALC_PAIR_PROB;
  info.sendType = SEND_ANY;
  MSA msaY(msaX);
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void 
Job::
get_CalcPairProb(GuideTree& guideTree, Weight& weights, MSA& msaX) const
{
  Assert(info.command == CMD_CALC_PAIR_PROB);
  Assert(info.sendType == SEND_ANY);
  MSA msaY(msaX);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void 
Job::
set_BcastPairProb(const int& seqId, const SparseProb& sp)
{
  clear();
  info.command  = CMD_BCAST_PAIR_PROB;
  info.sendType = SEND_ALL;
  info.nInts[0] = 1;
  intData.resize(1);
  intData[0] = seqId;
  const string& str = sp.toString();
  info.nChars[0] = str.size();
  charData.assign(str.begin(), str.end());
}
void 
Job::
get_BcastPairProb(int& seqId, SparseProb& sp)
{
  Assert(info.command == CMD_BCAST_PAIR_PROB);
  Assert(info.sendType == SEND_ALL);
  Assert(info.nChars[0] == (int)charData.size());
  Assert(intData.size() == 1);
  seqId = intData[0];
  sp.setFromString(string(charData.begin(), charData.end()));
}
void 
Job::
set_CalcAlignProb(const GuideTree& guideTree, const Weight& weights,
		  const MSA& msaX, const MSA& msaY)
{
  clear();
  info.command  = CMD_CALC_ALIGN_PROB;
  info.sendType = SEND_ANY;
  DPPath dpPath;
  float score;
  Vector<float> params;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void 
Job::
get_CalcAlignProb(GuideTree& guideTree, Weight& weights,
			 MSA& msaX, MSA& msaY) const
{
  Assert(info.command == CMD_CALC_ALIGN_PROB);
  Assert(info.sendType == SEND_ANY);
  DPPath dpPath;
  float score;
  Vector<float> params;
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
}
void 
Job::
set_BcastAlignProb(const int& seqId, const int& seqId1,
		   const SparseProb& sp)
{
  clear();
  info.command  = CMD_BCAST_ALIGN_PROB;
  info.sendType = SEND_ALL;
  info.nInts[0] = 2;
  intData.resize(2);
  intData[0] = seqId;
  intData[1] = seqId1;
  const string& str = sp.toString();
  info.nChars[0] = str.size();
  charData.assign(str.begin(), str.end());
}
void 
Job::
get_BcastAlignProb(int& seqId, int& seqId1, SparseProb& sp) const
{
  Assert(info.command == CMD_BCAST_ALIGN_PROB);
  Assert(info.sendType == SEND_ALL);
  Assert(info.nChars[0] == (int)charData.size());
  Assert(intData.size() == 2);
  seqId = intData[0];
  seqId1 = intData[1];
  sp.setFromString(string(charData.begin(), charData.end()));
}
void 
Job::
set_SmoothAlignProb(const Weight& weights, const SeqIds& seqIds, 
		    const SeqId& seqId1, const SeqId& seqId2,
		    const Vector<float>& params)
{
  clear();
  info.command  = CMD_SMOOTH_ALIGN_PROB;
  info.sendType = SEND_ANY;
  charData = weights.getPackedData();
  info.nChars[0] = charData.size();
  info.nInts[0] = seqIds.size();
  info.nInts[1] = 1;
  info.nInts[2] = 1;
  intData.assign(seqIds.begin(), seqIds.end());
  intData.push_back(seqId1);
  intData.push_back(seqId2);

  info.nFloats[0] = params.size();
  floatData = params;
}
void 
Job::
get_SmoothAlignProb(Weight& weights, SeqIds& seqIds,
		    SeqId& seqId1, SeqId& seqId2, Vector<float>& params) const
{
  Assert(info.command == CMD_SMOOTH_ALIGN_PROB);
  Assert(info.sendType == SEND_ANY);
  weights.setFromPackedData(charData);
  int n = info.nInts[0];
  seqIds.assign(intData.begin(), (intData.begin() + n));
  seqId1 = intData[n];
  seqId2 = intData[n + 1];
  params = floatData;
}
void 
Job::
set_CalcMLDistance(const MSA& msaX, const MSA& msaY,
		   const DPPath& dpPath, const float& score,
		   const Vector<float>& params)
{
  clear();
  info.command  = CMD_CALC_ML_DISTANCE;
  info.sendType = SEND_ANY;
  pack(msaX, msaY, dpPath, score, params);
}
void 
Job::
get_CalcMLDistance(MSA& msaX, MSA& msaY,
		   DPPath& dpPath, float& score,
		   Vector<float>& params) const
{
  Assert(info.command == CMD_CALC_ML_DISTANCE);
  Assert(info.sendType == SEND_ANY);
  unpack(msaX, msaY, dpPath, score, params);
}
void 
Job::
set_CalcAlignDistance(const MSA& msaX, const MSA& msaY,
		   const DPPath& dpPath, const float& score,
		   const Vector<float>& params)
{
  clear();
  info.command  = CMD_CALC_ALIGN_DISTANCE;
  info.sendType = SEND_ANY;
  pack(msaX, msaY, dpPath, score, params);
}
void 
Job::
get_CalcAlignDistance(MSA& msaX, MSA& msaY,
		   DPPath& dpPath, float& score,
		   Vector<float>& params) const
{
  Assert(info.command == CMD_CALC_ALIGN_DISTANCE);
  Assert(info.sendType == SEND_ANY);
  unpack(msaX, msaY, dpPath, score, params);
}
void 
Job::
set_CalcFeatureVect(const GuideTree& guideTree, const Weight& weights,
		    const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
		    const float& score, const Vector<float>& params,
		    const bool& dpAlongPath)
{
  clear();
  info.command  = CMD_CALC_FEATURE_VECT;
  info.sendType = SEND_ANY;
  pack(guideTree, weights, msaX, msaY, dpPath, score, params);
  info.nInts[3] = 1;
  intData.push_back(dpAlongPath ? 1 : 0);
}
void 
Job::
get_CalcFeatureVect(GuideTree& guideTree, Weight& weights,
		    MSA& msaX, MSA& msaY, DPPath& dpPath,
		    float& score, Vector<float>& params,
		    bool& dpAlongPath) const
{
  Assert(info.command == CMD_CALC_FEATURE_VECT);
  Assert(info.sendType == SEND_ANY);
  unpack(guideTree, weights, msaX, msaY, dpPath, score, params);
  Assert(info.nInts[3] == 1);
  dpAlongPath = (intData[info.nInts[0] + info.nInts[1] + info.nInts[2]] > 0);
}
