/*
 * JobRunner.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "LoadStat.hpp"
#include "Rect.hpp"
#include "Trapezoid.hpp"
#include "EScript.hpp"
#include "Alpha.hpp"
#include "ScoreParam.hpp"
#include "Dmtx.hpp"
#include "Weight.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"
#include "SeqData.hpp"
#include "PairProbData.hpp"
#include "DPPath.hpp"
#include "MSA.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
#include "Aligner.hpp"
#include "PairProbData.hpp"
#include "AlignProbData.hpp"
#include "Job.hpp"
#include "JobRunner.hpp"

#ifndef RELEASE0
//#include "EvalAlign.hpp"
//#include "EDistTuner.hpp"
#endif

JobRunner::
JobRunner(SeqData& seqData, 
	  Vector<ScoreParam>& scoreParamList,
	  GuideTree& guideTree,
	  Weight& weights,
	  Aligner& aligner,
	  Aligner& aligner2,
	  Aligner& folder,
	  Aligner& folder2,
	  PairProbData& pairProbData,
	  AlignProbData& alignProbData,
	  bool pair_prob_from_msa)
  : seqData(seqData),
    scoreParamList(scoreParamList),
    guideTree(guideTree),
    weights(weights),
    aligner(aligner),
    aligner2(aligner2),
    folder(folder),
    folder2(folder2),
    pairProbData(pairProbData),
    alignProbData(alignProbData),

    job(),

    msaX(seqData),
    msaY(seqData),
    dpPath(),
    score(IMPOSSIBLE),
    params(),

    // evalAlign(),
    seqDataSbj(),
    msaXSbj(seqDataSbj),
    msaYSbj(seqDataSbj),
    msaRef(seqData),
    use_pair_prob(true),
    pair_prob_from_msa(pair_prob_from_msa)
  
{
}
JobRunner::
JobRunner(const JobRunner& jobRunner)
  : seqData(jobRunner.seqData),
    scoreParamList(jobRunner.scoreParamList),
    guideTree(jobRunner.guideTree),
    weights(jobRunner.weights),
    aligner(jobRunner.aligner),
    aligner2(jobRunner.aligner2),
    folder(jobRunner.folder),
    folder2(jobRunner.folder2),
    pairProbData(jobRunner.pairProbData),
    alignProbData(jobRunner.alignProbData),

    job(jobRunner.job),

    msaX(jobRunner.msaX),
    msaY(jobRunner.msaY),
    dpPath(jobRunner.dpPath),
    score(jobRunner.score),
    params(jobRunner.params),

    //evalAlign(jobRunner.evalAlign),
    seqDataSbj(jobRunner.seqDataSbj),
    msaXSbj(jobRunner.msaXSbj),
    msaYSbj(jobRunner.msaYSbj),
    msaRef(jobRunner.msaRef),
    use_pair_prob(jobRunner.use_pair_prob),
    pair_prob_from_msa(jobRunner.pair_prob_from_msa)
{
}
JobRunner::
~JobRunner()
{
}
void
JobRunner::
setJob(const Job& job1)
{
  job = job1;
}
Job
JobRunner::
getJob() const
{
  return job;
}
bool
JobRunner::
hasJob() const
{
  return (job.getCommand() != Job::CMD_EMPTY);
}
void
JobRunner::
run()
{
  switch (job.getCommand()) {
  case Job::CMD_WEIGHT:            job.get_Weight(weights);      break;
  case Job::CMD_GUIDE_TREE:        job.get_GuideTree(guideTree); break;
  case Job::CMD_ALIGN:             align();                      break;
  case Job::CMD_ALIGN1:            align1();                     break;
  case Job::CMD_ALIGN2:            align2();                     break;
  case Job::CMD_ALIGN3:            align3();                     break;
  case Job::CMD_REFINE:            refine();                     break;
    //case Job::CMD_REFINE2:           refine2();                    break;
  case Job::CMD_MERGE:             merge();                      break;
  case Job::CMD_CALC_LL:           calcLL();                     break;
    // case Job::CMD_CALC_SPS:          calcSPS();                    break;
    // case Job::CMD_EVAL_SCORES:       calcEvalScores();             break;
  case Job::CMD_CALC_FDF:          calcFDF();                    break;
  case Job::CMD_FOLD:              fold();                       break;
  case Job::CMD_TEST_PARAM:        testParam();                  break;
  case Job::CMD_CALC_PAIR_PROB:    calcPairProb();               break;
  case Job::CMD_BCAST_PAIR_PROB:   bcastPairProb();              break;
  case Job::CMD_CALC_ALIGN_PROB:   calcAlignProb();              break;
  case Job::CMD_BCAST_ALIGN_PROB:  bcastAlignProb();             break;
  case Job::CMD_CALC_FEATURE_VECT: calcFeatureVect();            break;
  case Job::CMD_FOLD_2STEP:        fold2Step();                  break;
    // case Job::CMD_CALC_ML_DISTANCE:  calcMLDistance();             break;
  case Job::CMD_CALC_ALIGN_DISTANCE:  calcAlignDistance();             break;
  default: Die("bad command=%d", job.getCommand());              break;
  }
}
void
JobRunner::
align()
{
  job.get_Align(guideTree, weights, msaX, msaY);

  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  NMatrix<float> alignProb;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);

  aligner2.setPurpose("CYK");
  aligner2.setScoreMode("SCORE_MODE_MEA");
  if (use_pair_prob) {
    aligner2.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner2.setMSAs(msaX, msaY);
  }
  aligner2.setAlignProb(alignProb);//should be set after setMSAs
  if (msaY.size() > 0) {
    // currently not applicable to SCFG
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }
  aligner2.setDPPath(dpPath);
  LoadStat::setAccountForWidth(false);
  aligner2.run();
  LoadStat::setAccountForWidth(true);
  dpPath = aligner2.getDPPath();

  SeqData seqData1;// current family should be copied
  for (int i = 0; i < (int)msaX.size(); i++) {seqData1.addSeq(msaX.getSeq(i));}
  for (int i = 0; i < (int)msaY.size(); i++) {seqData1.addSeq(msaY.getSeq(i));}
  MSA msaX1(seqData1);
  MSA msaY1(seqData1);
  msaX1.setByNames(msaX);
  msaY1.setByNames(msaY);
  msaX1.merge(msaY1, dpPath);
  msaX1.setByNames(msaX);
  msaY1.setByNames(msaY);
  float seqSim = msaX1.computeSeqIdentity(msaY1);

  aligner.setPurpose("CYK");
  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setSequenceSimilarity(seqSim);
  aligner.setAlignProb(alignProb);//should be set after setMSAs
  aligner.setDPPath(dpPath);
  aligner.run();
  dpPath = aligner.getDPPath();
  score = aligner.getScore();

  ostringstream oss;
  oss << aligner.getLogString();

  //bool print_sci = true;
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci: " << sci << "\n";
  //}
  
  job.set_AlignResult(msaX, msaY, dpPath, score, params, oss.str());
}
void
JobRunner::
align1()
{
  job.get_Align1(guideTree, weights, msaX, msaY);

  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  NMatrix<float> alignProb;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);

  aligner2.setPurpose("CYK");
  aligner2.setScoreMode("SCORE_MODE_MEA");
  if (use_pair_prob) {
    aligner2.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner2.setMSAs(msaX, msaY);
  }
  aligner2.setAlignProb(alignProb);//should be set after setMSAs
  if (msaY.size() > 0) {
    // currently not applicable to SCFG
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }
  aligner2.setDPPath(dpPath);
  LoadStat::setAccountForWidth(false);
  aligner2.run();
  LoadStat::setAccountForWidth(true);
  dpPath = aligner2.getDPPath();

  SeqData seqData1;// current family should be copied
  for (int i = 0; i < (int)msaX.size(); i++) {seqData1.addSeq(msaX.getSeq(i));}
  for (int i = 0; i < (int)msaY.size(); i++) {seqData1.addSeq(msaY.getSeq(i));}
  MSA msaX1(seqData1);
  MSA msaY1(seqData1);
  msaX1.setByNames(msaX);
  msaY1.setByNames(msaY);
  msaX1.merge(msaY1, dpPath);
  msaX1.setByNames(msaX);
  msaY1.setByNames(msaY);
  float seqSim = msaX1.computeSeqIdentity(msaY1);

  aligner.setPurpose("CYK");
  aligner.setScoreMode("SCORE_MODE_MEA");
  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setAlignProb(alignProb);//should be set after setMSAs
  aligner.setDPPath(dpPath);
  aligner.setSequenceSimilarity(seqSim); 
  aligner.run();
  dpPath = aligner.getDPPath();
  score = aligner.getScore();

  ostringstream oss;
  oss << aligner.getLogString();

  //bool print_sci = true;
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci: " << sci << "\n";
  //}
  
  job.set_AlignResult(msaX, msaY, dpPath, score, params, oss.str());
}
void
JobRunner::
align2()
{
  job.get_Align2(guideTree, weights, msaX, msaY);

  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  NMatrix<float> alignProb;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);

  aligner.setPurpose("CYK");
  aligner.setScoreMode("SCORE_MODE_MEA");
  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setAlignProb(alignProb);// should be set after setMSAs
  if (msaY.size() > 0) {
    // currently not applicable to SCFG
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }
  aligner.setDPPath(dpPath);
  aligner.run();
  dpPath = aligner.getDPPath();
  score = aligner.getScore();

  ostringstream oss;
  oss << aligner.getLogString();

  //bool print_sci = true;
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci: " << sci << "\n";
  //}
  
  job.set_AlignResult(msaX, msaY, dpPath, score, params, oss.str());
}
void
JobRunner::
align3()
{
  job.get_Align3(guideTree, weights, msaX, msaY);

  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  NMatrix<float> alignProb;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);

  aligner2.setPurpose("CYK");
  aligner2.setScoreMode("SCORE_MODE_MEA");
  if (use_pair_prob) {
    aligner2.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner2.setMSAs(msaX, msaY);
  }
  aligner2.setAlignProb(alignProb);// should be set after setMSAs
  if (msaY.size() > 0) {
    // currently not applicable to SCFG
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }
  aligner2.setDPPath(dpPath);
  LoadStat::setAccountForWidth(false);
  aligner2.run();
  LoadStat::setAccountForWidth(true);
  dpPath = aligner2.getDPPath();
  score = aligner2.getScore();

  ostringstream oss;
  oss << aligner2.getLogString();

  //bool print_sci = true;
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci: " << sci << "\n";
  //}
  
  job.set_AlignResult(msaX, msaY, dpPath, score, params, oss.str());
}
//void
//JobRunner::
//align1()
//{
//  job.get_Align1(guideTree, weights, msaX, msaY);
//
//  aligner.setPurpose("CYK");
//  aligner.setMSAs(msaX, msaY);
//  if (msaY.size() > 0) {
//    // currently not applicable to SCFG
//    dpPath.setDiagonalPath(msaX.getUngappedLength(),
//			   msaY.getUngappedLength());
//  } else {
//    dpPath.setDiagonalPath(msaX.getUngappedLength(),
//			   msaX.getUngappedLength());
//  }
//  aligner.setDPPath(dpPath);
//  aligner.run();
//  dpPath = aligner.getDPPath();
//  //score = aligner.getScore();
//
//  SeqData seqData1;// current family should be copied
//  for (int i = 0; i < (int)msaX.size(); i++) {seqData1.addSeq(msaX.getSeq(i));}
//  for (int i = 0; i < (int)msaY.size(); i++) {seqData1.addSeq(msaY.getSeq(i));}
//  MSA msaX1(seqData1);
//  MSA msaY1(seqData1);
//  msaX1.setByNames(msaX);
//  msaY1.setByNames(msaY);
//  msaX1.merge(msaY1, dpPath);// Change msaX1 and msaY1. Needs instantiation of seqData
//  DPPath dpPath1;
//  dpPath1.setDiagonalPath(msaX1.getUngappedLength(),
//			  msaX1.getUngappedLength());
//  folder.setPurpose("CYK");
//  folder.setMSAs(msaX1, msaX1);
//  folder.setDPPath(dpPath1);
//  folder.run();
//  dpPath1 = folder.getDPPath();
//  score = folder.getScore();
//  dpPath.ct = dpPath1.ct;
//
//  ostringstream oss;
//  oss << aligner.getLogString();
//  bool print_sci = true;
//  if (print_sci) {
//    float sci = calcSCI(msaX, msaY, dpPath);
//    oss << "sci: " << sci << "\n";
//  }
//  string message = oss.str();
//  job.set_AlignResult(msaX, msaY, dpPath, score, params, message);
//}
//void
//JobRunner::
//align2()
//{
//  job.get_Align2(guideTree, weights, msaX, msaY);
//
//  aligner.setPurpose("CYK");
//  if (use_pair_prob) {
//    UpTriMtx<float> pairProbX;
//    UpTriMtx<float> pairProbY;
//    getPairProbs(pairProbX, pairProbY, msaX, msaY);
//    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
//  } else {
//    aligner.setMSAs(msaX, msaY);
//  }
//  if (msaY.size() > 0) {
//    // currently not applicable to SCFG
//    dpPath.setDiagonalPath(msaX.getUngappedLength(),
//			   msaY.getUngappedLength());
//  } else {
//    dpPath.setDiagonalPath(msaX.getUngappedLength(),
//			   msaX.getUngappedLength());
//  }
//  aligner.setDPPath(dpPath);
//  aligner.run();
//  dpPath = aligner.getDPPath();
//  score = aligner.getScore();
//
//  ostringstream oss;
//  oss << aligner.getLogString();
//  bool print_sci = true;
//  if (print_sci) {
//    float sci = calcSCI(msaX, msaY, dpPath);
//    oss << "sci: " << sci << "\n";
//  }
//  string message = oss.str();
//  job.set_AlignResult(msaX, msaY, dpPath, score, params, message);
//}
//void
//JobRunner::
//align3()
//{
//  job.get_Align3(guideTree, weights, msaX, msaY);
//  
//  // aligner2
//  aligner2.setPurpose("CYK");
//  aligner2.setMSAs(msaX, msaY);
//  DPPath dpPath0;
//  if (msaY.size() > 0) {
//    dpPath0.setDiagonalPath(msaX.getUngappedLength(),
//			    msaY.getUngappedLength());
//  } else {
//    dpPath0.setDiagonalPath(msaX.getUngappedLength(),
//			    msaX.getUngappedLength());
//  }
//  aligner2.setDPPath(dpPath0);
//
//  LoadStat::setAccountForWidth(false);
//  aligner2.run();
//  LoadStat::setAccountForWidth(true);
//  dpPath0 = aligner2.getDPPath();
//  //float score0 = aligner2.getScore();
//
//  // folder 1
//  //folder.setPurpose("LIKELIHOOD");
//  folder.setPurpose("CYK");
//  SeqData seqData1;// current family should be copied
//  for (int i = 0; i < (int)msaX.size(); i++) {
//    seqData1.addSeq(msaX.getSeq(i));
//  }
//  for (int i = 0; i < (int)msaY.size(); i++) {
//    seqData1.addSeq(msaY.getSeq(i));
//  }
//  MSA msaX1(seqData1);
//  MSA msaY1(seqData1);
//  msaX1.setByNames(msaX);
//  msaY1.setByNames(msaY);
//  msaX1.merge(msaY1, dpPath0);// Change msaX1 and msaY1. Needs instantiation of seqData
//  folder.setMSAs(msaX1, msaX1);
//  DPPath dpPath1;
//  dpPath1.setDiagonalPath(msaX1.getUngappedLength(),
//			  msaX1.getUngappedLength());
//  folder.setDPPath(dpPath1);
//  folder.run();
//  float score1 = folder.getScore();
//
//  //aligner
//  aligner.setPurpose("CYK");
//  aligner.setMSAs(msaX, msaY);
//  if (msaY.size() > 0) {
//    msaX1.setByNames(msaX);
//    msaY1.setByNames(msaY);
//    float seqSim = msaX1.computeSeqIdentity(msaY1);
//    aligner.setSequenceSimilarity(seqSim); 
//  }
//  aligner.setDPPath(dpPath0); //use aligner2 dpPath1
//  aligner.run();
//  DPPath dpPath2 = aligner.getDPPath();
//  //float score2 = aligner.getScore();
//
//  // folder 2
//  //folder.setPurpose("LIKELIHOOD");
//  folder.setPurpose("CYK");
//  msaX1.setByNames(msaX);
//  msaY1.setByNames(msaY);
//  msaX1.merge(msaY1, dpPath2);// change msaX1 and msaY1. needs instanciation of seqData
//  folder.setMSAs(msaX1, msaX1);
//  dpPath1.setDiagonalPath(msaX1.getUngappedLength(),
//			  msaX1.getUngappedLength());
//  folder.setDPPath(dpPath1);
//  folder.run();
//  float score3 = folder.getScore();
//
//  if (score1 > score3) {
//    score = score1;
//    dpPath = dpPath0;
//  } else {
//    score = score3;
//    dpPath = dpPath2;
//  }
//
//  ostringstream oss;
//  oss << aligner.getLogString();
//  bool print_sci = true;
//  if (print_sci) {
//    float sci = calcSCI(msaX, msaY, dpPath);
//    oss << "sci: " << sci << "\n";
//  }
//  string message = oss.str();
//  job.set_AlignResult(msaX, msaY, dpPath, score, params, message);
//}
void
JobRunner::
refine()
{
  job.get_Refine(guideTree, weights, msaX, msaY, dpPath);
  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  NMatrix<float> alignProb;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);

#if 0
  DPPath dpPath2;
  if (msaY.size() > 0) {
    folder.setPurpose("CYK");
    folder.setScoreMode("SCORE_MODE_MEA");
    folder.setMSAs(msaY, msaY, pairProbY, pairProbY);
    dpPath2.setDiagonalPath(msaY.getUngappedLength(),
  			    msaY.getUngappedLength());
    folder.setDPPath(dpPath2);
    folder.run();
    dpPath2 = folder.getDPPath();
  }
#endif

  aligner.setPurpose("CYK");
  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setAlignProb(alignProb);// should be set after setMSAs
  if (msaY.size() > 0) {
    float seqSim = msaX.computeSeqIdentity(msaY);
    aligner.setSequenceSimilarity(seqSim); //should be called after setMSAs
  }
  if (msaY.size() > 0) {
    dpPath.setFromMSAs(msaX, msaY);

#if 0
    if (!dpPath2.empty()) {
      dpPath.setCTY(dpPath2.ct);
    }
#endif
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }

  ostringstream oss;
  //bool print_sci = true;
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci0: " << sci << "\n";
  //}

  aligner.setDPPath(dpPath);
  aligner.setMaxStripWidth(1);// this is reset for each aligner.run
  aligner.setDoRefine(true); // this is reset for each aligner.run
  LoadStat::setAccountForWidth(false);
  aligner.run();
  LoadStat::setAccountForWidth(true);
  float scBefore = aligner.getScore();

  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setAlignProb(alignProb);// should be set after setMSAs
  if (msaY.size() > 0) {
    float seqSim = msaX.computeSeqIdentity(msaY);
    aligner.setSequenceSimilarity(seqSim); //should be called after setMSAs
  }
  if (msaY.size() > 0) {
    dpPath.setFromMSAs(msaX, msaY);
  } else {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
  }
  aligner.setDPPath(dpPath);
  aligner.setDoRefine(true); // this is reset for each aligner.run
  aligner.run();
  float scAfter = aligner.getScore();
  float scRatio = 
    (scAfter - scBefore) / (1.0 + abs(scBefore) + abs(scAfter));
  dpPath = aligner.getDPPath();
  score = scRatio;
  params.clear();
  params.push_back(scBefore);
  params.push_back(scAfter);

#ifdef DEBUG
  // three consistency checks

  cout << "dpPath flip X <-> Y" << endl;
  DPPath dpPath0;
  dpPath0.setFromMSAs(msaX, msaY);
  DPPath dpPath1;
  dpPath1.setFromMSAs(msaY, msaX);
  dpPath1.flipXY();
  if (!(dpPath0 == dpPath1)) {
    dpPath0.print();
    dpPath1.print();
    dpPath0.draw();
    dpPath1.draw();
    Die("dpPath0 != dpPath1");
  }

  // this is no longer true since skipSet is different
  //cout << "score along dpPath" << endl;
  //if (use_pair_prob) {
  //  aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  //} else {
  //  aligner.setMSAs(msaX, msaY);
  //}
  //aligner.setAlignProb(alignProb);// should be set after setMSAs
  //aligner.setDPPath(dpPath);
  //aligner.setMaxStripWidth(1);
  //LoadStat::setAccountForWidth(false);
  //aligner.run();
  //LoadStat::setAccountForWidth(true);
  //float score1 = aligner.getScore();
  //Check(abs(score1 - scAfter) < (0.1 / (1.0 + abs(score1) + abs(scAfter))),
  //	 "score1=%f, scAfter=%f", score1, scAfter);

  // if skipSize > 1 score system is asymmetric
  // skip points should also be flipped
  // so below is incorrect for skipSize > 1
  // determination of strip region is also slightly asymmetric
  // 
  //cout << "flip X <-> Y" << endl;
  //aligner.setMSAs(msaY, msaX);
  //DPPath dpPath2;
  //dpPath2.setFromMSAs(msaY, msaX);
  //aligner.setDPPath(dpPath2);
  //aligner.run();
  //float score2 = aligner.getScore();
  //Check(abs(score2 - scAfter) < 0.01,
  //	 "score2=%f, scAfter=%f", score2, scAfter);
  // if multiple paths with the same score
  // dpPath can be different. 
  //
  //dpPath2 = aligner.getDPPath();
  //dpPath2.flipXY();
  //if (!(dpPath2 == dpPath)) {
  //  dpPath2.print();
  //  dpPath.print();
  //  dpPath2.draw();
  //  dpPath.draw();
  //  Die("dpPath2 != dpPath");
  //}
#endif

  oss << aligner.getLogString();
  //if (print_sci) {
  //  float sci = calcSCI(msaX, msaY, dpPath);
  //  oss << "sci: " << sci << "\n";
  //}
  string message = oss.str();
  job.set_AlignResult(msaX, msaY, dpPath, score, params, message);
}
//void
//JobRunner::
//refine2()
//{
//  job.get_Refine2(guideTree, weights, msaX, msaY, dpPath);
//
//  // folder 1
//  //folder.setPurpose("LIKELIHOOD");
//  folder.setPurpose("CYK");
//  MSA msaXY(msaX);
//  msaXY.merge(msaY);
//  folder.setMSAs(msaXY, msaXY);
//  DPPath dpPath1;
//  dpPath1.setDiagonalPath(msaXY.getUngappedLength(),
//			  msaXY.getUngappedLength());
//  folder.setDPPath(dpPath1);
//  folder.run();
//  float scBefore = folder.getScore();
//
//  // aligner
//  aligner.setPurpose("CYK");
//  aligner.setMSAs(msaX, msaY);
//  if (msaY.size() > 0) {
//    float seqSim = msaX.computeSeqIdentity(msaY);
//    aligner.setSequenceSimilarity(seqSim); //should be called after setMSAs
//  }
//  if (msaY.size() > 0) {
//    dpPath.setFromMSAs(msaX, msaY);
//  } else {
//    dpPath.setDiagonalPath(msaX.getUngappedLength(),
//			   msaX.getUngappedLength());
//  }
//
//  ostringstream oss;
//  //bool print_sci = true;
//  //if (print_sci) {
//  //  float sci = calcSCI(msaX, msaY, dpPath);
//  //  oss << "sci0: " << sci << "\n";
//  //}
//
//  aligner.setDPPath(dpPath);
//  aligner.run();
//  float score1 = aligner.getScore();
//  dpPath = aligner.getDPPath();
//
//  Die();
//  // folder 2 
//  // wrong! guideTree and weights should be copied 
//  // since seqId is only unique in each instance of seqData's
//  SeqData seqData1;// current family should be copied
//  for (int i = 0; i < (int)msaX.size(); i++) {
//    seqData1.addSeq(msaX.getSeq(i));
//  }
//  for (int i = 0; i < (int)msaY.size(); i++) {
//    seqData1.addSeq(msaY.getSeq(i));
//  }
//  MSA msaX1(seqData1);
//  MSA msaY1(seqData1);
//  msaX1.setByNames(msaX);
//  msaY1.setByNames(msaY);
//  msaX1.merge(msaY1, dpPath);// Change msaX1 and msaY1.
//  folder.setPurpose("CYK");
//  folder.setMSAs(msaX1, msaX1);
//  dpPath1.setDiagonalPath(msaX1.getUngappedLength(),
//			  msaX1.getUngappedLength());
//  folder.setDPPath(dpPath1);
//  folder.run();
//  float scAfter = folder.getScore();
//  dpPath1 = folder.getDPPath();
//
//  float scRatio = (scAfter - scBefore) / (1.0 + abs(scBefore) + abs(scAfter));
//
//  Assert(dpPath.getAlignLen() == dpPath1.getAlignLen());
//  dpPath.ct = dpPath1.ct;
//  score = scRatio;
//  params.clear();
//  params.push_back(scBefore);
//  params.push_back(scAfter);
//  params.push_back(score1);
//
//  //oss << aligner.getLogString();
//  //if (print_sci) {
//  //  float sci = calcSCI(msaX, msaY, dpPath);
//  //  oss << "sci: " << sci << "\n";
//  //}
//  string message = oss.str();
//  job.set_AlignResult(msaX, msaY, dpPath, score, params, message);
//}
void
JobRunner::
merge()
{
  job.get_Merge(msaX, msaY, dpPath);

  msaX.merge(msaY, dpPath);
  msaY.clear();
  dpPath.clear();

  job.set_Merge(msaX, msaY, dpPath);
}
void
JobRunner::
calcLL()
{
  job.get_CalcLL(guideTree, weights,
		 msaX, msaY, dpPath, score, params);

  aligner.setPurpose("LIKELIHOOD");
  aligner.setParamVector(params);

  aligner.setMSAs(msaX, msaY);
  aligner.setDPPath(dpPath);
  aligner.setDPAlongPath(true);
  LoadStat::setAccountForWidth(false);
  aligner.run();
  LoadStat::setAccountForWidth(true);//dirty method
  const float& z1 = aligner.getScore();
  aligner.setDPPath(dpPath);
  aligner.run();
  const float& z = aligner.getScore();
  const float& len = 0.5 * (msaX.getUngappedLength() 
			    + msaY.getUngappedLength());
  score = ((z1 - z) / (float) len);

  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}
// void
// JobRunner::
// calcSPS()
// {
//   job.get_CalcSPS(guideTree, weights, msaX, msaY,
// 		  dpPath, score, params);
// 
//   aligner.setPurpose("CYK");
//   aligner.setParamVector(params);
// 
//   seqDataSbj.clear();
//   for (int i = 0; i < (int) msaX.size(); i++) {
//     seqDataSbj.addSeq(msaX.getSeq(i));
//   }
//   for (int i = 0; i < (int) msaY.size(); i++) {
//     seqDataSbj.addSeq(msaY.getSeq(i));
//   }
//   msaXSbj.setByNames(msaX);
//   msaYSbj.setByNames(msaY);    
//   msaXSbj.removeGapCols();
//   msaYSbj.removeGapCols();
// 
//   if (msaYSbj.size() > 0) {
//     // currently not applicable to SCFG
//     dpPath.setDiagonalPath(msaXSbj.getUngappedLength(),
// 			   msaYSbj.getUngappedLength());
//   } else {
//     dpPath.setDiagonalPath(msaXSbj.getUngappedLength(),
// 			   msaXSbj.getUngappedLength());
//   }
// 
//   aligner.setDPPath(dpPath);
//   aligner.setMSAs(msaXSbj, msaYSbj);
//   aligner.run();
// 
//   msaXSbj.merge(msaYSbj, aligner.getDPPath());
//   msaRef.setByNames(msaXSbj);
// 
//   evalAlign.setMSAs(msaRef, msaXSbj);
//   evalAlign.setCTRef(CT(msaRef.getSSCons()));
// 
//   dpPath = aligner.getDPPath();
//   score = evalAlign.getSPS();
// 
//   job.set_AlignResult(msaX, msaY, dpPath, score, params);
// }
void
JobRunner::
getPairProbs(UpTriMtx<float>& pairProbX, UpTriMtx<float>& pairProbY,
	     const MSA& msaX, const MSA& msaY)
{
  //bool pair_prob_from_msa = false;
  if (pair_prob_from_msa) {
    DPPath dpPath1;
    folder2.setPurpose("INSIDE_OUTSIDE");
    folder2.setScoreMode("SCORE_MODE_SIM_0");
    folder2.setMSAs(msaX, msaX); 
    dpPath1.setDiagonalPath(msaX.getUngappedLength(),
			    msaX.getUngappedLength());
    folder2.setDPPath(dpPath1);
    folder2.run();
    folder2.getPairProb(pairProbX);
    if (msaY.size() > 0) {
      if (&msaX != &msaY) {
	folder2.setMSAs(msaY, msaY);
	dpPath1.setDiagonalPath(msaY.getUngappedLength(),
				msaY.getUngappedLength());
	folder2.setDPPath(dpPath1);
	folder2.run();
	folder2.getPairProb(pairProbY);
      }
    } else {
      pairProbY.clear();
    }
  } else {
    pairProbData.getPairProb(pairProbX, weights, msaX);
    if (msaY.size() > 0) {
      if (&msaX != &msaY) {
	pairProbData.getPairProb(pairProbY, weights, msaY);
      }
    } else {
      pairProbY.clear();
    }
  }
}
void
JobRunner::
calcFDF()
{
  job.get_CalcFDF(guideTree, weights, msaX, msaY, dpPath, score, params);

  aligner.setPurpose("INSIDE_OUTSIDE");
  aligner.setScoreMode("SCORE_MODE_SIM_0");
  aligner.setParamVector(params);

  if (use_pair_prob) {
    UpTriMtx<float> pairProbX;
    UpTriMtx<float> pairProbY;
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  if (msaY.size() > 0) {
    //float seqSim = msaX.computeSeqIdentity(msaY);
    //aligner.setSequenceSimilarity(seqSim);
  }
  aligner.setDPPath(dpPath);
  aligner.setDPAlongPath(true);// this is reset for each aligner.run
  int skipSize = aligner.getSkipSize();
  aligner.setSkipSize(1);
  LoadStat::setAccountForWidth(false);
  aligner.run();
  LoadStat::setAccountForWidth(true);
  const float& scBefore = aligner.getScore();
  Vector<float> fvec0 = aligner.computeFeatureVector();

  if (IsPossible(scBefore)) {
    aligner.setDPPath(dpPath);
    aligner.setSkipSize(skipSize);
    aligner.run();
    const float& scAfter = aligner.getScore();
    Vector<float> fvec = aligner.computeFeatureVector();

    params.resize(fvec0.size());
    transform(fvec0.begin(), fvec0.end(), fvec.begin(), 
	      params.begin(), minus<float>());
    score = (scBefore - scAfter);

    // normalize by length
    float len;
    if (msaY.size() > 0) {
      len = 0.5 * (msaX.getUngappedLength() + msaY.getUngappedLength());
    } else {
      len = msaX.getUngappedLength();
    }
    float len_factor = 100.0 * (1.0 / (float)len);
    transform(params.begin(), params.end(), params.begin(),
	      bind2nd(multiplies<float>(), len_factor));
    score *= len_factor;
  } else {
    score = scBefore;
    params = fvec0;
  }
  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}
void
JobRunner::
testParam()
{
  job.get_TestParam(guideTree, weights, msaX, msaY, 
		    dpPath, score, params);

  aligner.setPurpose("CYK");
  aligner.setParamVector(params);

  aligner.setMSAs(msaX, msaY);
  aligner.setDPPath(dpPath);
  aligner.run();
  dpPath = aligner.getDPPath();
  score = aligner.getScore();

#if 0
  msaX.print();
  msaY.print();
  dpPath.print();
#endif

  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}

#ifndef RELEASE0
void
JobRunner::
calcEvalScores()
{
  job.get_EvalScores(guideTree, weights, 
		     msaX, msaY, dpPath, score, params);

  Die("wrong since weigts and guideTrees have different id mappings");
  aligner.setPurpose("CYK");
  seqDataSbj.clear();
  for (int i = 0; i < (int) msaX.size(); i++) {
    seqDataSbj.addSeq(msaX.getSeq(i));
  }
  for (int i = 0; i < (int) msaY.size(); i++) {
    seqDataSbj.addSeq(msaY.getSeq(i));
  } 
  msaXSbj.setByNames(msaX);
  msaYSbj.setByNames(msaY);

  if (msaYSbj.size() > 0) {
    // currently not applicable to SCFG
    dpPath.setDiagonalPath(msaXSbj.getUngappedLength(),
			   msaYSbj.getUngappedLength());
  } else {
    dpPath.setDiagonalPath(msaXSbj.getUngappedLength(),
			   msaXSbj.getUngappedLength());
  }

  aligner.setDPPath(dpPath);

  aligner.setMSAs(msaXSbj, msaYSbj);
  aligner.run();

  msaXSbj.merge(msaYSbj, aligner.getDPPath());
  msaRef.setByNames(msaXSbj);

  evalAlign.setMSAs(msaRef, msaXSbj);
  evalAlign.setCTRef(CT(msaRef.getSSCons()));

  dpPath = aligner.getDPPath();
  score = evalAlign.getSPS();
  //params[0] = evalAlign.getSCISbj();
  //params[1] = evalAlign.getSCIRef();
    
#ifdef DEBUG
  cout << "msaXSbj\n";
  msaXSbj.print();
  //cout << "sci: " << evalAlign.getSCISbj() << "\n";
  cout << "msaRef\n";
  msaRef.print();
#endif

  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}
#endif

void
JobRunner::
fold()
{
  job.get_Fold(guideTree, weights, msaX, dpPath);

  UpTriMtx<float> pairProbX;
  getPairProbs(pairProbX, pairProbX, msaX, msaX);
  folder.setPurpose("CYK");
  folder.setScoreMode("SCORE_MODE_MEA");
  folder.setMSAs(msaX, msaX, pairProbX, pairProbX);
  if (dpPath.empty()) {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
    folder.setDPAlongPath(false);
  } else {
    folder.setDPAlongPath(true);
  }
  folder.setDPPath(dpPath);
  folder.run();
  dpPath = folder.getDPPath();
  Assert((int)dpPath.ct.size() == msaX.getUngappedLength());
  score = folder.getScore();

#if 0
  dpPath.print();
#endif

  job.set_AlignResult(msaX, msaX, dpPath, score, params);
}
void
JobRunner::
fold2Step()
{
  job.get_Fold2Step(guideTree, weights, msaX, dpPath);

  UpTriMtx<float> pairProb;
  //folder.getPairProb(pairProb);
  getPairProbs(pairProb, pairProb, msaX, msaX);

  folder.setPurpose("CYK");
  folder.setPurpose("SCORE_MODE_MEA");
  folder.setMSAs(msaX, msaX, pairProb, pairProb);
  if (dpPath.empty()) {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaX.getUngappedLength());
    folder.setDPAlongPath(false);
  } else {
    folder.setDPAlongPath(true);
  }
  folder.setDPPath(dpPath);
  folder.run();
  dpPath = folder.getDPPath();
  score = folder.getScore();
  params = pairProb.toVector();

  job.set_AlignResult(msaX, msaX, dpPath, score, params);
}
void 
JobRunner::
calcPairProb()
{
  UpTriMtx<float> pairProb;
  job.get_CalcPairProb(guideTree, weights, msaX);
  folder2.setPurpose("INSIDE_OUTSIDE");
  folder2.setScoreMode("SCORE_MODE_SIM_0");
  folder2.setMSAs(msaX, msaX);
  dpPath.setDiagonalPath(msaX.getUngappedLength(),
			 msaX.getUngappedLength());
  folder2.setDPPath(dpPath);
  folder2.run();
  folder2.getPairProb(pairProb);
  params = pairProb.toVector();
  job.set_AlignResult(msaX, msaX, dpPath, score, params);
}
void 
JobRunner::
bcastPairProb()
{
  int seqId;
  SparseProb sp;
  job.get_BcastPairProb(seqId, sp);
  SparseProb& mtx = pairProbData.ref(seqId);
  mtx = sp;
  sp.clear();
  job.set_BcastPairProb(seqId, sp);
}
void
JobRunner::
calcAlignProb()
{
  NMatrix<float> mtx;
  job.get_CalcAlignProb(guideTree, weights, msaX, msaY);
  aligner2.setPurpose("INSIDE_OUTSIDE");
  aligner2.setScoreMode("SCORE_MODE_SIM_0");
  dpPath.setFromMSAs(msaX, msaY);
  aligner2.setDPPath(dpPath);
  aligner2.setDPAlongPath(false);
  aligner2.setMSAs(msaX, msaY);
  LoadStat::setAccountForWidth(false);
  aligner2.run();
  LoadStat::setAccountForWidth(true);
  aligner2.getAlignProb(mtx);
  params.assign(mtx.begin(), mtx.end());
  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}
void
JobRunner::
bcastAlignProb()
{
  int seqId1;
  int seqId2;
  SparseProb sp;
  job.get_BcastAlignProb(seqId1, seqId2, sp);
  SparseProb& mtx = alignProbData.ref(seqId1, seqId2);
  mtx = sp;
  SparseProb& mtx_r = alignProbData.ref(seqId2, seqId1);
  mtx.transpose(mtx_r);

  sp.clear();
  job.set_BcastAlignProb(seqId1, seqId2, sp);
}
//void
//JobRunner::
//smoothAlignProb()
//{
//  int seqId1;
//  int seqId2;
//  SeqIds seqIds;
//  job.get_SmoothAlignProb(weights, seqIds, seqId1, seqId2, params);
//  NMatrix<float> mtx;
//  alignProbData.smoothAlignProb(mtx, seqIds, seqId1, seqId2);
//  job.set_SmoothAlignProb(weights, seqIds, seqId1, seqId2, mtx.toVector());
//}
// void
// JobRunner::
// calcMLDistance()
// {
//   job.get_CalcMLDistance(msaX, msaY, dpPath, score, params);
//   Assert(msaX.size() == 1 && msaY.size() == 1);
//   SeqIds seqIdsXY(msaX.getSeqIds());
//   seqIdsXY.append(msaY.getSeqIds());
//   guideTree.setFromTupleCount(seqData, seqIdsXY);
//   weights.setFlatWeight(seqIdsXY);
//   if (dpPath.empty()) {
//     Die("dpPath.empty()");
//     aligner2.setPurpose("CYK");
//     dpPath.setDiagonalPath(msaX.getUngappedLength(),
// 			   msaY.getUngappedLength());
//     aligner2.setMSAs(msaX, msaY);
//     aligner2.setDPPath(dpPath);
//     LoadStat::setAccountForWidth(false);
//     aligner2.run();
//     LoadStat::setAccountForWidth(true);
//     dpPath = aligner2.getDPPath();
//   }
//   aligner2.setMSAs(msaX, msaY);
//   aligner2.setDPPath(dpPath);
//   aligner2.setMaxLWRatio(0.01);
//   EDistTuner edt(aligner2, guideTree, msaX, msaY, dpPath);
//   edt.run();
//   score = edt.getScore();
//   float distance = edt.getDistance();
//   Check(distance > 0.0);
//   
//   params.resize(1);
//   params[0] = distance;
//   job.set_CalcMLDistance(msaX, msaY, dpPath, score, params);
// }
void
JobRunner::
calcAlignDistance()
{
  job.get_CalcAlignDistance(msaX, msaY, dpPath, score, params);
  Assert(msaX.size() == 1 && msaY.size() == 1);
  SeqIds seqIdsXY(msaX.getSeqIds());
  seqIdsXY.append(msaY.getSeqIds());
  // guideTree.setFromTupleCount(seqData, seqIdsXY);
  weights.setFlatWeight(seqIdsXY);
  NMatrix<float> alignProb;
  alignProbData.getAlignProb(alignProb, weights, msaX, msaY);
  UpTriMtx<float> pairProbX;
  UpTriMtx<float> pairProbY;
  if (use_pair_prob) {
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
  }

  float seqIdentity = 0.0;
  float score1 = IMPOSSIBLE;
  if (dpPath.empty()) {
    dpPath.setDiagonalPath(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
    aligner2.setMSAs(msaX, msaY);
    aligner2.setAlignProb(alignProb);// should be set after setMSAs
    aligner2.setDPPath(dpPath);
    aligner2.setPurpose("CYK");
    aligner2.setScoreMode("SCORE_MODE_MEA");// this is
    LoadStat::setAccountForWidth(false);
    aligner2.run();
    LoadStat::setAccountForWidth(true);
    dpPath = aligner2.getDPPath();
    score1 = aligner2.getScore();// maximal expected accuracy score

    SeqData seqData1;// current family should be copied
    for (int i = 0; i < (int)msaX.size(); i++) {seqData1.addSeq(msaX.getSeq(i));}
    for (int i = 0; i < (int)msaY.size(); i++) {seqData1.addSeq(msaY.getSeq(i));}
    MSA msaX1(seqData1);
    MSA msaY1(seqData1);
    msaX1.setByNames(msaX);
    msaY1.setByNames(msaY);
    msaX1.merge(msaY1, dpPath);// Change msaX1 and msaY1.
    //msaX1.print();
    msaX1.setByNames(msaX);
    msaY1.setByNames(msaY);
    seqIdentity = msaX1.computeSeqIdentity(msaY1);
  } else {
    seqIdentity = msaX.computeSeqIdentity(msaY);
  }
  aligner.setPurpose("CYK");
  if (use_pair_prob) {
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  aligner.setAlignProb(alignProb);// should be set after setMSAs
  if (msaY.size() > 0) {
    aligner.setSequenceSimilarity(seqIdentity); //should be called after setMSAs
  }
  aligner.setDPPath(dpPath);
  aligner.setMaxStripWidth(1);// this is reset for each aligner.run
  LoadStat::setAccountForWidth(false);
  aligner.run();
  LoadStat::setAccountForWidth(true);
  score = aligner.getScore();

#if 0
  msaX1.print();
  cout << "score" << score << endl;
  cout << "seqIdent: " << seqIdentity << endl;
#endif

  params.resize(2);
  params[0] = seqIdentity;
  params[1] = score1;//PHMM Score
  job.set_CalcAlignDistance(msaX, msaY, dpPath, score, params);
}
void
JobRunner::
calcFeatureVect()
{
  bool dpAlongPath;
  job.get_CalcFeatureVect(guideTree, weights, msaX, msaY, dpPath, score, params, dpAlongPath);
  aligner.setPurpose("INSIDE_OUTSIDE");
  aligner.setScoreMode("SCORE_MODE_SIM_0");

  if (!params.empty()) {
    aligner.setParamVector(params);
  }
  if (use_pair_prob) {
    UpTriMtx<float> pairProbX;
    UpTriMtx<float> pairProbY;
    getPairProbs(pairProbX, pairProbY, msaX, msaY);
    aligner.setMSAs(msaX, msaY, pairProbX, pairProbY);
  } else {
    aligner.setMSAs(msaX, msaY);
  }
  if (msaY.size() > 0) {
    float seqSim = msaX.computeSeqIdentity(msaY);
    aligner.setSequenceSimilarity(seqSim);
  }
  if (dpPath.empty()) {
    if (msaY.size() > 0) {
      dpPath.setDiagonalPath(msaX.getUngappedLength(),
			     msaY.getUngappedLength());
    } else {
      dpPath.setDiagonalPath(msaX.getUngappedLength(),
			     msaX.getUngappedLength());
    }
  }
  aligner.setDPPath(dpPath);
  aligner.setDPAlongPath(dpAlongPath);
  int skipSize = aligner.getSkipSize();
  if (dpAlongPath) {
    aligner.setSkipSize(1);
  }
  aligner.run();
  aligner.setSkipSize(skipSize);
  score = aligner.getScore();
  params = aligner.computeFeatureVector();

  // normalize by length
  //float len;
  //if (msaY.size() > 0) {
  //  len = 0.5 * (msaX.getUngappedLength() + msaY.getUngappedLength());
  //} else {
  //  len = msaX.getUngappedLength();
  //}
  //float len_factor = 100.0 * (1.0 / (float)len);
  //transform(params.begin(), params.end(), params.begin(),
  //	    bind2nd(multiplies<float>(), len_factor));
  //score *= len_factor;

  job.set_AlignResult(msaX, msaY, dpPath, score, params);
}

// should be treated by seqFamAligners
//float
//JobRunner::
//calcSCI(const MSA& msaX1, const MSA& msaY1, const DPPath& dpPath1)
//{
//  DPPath dpPath2;
//
//  // weights and guideTree should also be copied if used in folder
//  
//  folder.setPurpose("CYK");
//  folder.setMSAs(msaX1, msaX1);
//  dpPath2.setDiagonalPath(msaX1.getUngappedLength(),
//			  msaX1.getUngappedLength());
//  folder.setDPPath(dpPath2);
//  folder.run();
//  float scoreX = folder.getScore();
//
//  folder.setPurpose("CYK");
//  folder.setMSAs(msaY1, msaY1);
//  dpPath2.setDiagonalPath(msaY1.getUngappedLength(),
//			  msaY1.getUngappedLength());
//  folder.setDPPath(dpPath2);
//  folder.run();
//  float scoreY = folder.getScore();
//
//  
//  SeqData seqData2;// current family should be copied
//  for (int i = 0; i < (int)msaX1.size(); i++) {
//    seqData2.addSeq(msaX1.getSeq(i));
//  }
//  for (int i = 0; i < (int)msaY1.size(); i++) {
//    seqData2.addSeq(msaY1.getSeq(i));
//  }
//  MSA msaX2(seqData2);
//  MSA msaY2(seqData2);
//  msaX2.setByNames(msaX1);
//  msaY2.setByNames(msaY1);
//  msaX2.merge(msaY2, dpPath1);// Change msaX2 and msaY2.
//  folder.setMSAs(msaX2, msaX2);
//  dpPath2.setDiagonalPath(msaX2.getUngappedLength(),
//			  msaX2.getUngappedLength());
//  folder.setDPPath(dpPath2);
//  folder.run();
//  float scoreXY = folder.getScore();
//
//  return (scoreXY / (0.5 * (scoreX + scoreY)));
//}

