/*
 * Opt.cpp
 */
#include "Util.hpp"
#include "Opt.hpp"

Opt::EntryData::
EntryData()
  : name(), type(NO_ARGUMENT), dflt(), value(), valIsSet(false) {}

Opt::EntryData::
EntryData(const Entry& entry)
  : name(entry.name), type(entry.type), dflt(entry.dflt), value(),
    valIsSet(false)
{
}

Opt::EntryData::
EntryData(const EntryData& entryData)
 : name(entryData.name), type(entryData.type), dflt(entryData.dflt),
   value(entryData.value), valIsSet(entryData.valIsSet)
{
}

Opt::EntryData::
~EntryData()
{
}

Opt::EntryData&
Opt::EntryData::
operator=(const EntryData& entryData)
{
  if (this == &entryData) return *this;

  name = entryData.name;
  type = entryData.type;
  dflt = entryData.dflt;
  value = entryData.value;
  valIsSet = entryData.valIsSet;
  return *this;
}

void
Opt::EntryData::
setValue(const string& value0)
{
  if (value0.size() >= 2
      && ((value0[0] == '"' && value0[value0.size() - 1] == '"')
	  || (value0[0] == '\'' && value0[value0.size() - 1] == '\''))
      ) {
    value = value0.substr(1, value0.size() - 2);
  } else {	    
    value = value0;
  }
  valIsSet = true;
}

const string&
Opt::EntryData::
getValue() const
{
  return (valIsSet ? value : dflt); 
}

string 
Opt::EntryData::
toString() const
{
  ostringstream oss;
  oss << "{ name: " << name << ", ";
  oss << "dflt: " << dflt << ", ";
  oss << "value: " << value << ", ";
  oss << "valIsSet: " << valIsSet << " }";
  return oss.str();
}

void
Opt::EntryData::
print() const
{
  cout << toString() << flush;
}

Opt::
Opt(const Entry* const optTable) : fname(), table()
{
  for (int i = 0; optTable[i].name != NULL; i++) {
    const Entry& entry = optTable[i];
    Assert(table.find(entry.name) == table.end(),
	   "duplicated entry name %s", entry.name);
    table.insert(make_pair(string(entry.name), EntryData(entry)));
  }
}

Opt::
~Opt()
{
}

void
Opt::
setFromFile(const string& filename)
{
  ifstream fi(filename.c_str());
  string buf;

  if (!fi) {
    Warn("cannot open paramfile=%s\n", filename.c_str());
    return;
  }
  fname = filename;

  while (getline(fi, buf)) {
    StringTokenizer st(buf);
    string key = st.nextToken();
    string val = st.nextToken();
    if (key.empty() 
	|| key[0] == '#' 
	|| key.find("---") == 0
	|| key == "options:"
	|| key == "fname:") continue;
    
    bool sepFound = false;
    string::size_type pos = key.size() - 1;
    if (key[pos] == ':') {
      sepFound = true;
      key.erase(pos);
    }
    if (!hasKey(key)) {
      Die("unrecognized parameter %s in file %s",
	  key.c_str(), filename.c_str());
    }
    if (!sepFound && val == ":") {
      sepFound = true;
      val = st.nextToken();
    }
    if (!sepFound) Die("separator ':' not found in line\n  %s", 
		       buf.c_str());

    table[key].setValue(val);
  }
}
void
Opt::
setValue(const string& key, const string& value)
{
  Check(hasKey(key));
  table[key].setValue(value);
}
    
void
Opt::
setFromArgs(int& argc, char**& argv)
{
  static const string paramFile = "paramfile";
  int i = 1;
  while ((i < argc) && (argv[i][0] == '-')) {
    const string& str = string(argv[i] + (argv[i][1] == '-' ? 2 : 1));
    bool hasEqual = (str.find('=') != string::npos);
    StringTokenizer st(str);
    const string& key = st.nextToken("=");
    string val = st.nextToken();

    if (!hasKey(key) && key != paramFile) {
      Die("unrecognized option %s", key.c_str());
    }
    i++;
    
    if (val.empty()) {
      if (table[key].requiredArgument() || key == paramFile) {
	if (hasEqual || (i >= argc) || (argv[i][0] == '-')) {
	  Die("no argument for parameter %s", key.c_str());
	} else {
	  val = argv[i];
	  i++;
	}
      } else {
	val = (hasEqual ? "false" : "true");
      }
    }

    if (key == paramFile) {
      setFromFile(val);
    } else {
      table[key].setValue(val);
    }
  }

  for (int j = i; j < argc; j++) {
    if (argv[j][0] == '-') {
      Die("parsing stopped at argv[%d] = %s", i, argv[i]);
    }
  }

  argc -= i;
  argv += i;
  return;
}

bool
Opt::
hasKey(const string& name) const
{
  return (table.find(name) != table.end());
}

const string&
Opt::
getStr(const string& name) const
{
  if (!hasKey(name)) Die("unrecognized option %s", name.c_str());

  return table.find(name)->second.getValue();
}

int
Opt::
getInt(const string& name) const
{
  if (!hasKey(name)) Die("unrecognized option %s", name.c_str());

  return atoi(table.find(name)->second.getValue().c_str());
}

double
Opt::
getDbl(const string& name) const
{
  if (!hasKey(name)) Die("unrecognized option %s", name.c_str());

  return atof(table.find(name)->second.getValue().c_str());
}

bool
Opt::
getBool(const string& name) const
{
  if (!hasKey(name)) Die("unrecognized option %s", name.c_str());

  static const string trueValues[] = {
    "true", "1", "yes", "on", "t", "y", ""
  };

  const string& value = table.find(name)->second.getValue();
  for (const string* p = trueValues; !p->empty(); p++) {
    if (value == *p) return true;
  }
  return false;
}

void
Opt::
checkIntLowerBound(const string& name, int minValue) const
{
  if (getInt(name) < minValue)
    Die("%s < %d is not a valid option", name.c_str(), minValue);
}

void
Opt::
checkIntUpperBound(const string& name, int maxValue) const
{
  if (getInt(name) > maxValue)
    Die("%s > %d is not a valid option", name.c_str(), maxValue);
}

void
Opt::
checkIntRange(const string& name, int minValue, int maxValue) const
{
  checkIntLowerBound(name, minValue);
  checkIntUpperBound(name, maxValue);
}

void
Opt::
checkDblLowerBound(const string& name, double minValue) const
{
  if (getDbl(name) < minValue)
    Die("%s < %lf is not a valid option", name.c_str(), minValue);
}

void
Opt::
checkDblUpperBound(const string& name, double maxValue) const
{
  if (getDbl(name) > maxValue)
    Die("%s > %lf is not a valid option", name.c_str(), maxValue);
}

void
Opt::
checkDblRange(const string& name, double minValue, double maxValue) const
{
  checkDblLowerBound(name, minValue);
  checkDblUpperBound(name, maxValue);
}

void
Opt::
checkStrEntry(const string& name, const string& entries) const
{
  StringTokenizer st(entries);
  const string& value = getStr(name);
  while (st.hasMoreTokens()) {
    if (value == st.nextToken()) return;
  }
  Die("%s=%s is not a valid option", name.c_str(), value.c_str());
}

void
Opt::
printHeader(ostream& fo) const
{
  for (Table::const_iterator ti = table.begin(); ti != table.end(); ti++) {
    const string& key = ti->first;
    const string& value = ti->second.getValue();
    fo << "#" << key << "\t" << value << endl;
  }
}

void
Opt::
printYAML(ostream& fo) const
{
  const string& indent = "  ";
  fo << "---\noptions:\n";
  fo << indent << "fname: " << fname << "\n";
  for (Table::const_iterator ti = table.begin(); ti != table.end(); ti++) {
    const EntryData& entryData = ti->second;
    if (entryData.valueIsSet()) {
      fo << indent << entryData.getName() << ": ";
      fo << entryData.getValue() << "\n";
    }
  }
  fo << flush;
}

void
Opt::
setFromYAML(istream& fi)
{
  string buf;
  while (getline(fi, buf)) {
    StringTokenizer st(buf);
    while (st.hasMoreTokens()) {
      const string& tok = st.nextToken();
      if (tok == "---" || tok == "options:") continue;
      if (tok == "fname:") {
	fname = st.nextToken();
	continue;
      }

      const string& name = tok.substr(0, tok.size() - 1);
      if (hasKey(name)) {
	table[name].setValue(st.nextToken());
      } else {
	Die("unrecognized option %s", name.c_str());
      }
    }
  }
}

string
Opt::
toString() const
{
  ostringstream oss;
  oss << "fname: " << fname << "\n";
  oss << "nopt: " << table.size() << "\n";
  for (Table::const_iterator ti = table.begin(); ti != table.end(); ti++) {
    const string& key = ti->first;
    const string& value = ti->second.toString();
    oss << key << ": " << value << "\n";
  }
  return oss.str();
}
void
Opt::
print() const 
{
  cout << toString() << flush;
}
  
#if false
// check behavior
Opt::Entry optTable[] = {
  {"leneage",        Opt::REQUIRED_ARGUMENT, "WAVG"},
  {"idxbit",         Opt::REQUIRED_ARGUMENT, "2"},
  {"wdsiz",          Opt::REQUIRED_ARGUMENT, "3"},
  {"lensft",         Opt::REQUIRED_ARGUMENT, "1"},
  {"lenbit",         Opt::REQUIRED_ARGUMENT, "8"},
  {"block_size",     Opt::REQUIRED_ARGUMENT, "10"},
  {"init_size",      Opt::REQUIRED_ARGUMENT, "20"},
  {"gapOpenScore",   Opt::REQUIRED_ARGUMENT, "-5.0"},
  {"gapExtendScore", Opt::REQUIRED_ARGUMENT, "-1.0"},
  {"lambda",         Opt::REQUIRED_ARGUMENT, "5.0"},
  {"smtxfile",       Opt::REQUIRED_ARGUMENT, "RIBOSUM45.mat"},
  {"seqfile",        Opt::REQUIRED_ARGUMENT, "seqdata1.txt"},
  {"outfile",        Opt::REQUIRED_ARGUMENT, "outfile.txt"},
  {"RefineTreeMaxIter", Opt::REQUIRED_ARGUMENT, "3"},
  {"flag",           Opt::NO_ARGUMENT, NULL},
  {NULL,             Opt::NO_ARGUMENT,       NULL}
};

int
main()
{
#ifdef DMALLOC
  dmalloc_debug_setup(getenv("DMALLOC_OPTIONS"));
#endif
  Opt opt(optTable);
  ostream& fo = cout;
  char* args[] = {
    "main", "-idxbit", "1000", "-seqfile=data.txt", "file"
  };
  int argc = sizeof(args) / sizeof(char*);
  char** argv = args;

  opt.print(fo);
  opt.setFromFile("../data/params.txt");
  opt.print(fo);
  opt.setFromArgs(argc, argv);
  opt.print(fo);

  fo << "remaining arguments:" << argc << "\n";
  for (int i = 0; i < argc; i++) {
    fo << argv[i] << endl;
  }


  fo << "size:" << opt.getInt("size") << "\n";
  fo << "double:" << opt.getDbl("lambda") << "\n";
  fo << "flag:" << opt.getBool("flag") << "\n";
  fo << "outfile:" << opt.getStr("outfile") << endl;

  return 0;
}

#endif
