/*
 * Rect.hpp
 *
 */
#ifndef RECT_HPP
#define RECT_HPP
#include "Util.hpp"

struct Point {
  int x;
  int y;
  string toString() const {
    ostringstream oss;
    oss << "[ " << x << ", " << y << " ]";
    return oss.str();
  }
  void print() const {
    cout << toString() << endl;
  }
  bool operator==(const Point& pt) const {
    return (x == pt.x && y == pt.y);
  }
  static const Point& null() {
    static const Point nullPoint = {-1, -1};
    return nullPoint;
  }
  bool isNull() const {return (*this == null());}
  bool isNormal() const {return (0 <= x && 0 <= y);}
  bool isValid() const {return (isNull() || isNormal());}
  void setNull() {*this = null();}
  void clean() {memset(this, 0, sizeof(*this));}
  Point& setVal(int i, int j) {
    x = i;
    y = j;
    return *this;
  }
};

struct Rect {
  union {
    Point b;
    struct {int bx, by;};
  };
  union {
    Point e;
    struct {int ex, ey;};
  };
  // int bx;
  // int by;
  // int ex;
  // int ey;

  string toString() const {
    ostringstream oss;
    oss << "[ " << bx << ", " << by << ", " << ex << ", " << ey << " ]";
    return oss.str();
  }
  void print() const {
    cout << toString() << endl;
  }
  bool operator==(const Rect& rect) const {
    return ((b == rect.b) && (e == rect.e));
  }
  static const Rect& null() {
    static const Rect nullRect = {{Point::null()}, {Point::null()}};
    return nullRect;
  }
  bool isNull() const {return (*this == Rect::null());}
  bool isNormal() const {
    return (0 <= bx && bx <= ex) && (0 <= by && by <= ey);
  }
  bool isValid() const {
    return (isNull() || isNormal());  } 
  void setNull() {*this = Rect::null();}
  void clean() {memset(this, 0, sizeof(*this));}
  Rect& setVal(int i, int j, int k, int l) {
    bx = i, by = j, ex = k, ey = l;
    return *this;
  }
  Rect& setVal(const Point& b1, const Point& e1) {
    b = b1, e = e1;
    return *this;
  }
  bool includes(int x, int y) const {
    return (bx <= x && x <= ex && by <= y && y <= ey);
  }
  bool includes(const Point& pt) const {
    return (bx <= pt.x && pt.x <= ex && by <= pt.y && pt.y <= ey);
  }
  bool includes(const Rect& rect) const {
    return (includes(rect.b) && includes(rect.e));
  }
  int lengthX() const {return (ex - bx + 1);}
  int lengthY() const {return (ey - by + 1);}
};


class RectIterator {
public:
  Rect rect0;

  RectIterator(Rect &rect, Point &pt) {
    init(rect, pt);
  };

  void init(Rect &rect, Point &pt) {
    rect0 = rect;
    initPoint(pt);
  }

  void initPoint(Point &pt) {
    pt.x = rect0.bx;
    pt.y = rect0.by - 1;
  }

  bool getNext(Point &pt) {
    pt.y++;
    if (pt.y < rect0.ey) {
      return true;
    } else {
      pt.x++;
      if (pt.x < rect0.ex) {
	pt.y = rect0.by;
	return true;
      } else {
	return false;
      }
    }
  }
  
};

struct Hexagon {
  union {
    Rect rect;
    struct {
      union {
	Point b;
	struct {int bx, by;};
      };
      union {
	Point e;
	struct {int ex, ey;};
      };
    };
  };
  union {
    Point d;
    struct {int dx, dy;};
  };

  Hexagon() {clear();}
  Hexagon(int bx, int by, int ex, int ey, int dx, int dy)
    : bx(bx), by(by), ex(ex), ey(ey), dx(dx), dy(dy) {};
  Hexagon(Point& b, Point& e, int dx, int dy)
    : b(b), e(e), dx(dx), dy(dy) {};
  Hexagon(Rect& rect, int dx, int dy) : rect(rect), dx(dx), dy(dy) {};
  void clear() {memset(this, 0, sizeof(*this));}
  void setVal(int bx0, int by0, int ex0, int ey0, int dx0, int dy0) {
    bx = bx0, by = by0, ex = ex0, ey = ey0, dx = dx0, dy = dy0;
  }
  void setVal(Point& b0, Point& e0, int dx0, int dy0) {
    b = b0, e = e0, dx = dx0, dy = dy0;
  }
  void setVal(Rect& rect0, int dx0, int dy0) {
    rect = rect0, dx = dx0, dy = dy0;
  }
  void setVal(Hexagon& hex) {
    *this = hex;
  }

  void checkConsistency() {
#ifdef DEBUG
    int lx = ex - bx;
    int ly = ey - by;
    Assert(lx > 0);
    Assert(ly > 0);
    Assert(dx > 0);
    Assert(dy > 0);
    Assert(dx > lx - ly);
    Assert(dy > ly - lx);
#endif
  }
  bool isType1() {return (dx + dy) > (ey - by);}
  bool includes(int x, int y) const {
    const int d = (x - bx) - (y - by);
    return ((dx < d && d < dy) && rect.includes(x, y));
  }
  int getArea() const {
    const int lx = ex - bx + 1;
    const int ly = ey - by + 1;
    const int dlx = lx - dx;
    const int dly = lx - dy;
    return ((lx * ly) 
	    - ((dlx * (dlx + 1)) / 2) 
	    - ((dly * (dly + 1)) / 2));
  }
  int getAreaBefore(int x, int y) {
    Assert(includes(x, y));
    const int lx1 = x - bx;
    const int dlx1 = lx1 - dx + 1;
    const int ly = ex - bx + 1;
    const int dly = ly - dy;
    int n = lx1 * ly + (y - by);
  
    if (dlx1 > 0) {
      n -= dlx1 * (dlx1 + 1) / 2;
    }

    if (lx1 >= dly) {
      n -= dly * (dly - 1) / 2;
    } else {
      n -= lx1 * dly;
      n += lx1 * (lx1 - 1) / 2;
    }
    return n;
  }
};

#endif
