/*
 * Scheduler.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "LoadStat.hpp"
#include "Rect.hpp"
#include "Trapezoid.hpp"
#include "EScript.hpp"
#include "Alpha.hpp"
#include "Profile.hpp"
#include "ScoreParam.hpp"
#include "Dmtx.hpp"
#include "Weight.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"
#include "SeqData.hpp"
#include "BasePairData.hpp"
#include "DPPath.hpp"
#include "MSA.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
#include "Aligner.hpp"
// #include "EvalAlign.hpp"
#include "Job.hpp"
#include "JobRunner.hpp"
// #include "MPIHandle.hpp"
#include "Scheduler.hpp"

#ifndef HAVE_MPI_CPP
#define SEQUENTIAL
#endif

// pthread_mutex_t Scheduler::mutex = PTHREAD_MUTEX_INITIALIZER;
// #define MUTEX (Scheduler::mutex)
#define MUTEX

#ifdef DEBUG
#define dmsg(obj) do {                                             \
  cerr << "dmsg: " << __PRETTY_FUNCTION__ << ": " << obj << endl;  \
} while (0)
#define dmsg2(obj1, obj2) do {                     \
  cerr << "dmsg: " << __PRETTY_FUNCTION__ << ": "; \
  cerr << obj1 << obj2 << endl;                    \
} while (0)
#else
#define dmsg(obj)
#define dmsg2(obj1, obj2)
#endif

// #define LOCK() do {           
//   /* pthread_mutex_lock(&MUTEX); \ */
//   /* dmsg("lock");               \ */
// } while (0)
// #define UNLOCK() do {           
//   /* pthread_mutex_unlock(&MUTEX); \ */
//   /* dmsg("unlock");               \ */
// } while (0)
// 
// #ifdef __CYGWIN__
// #define COND_WAIT(cv) do {      
//   /* pthread_mutex_unlock(&MUTEX); \ */
//   /* usleep(500000);  0.5sec  \      */
//   /* pthread_mutex_lock(&MUTEX);   \ */
// } while (0)
// #else
// #define COND_WAIT(cv) do {        
//   /* dmsg("cond wait in");           \ */
//   /* pthread_cond_wait(&cv, &MUTEX); \ */
//   /* dmsg("cond wait out");          \ */
// } while (0)
// #endif
// 
// #define TIMED_WAIT() do {      
//   /* pthread_mutex_unlock(&MUTEX);\ */
//   /* sleep(1);			  \ */
//   /* pthread_mutex_lock(&MUTEX);  \ */
// } while (0)

#define LOCK() 
#define UNLOCK()
#define COND_WAIT(cv) 
#define TIMED_WAIT() 

Scheduler::TaggedJob::
TaggedJob(const Job& job, int id, int clientID, int servantID)
  : job(job), id(id), clientID(clientID), servantID(servantID)
{
}
Scheduler::TaggedJob::
TaggedJob(const TaggedJob& taggedJob)
  : job(taggedJob.job),
    id(taggedJob.id),
    clientID(taggedJob.clientID),
    servantID(taggedJob.servantID)
{
}
Scheduler::TaggedJob::
~TaggedJob()
{
}
Scheduler::ClientState::
ClientState()
  : isWaiting(false), results(), remaining(0)
{
  // pthread_cond_init(&cv, NULL);
}
Scheduler::ClientState::
~ClientState()
{
  // pthread_cond_destroy(&cv);
}
string
Scheduler::ClientState::
toString() const
{
  ostringstream oss;
  oss << "{ isWaiting: " << isWaiting << ", ";
  oss << "results.size(): " << results.size() << ", ";
  oss << "remaining: " << remaining << " }";
  return oss.str();
}
Scheduler::ServantState::
ServantState()
  : isBusy(false), jobQ()
{
}
Scheduler::ServantState::
~ServantState()
{
}
string
Scheduler::ServantState::
toString() const
{
  ostringstream oss;
  oss << "{ isBusy: " << isBusy << ", ";
  oss << "jobQ.size(): " << jobQ.size() << " }";
  return oss.str();
}
Scheduler::Channel::
Channel(Scheduler& scheduler)
  : scheduler(scheduler), shutdownRequested(false)
{
  // pthread_cond_init(&cv, NULL);
}
Scheduler::Channel::
~Channel()
{
  // pthread_cond_destroy(&cv);
}
string
Scheduler::Channel::
toString() const
{
  ostringstream oss;
  oss << "{ shutdownRequested: " << shutdownRequested << " }";
  return oss.str();
}
void*
Scheduler_Channel_run(void* pScheduler)
{
  Scheduler& scheduler = *(Scheduler*)pScheduler;
  Scheduler::Channel& channel = scheduler.channel;
  channel.run();
  return NULL;
}
void
Scheduler::Channel::
run()
{
  LOCK();
  while (!shutdownRequested || scheduler.remaining > 0) {
    int nProcessed = 0;
    nProcessed += collectResults();
    nProcessed += throwJobs();
    dmsg2("nProcessed=", nProcessed);
    if (nProcessed > 0) continue;

    if (servantsAreAllBusy()) {
      dmsg("servants are all busy");
      waitAnyServant();
    } else if (clientsAreAllWaiting()) {
      dmsg("clients are all waiting");
      waitAnyServant();
    } else if (!shutdownRequested && scheduler.remaining == 0) {
      dmsg("remaining == 0");
      waitAnyClient();
    } else {
      timedWaitAnyClient();
    }
  }
  UNLOCK();
}
int
Scheduler::Channel::
collectResults()
{
  int nProcessed = 0;
  for (int i = 0; i < (int)scheduler.servantStates.size(); i++) {
    ServantState& servantState = scheduler.servantStates[i];
    if (!servantState.isBusy) continue;

    if (resultIsArrived(i)) {
      TaggedJob& taggedJob = servantState.jobQ.front();
      taggedJob.job = recvResult(i);
      if (taggedJob.clientID != TaggedJob::CHANNEL_CLIENT) {
	ClientState& clientState = 
	  scheduler.clientStates.find(taggedJob.clientID)->second;
	if (!taggedJob.job.sendAll()) {
	  clientState.results.push_back(taggedJob);
	}
	clientState.remaining--;
	clientState.isWaiting = false;
	// pthread_cond_signal(&clientState.cv);
      }
      servantState.jobQ.pop_front();
      scheduler.remaining--;
      servantState.isBusy = false;
      nProcessed++;
    }
  }
  return nProcessed;
}
int
Scheduler::Channel::
throwJobs()
{
  int nProcessed = 0;
  while (!scheduler.jobQ.empty()) {
    const TaggedJob& taggedJob = scheduler.jobQ.front();
    if (taggedJob.servantID == TaggedJob::ANY_SERVANT) break;

    ServantState& servantState = 
      scheduler.servantStates[taggedJob.servantID];
    servantState.jobQ.push_back(taggedJob);
    scheduler.jobQ.pop_front();
    nProcessed++;
  }
  for (int i = 0; i < (int)scheduler.servantStates.size(); i++) {
    ServantState& servantState = scheduler.servantStates[i];
    if (servantState.isBusy) continue;

    if (servantState.jobQ.empty() && !scheduler.jobQ.empty()) {
      TaggedJob& taggedJob = scheduler.jobQ.front();
      if (taggedJob.servantID == TaggedJob::ANY_SERVANT) {
	taggedJob.servantID = i;
	servantState.jobQ.push_back(taggedJob);
	scheduler.jobQ.pop_front();
	nProcessed++;
      }
    }
    if (!servantState.jobQ.empty()) {
      const TaggedJob& taggedJob = servantState.jobQ.front();
      sendRequest(i, taggedJob.job);
      servantState.isBusy = true;
      nProcessed++;
    }
  }
  return nProcessed;
}
bool
Scheduler::Channel::
servantsAreAllBusy() const
{
  if (scheduler.servantStates.empty()) return false;

  for (int i = 0; i < (int) scheduler.servantStates.size(); i++) {
    if (!scheduler.servantStates[i].isBusy) return false;

  }
  return true;
}
bool
Scheduler::Channel::
clientsAreAllWaiting() const
{
  if(scheduler.clientStates.empty()) return false;

  for (ClientStates::iterator ci = scheduler.clientStates.begin();
       ci != scheduler.clientStates.end(); ci++) {
    if (!ci->second.isWaiting) return false;

  }
  return true;
}
void
Scheduler::Channel::
waitAnyServant()
{
  if (scheduler.isMultiProcess) {
    UNLOCK();
    Die();
    // MPIHandle::waitAny();
    LOCK();
  } else {
    while (!scheduler.servant.jobIsCompleted) {
      COND_WAIT(cv);
    }
  }
}
void
Scheduler::Channel::
waitAnyClient()
{
  COND_WAIT(cv);
}
void
Scheduler::Channel::
timedWaitAnyClient()
{
  TIMED_WAIT();
}
void
Scheduler::Channel::
sendRequest(int servantID, const Job& job)
{
  if (scheduler.isMultiProcess) {
    Die();
    // MPIHandle::sendRequest(servantID, job);
  } else {
    Assert(scheduler.servant.jobIsCompleted);
    if (job.sendAll()) {// dispose sendAll job
      dmsg("signal sendRequest send_all job");
    } else {
      dmsg("signal sendRequest send_any job");
      //when this signal is executed?
      // -> waitServants/Clients, collectResults
      // pthread_cond_signal(&scheduler.servant.cv);
      scheduler.servant.setJob(job);
      scheduler.servant.jobIsCompleted = false;
    }
  }
}
bool
Scheduler::Channel::
resultIsArrived(int servantID) const
{
  if (scheduler.isMultiProcess) {
    Die();
    //return MPIHandle::resultIsArrived(servantID);
    return false;
  } else {
    dmsg2("resultisArrived=", scheduler.servant.jobIsCompleted);
    return scheduler.servant.jobIsCompleted;
  }
}
Job
Scheduler::Channel::
recvResult(int servantID)
{
  Job job;
  if (scheduler.isMultiProcess) {
    Die();
    // MPIHandle::recvResult(servantID, job);
    return Job();
  } else {
    job = scheduler.servant.getJob();
  }
  return job;
}
void*
Scheduler_Servant_run(void* pScheduler)
{
  Scheduler& scheduler = *(Scheduler*)pScheduler;
  Scheduler::Servant& servant = scheduler.servant;
  servant.run();
  return NULL;
}
Scheduler::Servant::
Servant(Scheduler& scheduler, const JobRunner& jobRunner)
  : scheduler(scheduler),
    jobRunner(jobRunner),
    // thread(),
    jobIsCompleted(true)
{
  // pthread_cond_init(&cv, NULL);
}
Scheduler::Servant::
~Servant()
{
  // pthread_cond_destroy(&cv);
}
string
Scheduler::Servant::
toString() const
{
  ostringstream oss;
  return oss.str();
}
void
Scheduler::Servant::
run()
{
  bool stop = false;
  while (!stop) {
    recvRequest();
    if (jobRunner.hasJob()) {
      jobRunner.run();
    } else {
      stop = true;
    }
    sendResult();
  }
  dmsg("exiting");
}
void
Scheduler::Servant::
recvRequest()
{
  if (scheduler.isMultiProcess) {
    dmsg("waiting");
    Job job;
    Die();
    //MPIHandle::recvRequest(job);
    jobRunner.setJob(job);
    dmsg("return");
  } else {
    LOCK();
    while (jobIsCompleted) {
      COND_WAIT(cv);
    }
    UNLOCK();
  }
}
void
Scheduler::Servant::
sendResult()
{
  if (scheduler.isMultiProcess) {
    Die();
    //MPIHandle::sendResult(jobRunner.getJob());
  } else {
    LOCK();
    jobIsCompleted = true;
    // pthread_cond_signal(&scheduler.channel.cv);
    UNLOCK();
  }
}
void
Scheduler::Servant::
executeCommand()
{
  jobRunner.run();
}
void
Scheduler::Servant::
setJob(const Job& job)
{
  jobRunner.setJob(job);
}
Job
Scheduler::Servant::
getJob() const
{
  return jobRunner.getJob();
}

Scheduler::
Scheduler(const JobRunner& jobRunner)
  : jobCounter(0),
    remaining(0),
    clientStates(),
    servantStates(),
    channel(*this),
    servant(*this, jobRunner)
{
  // MPIHandle::getSize() returns correct value
  // even in the absence of MPI
  //isMultiProcess = (MPIHandle::getSize() > 1); 
  isMultiProcess = false;

  servantStates.resize(getNServant());

#ifdef SEQUENTIAL
  return;
#endif

  // if (!isRoot()) {
  //   servant.run();
  //   return;
  // }
  // if (!isMultiProcess) {
  //   int err = pthread_create(&servant.thread, NULL,
  // 			     Scheduler_Servant_run, (void*)this);
  //   if (err) Die("cannot create Servant");
  // }
  // int err = pthread_create(&channel.thread, NULL,
  // 			   Scheduler_Channel_run, (void*)this);
  // if (err) Die("cannot create Channel");
}
Scheduler::
~Scheduler()
{
  if (!isRoot()) return;

  if (!clientStates.empty()) {
    for (ClientStates::const_iterator i = clientStates.begin();
	 i != clientStates.end(); i++) {
      Warn("client %d not removed", i->first);
      finish(i->first);
      removeClient(i->first);
    }
  }

#ifdef SEQUENTIAL
  return;
#endif

   // terminate();
   // if (!isMultiProcess) {
   //   pthread_join(servant.thread, NULL);
   // }
   // pthread_join(channel.thread, NULL);
}
string
Scheduler::
toString() const
{
  ostringstream oss;
  oss << "--- \n";
  oss << "jobQ.size(): " << jobQ.size() << "\n";
  oss << "remaining: " << remaining << "\n";
  oss << "jobCounter: " << jobCounter << "\n";
  oss << "channel: \n";
  oss << "  " << channel.toString() << "\n";
  oss << "clientStates \n";
  for (ClientStates::const_iterator i = clientStates.begin();
       i != clientStates.end(); i++) {
    oss << "  " << i->first << ": " << i->second.toString() << "\n";
  }
  oss << "servantStates \n";
  for (int i = 0; i < (int) servantStates.size(); i++) {
    oss << "  " << i << ": " << servantStates[i].toString() << "\n";
  }
  oss << "servant: " << servant.toString() << "\n";
  return oss.str();
}
int
Scheduler::
createClient()
{
  LOCK();
  int clientID = 0;
  if (!clientStates.empty()) {
    int maxID = clientStates.rbegin()->first;
    clientID = maxID + 1;
  }
  clientStates[clientID];
  // pthread_cond_signal(&channel.cv);
  UNLOCK();
  dmsg2("created client=", clientID);
  return clientID;
}
void
Scheduler::
removeClient(int clientID)
{
  LOCK();

#ifdef DEBUG
  cout << "client" << clientID << ": Lock removeClient" << endl;
  Assert(clientStates.find(clientID) != clientStates.end(),
	 "no client %d", clientID);
  ClientState& clientState = clientStates[clientID];
  Assert(clientState.remaining == 0, 
	 "client %d has %d remained jobs",
	 clientID, clientState.remaining);
#endif

  clientStates.erase(clientID);
  // pthread_cond_signal(&channel.cv);
  UNLOCK();
}

int
Scheduler::
addJob(const int& clientID, const Job& job)
{
  dmsg2("client=", clientID);
  LOCK();

  ClientState& clientState = clientStates[clientID];
  int retval = (-1);

  if (job.sendAll()) {
#ifdef SEQUENTIAL
    // do nothing
#else
    TaggedJob taggedJob(job, -1, clientID, TaggedJob::ANY_SERVANT);
    for (int i = 0; i < (int)servantStates.size(); i++) {
      taggedJob.id = jobCounter++;
      taggedJob.servantID = i;
      jobQ.push_back(taggedJob);
      clientState.remaining++;
      remaining++;
    }
#endif
  } else {// SEND_ANY
    TaggedJob taggedJob(job, jobCounter++, clientID,
			TaggedJob::ANY_SERVANT);
    retval = taggedJob.id;
#ifdef SEQUENTIAL
    servant.setJob(job);
    servant.executeCommand();
    taggedJob.job = servant.getJob();
    clientState.results.push_back(taggedJob);
#else
    jobQ.push_back(taggedJob);
    clientState.remaining++;
    remaining++;
#endif
  }
  dmsg2("client=", clientID);
  // pthread_cond_signal(&channel.cv);
  UNLOCK();
  return retval;
}
int
Scheduler::
getResult(const int& clientID, Job& job)
{
  dmsg2("client=", clientID);
  LOCK();
  ClientState& clientState = clientStates[clientID];
  while (clientState.results.empty()) {
#ifdef SEQUENTIAL
    Die("no results for client %d", clientID);
#endif
    Assert(clientState.remaining > 0,
	   "client%d: remaining=%d", clientID, clientState.remaining);
    clientState.isWaiting = true;
    //pthread_cond_signal(&channel.cv);
    COND_WAIT(clientState.cv);
  }
  TaggedJob& taggedJob = clientState.results.front();
  const int retval = taggedJob.id;
  job = taggedJob.job;
  clientState.results.pop_front();

  dmsg2("client=", clientID);
  // pthread_cond_signal(&channel.cv);
  UNLOCK();
  return retval;
}
void
Scheduler::
finish(int clientID)
{
#ifdef SEQUENTIAL
  return;
#endif

  // dmsg2("client=", clientID);
  // LOCK();
  // ClientState& clientState = clientStates[clientID];
  // Assert(clientState.results.empty());
  // while (clientState.remaining != 0) {
  //   Assert(clientState.remaining > 0);
  //   clientState.isWaiting = true;
  //   pthread_cond_signal(&channel.cv);
  //   COND_WAIT(clientState.cv);
  // }
  // dmsg2("client=", clientID);
  // pthread_cond_signal(&channel.cv);
  // UNLOCK();
}
int
Scheduler::
getRemaining(int clientID) const
{
  LOCK();
  Assert(clientStates.find(clientID) != clientStates.end());
  int val = clientStates.find(clientID)->second.remaining;
  UNLOCK();
  return val;
}
void
Scheduler::
terminate()
{
  LOCK();
  Job job;
  job.set_Terminate();
  for (int i = 0; i < (int)servantStates.size(); i++) {
    TaggedJob taggedJob(job, jobCounter++,
			TaggedJob::CHANNEL_CLIENT, i);
    jobQ.push_back(taggedJob);
    remaining++;
  }
  channel.shutdownRequested = true;
  // pthread_cond_signal(&channel.cv);
  UNLOCK();
}
bool
Scheduler::
isRoot() const
{
  //return MPIHandle::isRoot();
  return true;
}
int
Scheduler::
getNServant() const
{
  // return MPIHandle::getNServant();
  return 1;
}

#undef dmsg
#undef dmsg2
#undef LOCK
#undef UNLOCK
#undef COND_WAIT
#undef TIMED_WAIT
