/*
 * SeqData.hpp
 *
 */
#ifndef SEQ_DATA_HPP
#define SEQ_DATA_HPP
#include "Util.hpp"
#include "EScript.hpp"

class SeqData;
class MSA;

class Seq : private Vector<char> {
  typedef Vector<char> Super;
public:
  typedef int Id;

  Seq(const SeqData& seqData,
      const string& name = "", const string& str = "", const string& qname = "");
  Seq(const Seq& seq);
  Seq(const SeqData& seqData, const Seq& seq);
  ~Seq();
  Seq& operator=(const Seq& seq);
  using Super::size;
  using Super::empty;
  using Super::iterator;
  using Super::const_iterator;
  using Super::begin;
  using Super::end;
  using Super::operator[];
  const char& front() const {return Super::front();}
  const char& back() const {return Super::back();}

  void removeAt(int pos, int len = 1) {
    Assert(0 <= pos && pos < (int)size() && len >= 0);
    erase(Super::begin() + pos, Super::begin() + pos + len);
  }
  void assignSeqContent(const Seq& seq);
  const string& getName() const {return name;}
  const string& getQName() const {return qname;}
  string getSequenceString() const;
  string getString() const;
  Id getId() const;
  int getBaseLength() const;
  float computeSeqIdentity(const Seq& seq) const;
  float computeTupleCountSimilarity(const Seq& seq) const;
  PosMap getPosMap() const;
  string toString() const;
  void print() const;
private:
  const SeqData& seqData;
  string name;
  string qname;
  
  Seq();
  void clear();
  void concat(const string& str);
  bool readFasta(istream& fi, const string& famName = "");
  void writeFasta(ostream& fo) const;
  friend class SeqData;
  friend class MSA;
};

typedef Seq::Id SeqId;

//typedef Vector<SeqId> SeqIds;

class SeqIds : public IntPairs {
  typedef IntPairs Super;
public:
  SeqIds() : Super() {}
  SeqIds(const SeqIds& seqIds) : Super(seqIds) {}
  SeqIds(const Vector<int>& v) : Super(v) {}
  //SeqIds(int n) : Super(n) {}
  //SeqIds(int n, const SeqId& seqId) : Super(n, seqId) {}
  template <typename InpIt>
  SeqIds(InpIt first, InpIt last) : Super(first, last) {}
  ~SeqIds() {}
  int getIndex(const SeqId& id) const {
    return Super::indexOf(id);
  }
  bool operator<(const SeqIds& seqIds) const {
    return Super::operator<(seqIds);
  }
};

class SeqData : private Vector<Seq> {
  typedef Vector<Seq> Super;
public:
  SeqData();
  SeqData(const SeqData& seqData);
  SeqData(const string& seqFile, const string& seqFileType);
  ~SeqData();
  SeqData& operator=(const SeqData& seqData);
  using Super::size;
  using Super::empty;
    
  void clear();
  void addSeq(const Seq& seq);
  void addSeq(const string& name0, const string& str, const string& qname0);
  Seq& getSeq(const string& qname0);
  const Seq& getSeq(const string& qname0) const;
  Seq& getSeq(SeqId id0);
  const Seq& getSeq(SeqId id0) const;
  const string& getName(SeqId id0) const;
  const string& getQName(SeqId id0) const;
  SeqId getSeqId(const string& qname0) const;
  SeqId getSeqId(const string& famName0, const string& name0) const;
  SeqId getSeqId(const Seq& seq) const;
  SeqIds getSeqIds() const;
  string getAC(const SeqIds& seqIds = SeqIds()) const {
    const int& index = getFamilyIndex(seqIds);
    return (index >= 0 ? vFamilyAnnot[index].AC : "");
  }
  string getID(const SeqIds& seqIds = SeqIds()) const {
    const int& index = getFamilyIndex(seqIds);
    return (index >= 0 ? vFamilyAnnot[index].ID : "");
  }
  string getSSCons(const SeqIds& seqIds = SeqIds()) const {
    const int& index = getFamilyIndex(seqIds);
    return (index >= 0 ? vFamilyAnnot[index].sscons : "");
  }
  void setSSCons(const SeqIds& seqIds, const string& sscons) {
    const int& index = getFamilyIndex(seqIds);
    if (index < 0) Die("cannot set sscons");

    vFamilyAnnot[index].sscons = sscons;
  }
  const Seq& operator[](int index) const {return Super::get(index);}
  bool hasSeqId(SeqId id0) const;
  bool nColsAreEqual() const;
  bool nColsAreEqual(SeqId id_beg, SeqId id_end) const;
  bool nColsAreEqual(const SeqIds& seqIds) const;
  void readFasta(istream& fi, const string& fname = "");
  void writeFasta(ostream& fo) const;
  void writeFasta(ostream& fo, const SeqIds& seqIds, const string& sscons) const;
  void writeFastaAt(ostream& fo, Seq::Id id) const ;
  bool readStockholmEntry(istream& fi, const string& fname = "");
  void readStockholm(istream& fi, const string& fname = "");
  void writeStockholm(ostream& fo, const SeqIds& seqIds,
		      bool withSSCons = false) const;
  void writeStockholm(ostream& fo, const SeqIds& seqIds,
		      const string& sscons) const;
  void writeStockholm(ostream& fo, const SeqIds& seqIds,
		      const string& sscons, const float& score) const;
  void writeStockholm(ostream& fo, const SeqIds& seqIds,
		      const string& sscons, const float& score, const bool& printScore) const;
  void writeClustal(ostream& fo, const SeqIds& seqIds) const;
  string toStockholmString(const SeqIds& seqIds) const {
    ostringstream oss;
    writeStockholm(oss, seqIds);
    return oss.str();
  }
  string toString() const;
  void print() const;
  float computeMeanSeqLength(const SeqIds& seqIds) const {
    float sum = 0.0;
    for (int i = 0; i < (int) seqIds.size(); i++) {
      sum += get(seqIds[i]).getBaseLength();
    }
    return sum / (float) seqIds.size();
  }

  class FamilyIterator {
  public:
    FamilyIterator(SeqData& seqData) : seqData(seqData), familyIndex(0) {}
    ~FamilyIterator() {}
    void reset() {familyIndex = 0;}
    bool getNext(SeqIds& seqIds) {
      const Vector<FamilyAnnot>& vfa = seqData.vFamilyAnnot;
      if (familyIndex >= (int)vfa.size()) return false;

      const FamilyAnnot& fa = vfa[familyIndex];
      Vector<int> v(fa.index_end - fa.index_begin);
      generate(v.begin(), v.end(), Series<int>(fa.index_begin));
      seqIds.assign(v.begin(), v.end());
      familyIndex++;
      return true;
    }
 
  private:
    SeqData& seqData;
    int familyIndex;
    FamilyIterator();
    FamilyIterator(const FamilyIterator& familyIterator);
    FamilyIterator& operator=(const FamilyIterator& familyIterator);
  };
  
private:
  typedef map<string, int> NameTable;
  struct FamilyAnnot {
    string AC;
    string ID;
    string sscons;
    int index_begin;
    int index_end;

    FamilyAnnot() : AC(), ID(), sscons(), index_begin(0), index_end(0) {}
    ~FamilyAnnot() {}
  };

  Vector<FamilyAnnot> vFamilyAnnot;
  NameTable nameTable;

  int getFamilyIndex(const SeqIds& seqIds) const {
    if (!seqIds.empty() && !vFamilyAnnot.empty()) {
      Seq::Id minId = *min_element(seqIds.begin(), seqIds.end());
      Seq::Id maxId = *max_element(seqIds.begin(), seqIds.end());
      for (int i = 0; i < (int) vFamilyAnnot.size(); i++) {
	if (vFamilyAnnot[i].index_begin <= minId
	    && maxId < vFamilyAnnot[i].index_end) {
	  return i;
	}
      }
    }
    return (-1);
  }
  void resetNameTable();
  SeqId getSeqId0(const string& qname) const;
  friend class MSA;
};

#endif
