/*
 * StripRegion.cpp
 *
 */
#include "StripRegion.hpp"


void
Trapezoid::
print() const
{
  cout << toString() << flush;
}

StripRegion::
StripRegion()
  : Super(), rect0(Rect::null())
{
}
StripRegion::
StripRegion(const StripRegion& sr)
  : Super(sr), rect0(sr.rect0)
{
}
StripRegion::
~StripRegion()
{
}
StripRegion&
StripRegion::
operator=(const StripRegion& sr)
{
  if (this == &sr) return *this;

  Super::operator=(sr);
  rect0 = sr.rect0;
  return *this;
}
bool
StripRegion::
operator==(const StripRegion& sr) const
{
  return (rect0 == sr.rect0
	  && equal(Super::begin(), Super::end(), sr.begin()));
}
void
StripRegion::
clear()
{
  Super::clear();
  rect0.setNull();
}

string
StripRegion::
toString() const
{
  ostringstream oss;
  oss << "--- \n";
  oss << "rect0: " << rect0.toString() << "\n";
  for (int i = 0; i < (int) size(); i++) {
    oss << get(i).toString() << "\n";
  }
  return oss.str();
}
void
StripRegion::
print() const
{
  cout << toString() << flush;
}
void
StripRegion::
draw(ostream& fo) const
{
  int lx = rect0.lengthX();
  int ly = rect0.lengthY();
  int n = (lx + 2) * (ly + 3);
  int m = ly + 3;
  Vector<char> disp(n + 1);
  disp.fill(' ');
  disp[n] = '\0';
  int x0 = 1 - bx0;
  int y0 = 1 - by0;
  
  disp[0] = '+';
  disp[ly + 1] = '+';
  disp[(lx + 1) * m] = '+';
  disp[(lx + 1) * m + ly + 1] = '+';
  disp[0 * m + ly + 2] = '\n';
  disp[(lx + 1) * m + ly + 2] = '\n';
  for (int x = bx0; x <= ex0; x++) {
    disp[(x0 + x) * m] = '|';
    disp[(x0 + x) * m + ly + 1] = '|';
    disp[(x0 + x) * m + ly + 2] = '\n';
  }

  for (int y = by0; y <= ey0; y++) {
    disp[0 * m + y0 + y] = '-';
    disp[(lx + 1) * m + y0 + y] = '-';
  }
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    for (int y = t.exby; y <= t.ey; y++) {
      disp[(x0+t.ex)*m + (y0+y)] = '~';
    }
    for (int y = t.by; y <= t.bxey; y++) {
      disp[(x0+t.bx)*m + (y0+y)] = '_';
    }
    for (int x = t.bx + 1; x < t.ex; x++) {
      int y_beg = getBY(x);
      int y_last = getEY(x);
      disp[(x0+x) * m + (y0 + y_beg)] = (t.isLD ? '\\' : '|');
      disp[(x0+x) * m + (y0 + y_last)] = (t.isRD ? '\\' : '|');
      for (int y = y_beg + 1; y < y_last; y++) {
	disp[(x0+x) * m + (y0 + y)] = 'X';
      }
    }
  }
  fo << &disp.front() << flush;
}


void
StripRegion::
resetRect0()
{
  if (empty()) {
    rect0.setNull();
  } else {
    rect0.setVal(Super::front().b, Super::back().e);
  }
}
void
StripRegion::
push(const Trapezoid& t)
{
  Super::push_back(t);
  resetRect0();
  testConsistency();
}
bool
StripRegion::
includes(const Point& pt) const
{
  if (!pt.isNormal()) return false;
  if (pt.x < bx0 || ex0 < pt.x) return false;
  if (pt.y < by0 || ey0 < pt.y) return false;

  for (int i = 0; i < (int) size(); i++) {
    if (get(i).ex < pt.x) continue;

    return get(i).includes(pt);
  }
  Assert(false, "never reach here rect0=(%d,%d,%d,%d), (x,y)=(%d,%d)", 
	 bx0, by0, ex0, ey0, pt.x, pt.y);
  return false;
}
bool
StripRegion::
includes(const Rect& rect) const
{
  return (includes(rect.b) 
	  && includes(rect.e));
}
int
StripRegion::
getBX(int y) const
{
  return get(getBXTIndex(y)).getBX(y);
}
int
StripRegion::
getEX(int y) const
{
  return get(getEXTIndex(y)).getEX(y);
}
int
StripRegion::
getXTIndex(int x) const
{
  if (x < bx0 || ex0 < x) {
    Die("out of region");
    return (-1);
  }

  for (int i = 0; i < (int) size(); i++) {
    if (x <= get(i).ex) return i;
  }
  Die("never reach here");
  return (-1);
}
int
StripRegion::
getBXTIndex(int y) const
{
  if (y < by0 || ey0 < y) {
    Die("out of region");
    return (-1);
  }
  for (int i = 0; i < (int) size(); i++) {
    if (get(i).ey < y) continue;

    return i;
  }
  Die("never reach here");
  return (-1);
}
int
StripRegion::
getEXTIndex(int y) const
{
  if (y < by0 || ey0 < y) {
    Die("out of region");
    return (-1);
  }
  for (int i = (int) size() - 1; i >= 0; i--) {
    if (y < get(i).by) continue;

    return i;
  }
  Die("never reach here");
  return (-1);
}
int
StripRegion::
getBXEY() const
{
  if (empty()) return (-1);

  return get(0).bxey;
}
int
StripRegion::
getEXBY() const
{
  if (empty()) return (-1);

  return get(size() - 1).exby;
}
int
StripRegion::
getArea() const
{
  int area = 0;
  for (int i = 0; i < (int) size(); i++) {
    area += get(i).getArea();
  }
  return area;
}
int
StripRegion::
getMaxDiag() const
{
  int maxDiag = -1; //returns -1 if empty or no non-negative diagonals
  for (int i = 0; i < (int)size(); i++) {
    const Trapezoid& t = get(i);
    maxDiag = max(maxDiag, t.getMaxDiag());
  }
  return maxDiag;
}
int
StripRegion::
getMinDiag() const
{
  int minDiag = 1; //returns +1 if empty or no non-positive diagonals
    for (int i = 0; i < (int)size(); i++) {
    const Trapezoid& t = get(i);
    minDiag = min(minDiag, t.getMinDiag());
  }
  return minDiag;
}
int
StripRegion::
getDiagWidth() const {
  return (getMaxDiag() - getMinDiag() + 1);
}
int
StripRegion::
getAreaBefore(const Point& pt) const
{//excludes (x, y)
  int area = 0;
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    if (pt.x < t.bx) break;

    if (pt.x <= t.ex) {
      area += get(i).getAreaBefore(pt);
      break;
    } else {
      area += get(i).getArea();
    }
  }
  return area;
}
int
StripRegion::
getAreaBeforeT(const Point& pt) const
{
  Assert(includes(pt));
  return (getAreaLeft(pt.y - 1) + (pt.x - getBX(pt.y)));
}
int
StripRegion::
getAreaLeft(int y) const
{//includes column y
  int area = 0;
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    if (y < t.by) break;

    area += t.getAreaLeft(y);
  }
  return area;
}
int
StripRegion::
getAreaTopLeft(const Point& pt) const
{//includes row x , column y
  int area = 0;
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    if (pt.x < t.bx || pt.y < t.by) break;

    if (pt.x <= t.ex) {
      area += t.getAreaTopLeft(pt);
      break;
    } else if (pt.y <= t.ey) {
      area += t.getAreaLeft(pt.y);
    } else {
      area += t.getArea();
    }
  }
  return area;
}
int
StripRegion::
getAreaBottomRight(const Point& pt) const
{//includes row x , column y
  int area = 0;
  for (int i = (int) size() - 1; i >= 0; i--) {
    const Trapezoid& t = get(i);
    if (t.ex < pt.x || t.ey < pt.y) break;

    if (t.bx <= pt.x) {
      area += t.getAreaBottomRight(pt);
      break;
    } else if (t.by <= pt.y) {
      area += get(i).getAreaRight(pt.y);
    } else {
      area += t.getArea();
    } 
  }
  return area;
}
//bool
//StripRegion::
//hasLockedPath() const
//{
//  const Rect rect = {{bx0 + 1, by0 + 1}, {ex0 - 1, ey0 - 1}};
//  if (!includes(rect)) return false;
//  if (rect.b.equals(rect.e)) return true;
//  if (rect.bx == rect.ex || rect.by == rect.ey) return false;
//
//  return true;
//}
//StripRegion
//StripRegion::
//getInsidePart(const Rect& rect) const
//{//includes rect
//  Check(includes(rect));
//  StripRegion dpRegion0 = getInsidePartX(rect.bx, rect.ex);
//  const int lenX = rect.ex - rect.bx + 1;
//  Vector<int> y_begs;
//  Vector<int> y_lasts;
//  dpRegion0.getYRange(y_begs, y_lasts);
//  for (int i = 0; i < lenX; i++) {
//    y_begs[i] = max(y_begs[i], rect.by);
//    y_lasts[i] = min(y_lasts[i], rect.ey);
//  }
//  dpRegion0.setFromYRange(rect, y_begs, y_lasts);
//  return dpRegion0;
//}
void
StripRegion::
setFromYRange(const Rect& rect00,
	      const Vector<int>& y_begs,
	      const Vector<int>& y_lasts)
{
  Assert(y_begs.size() > 0 && y_begs.size() == y_lasts.size());
  Assert((rect00.ex - rect00.bx + 1) == (int)y_begs.size());
  clear();

  int NULL_TYPE = -1;
  int y_beg = y_begs[0];
  int y_last = y_lasts[0];
  int type = NULL_TYPE;
  Rect rect = {{{rect00.bx, y_beg}}, {{rect00.bx, y_last}}};
  for (int x = rect00.bx; x <= rect00.ex;) {
    if (x == rect00.ex) {
      rect.ex = x;
      rect.ey = y_last;
      push(Trapezoid(rect, type));
      break;
    }
    int y_beg1 = y_begs[(x - rect00.bx) + 1];
    int y_last1 = y_lasts[(x - rect00.bx) + 1];
    int dy_beg = y_beg1 - y_beg;
    int dy_last = y_last1 - y_last;

    if (x == rect.bx) {
      int flag_ld = -1;
      int flag_rd = -1;
      switch (dy_beg) {
      case 0: flag_ld = 0; break;
      case 1: flag_ld = Trapezoid::FLAG_LD; break;
      }
      switch (dy_last) {
      case 0: flag_rd = 0; break;
      case 1: flag_rd = Trapezoid::FLAG_RD; break;
      }
      if (flag_ld != -1 && flag_rd != -1) {
	y_beg = y_beg1;
	y_last = y_last1;
	x++;
	type = flag_ld | flag_rd;
      } else {
	rect.ex = x;
	rect.ey = y_last;
	push(Trapezoid(rect, Trapezoid::LV_RV));
	y_beg = y_beg1;
	y_last = y_last1;
	x++;
	type = NULL_TYPE;
	rect.bx = x;
	rect.by = y_beg;
      }
    } else {
      bool left_continues =
	(type & Trapezoid::FLAG_LD) ? (dy_beg == 1) : (dy_beg == 0);
      bool right_continues =
	(type & Trapezoid::FLAG_RD) ? (dy_last == 1) : (dy_last == 0);

      if (left_continues && right_continues) {
	y_beg = y_beg1;
	y_last = y_last1;
	x++;
      } else {
	rect.ex = x;
	rect.ey = y_last;
	push(Trapezoid(rect, type));
	y_beg = y_beg1;
	y_last = y_last1;
	x++;
	type = NULL_TYPE;
	rect.bx = x;
	rect.by = y_beg;
      }
    }
  }
  testConsistency();
}
void
StripRegion::
getYRange(Vector<int>& y_begs, Vector<int>& y_lasts) const
{
  y_begs.clear();
  y_lasts.clear();
  for (int i = 0; i < (int)size(); i++) {
    const Trapezoid& t = get(i);
    for (int x = t.bx; x <= t.ex; x++) {
      y_begs.push_back(t.getBY(x));
      y_lasts.push_back(t.getEY(x));
    }
  }
}
StripRegion
StripRegion::
getBoundedRegion(const Rect& bbox) const
{
  if (bbox.includes(rect0)) return *this;
  Vector<int> y_begs;
  Vector<int> y_lasts;
  for (int i = 0; i < (int)size(); i++) {
    const Trapezoid& t = get(i);
    if (bbox.ex < t.bx) break;
    if (t.ex < bbox.bx) continue;

    const int& x_beg = max(bbox.bx, t.bx);
    const int& x_last = min(bbox.ex, t.ex);
    for (int x = x_beg; x <= x_last; x++) {
      y_begs.push_back(max(bbox.by, t.getBY(x)));
      y_lasts.push_back(min(bbox.ey, t.getEY(x)));
    }
  }
  if (y_begs.empty()) return StripRegion();

  int bi = 0;
  for (; bi < (int)y_begs.size(); bi++) {
    if (y_begs[bi] <= y_lasts[bi]) break;
  }
  if (bi == (int)y_begs.size()) return StripRegion();

  int ei = (int) y_begs.size() - 1;
  for (; ei >= 0; ei--) {
    if (y_begs[ei] <= y_lasts[ei]) break;
  }
  y_begs.erase(y_begs.begin() + ei + 1, y_begs.end());
  y_lasts.erase(y_lasts.begin() + ei + 1, y_lasts.end());
  y_begs.erase(y_begs.begin(), y_begs.begin() + bi);
  y_lasts.erase(y_lasts.begin(), y_lasts.begin() + bi);
  const int& x0 = max(rect0.bx, bbox.bx) + bi;
  Rect rect = {
    {{x0, y_begs.front()}},
    {{x0 + y_begs.size() - 1, y_lasts.back()}}
  };
  StripRegion sr;
  sr.setFromYRange(rect, y_begs, y_lasts);
  sr.testConsistency();
  return sr;
}
StripRegion
StripRegion::
getBoundingRegion(const StripRegion& sr) const
{
  if (!(bx0 <= (sr.ex0 + 1) && sr.bx0 <= (ex0 + 1))) {
    Die("at present x ranges should not have a gap");
  }
  const Rect rect = {
    {{min(bx0, sr.bx0), min(by0, sr.by0)}},
    {{max(ex0, sr.ex0), max(ey0, sr.ey0)}}
  };
  const int& lx = rect.ex - rect.bx + 1;
  Vector<int> y_begs(lx, (-1));
  Vector<int> y_lasts(lx, (-1));

  const int& offset1 = (bx0 - rect.bx);
  Vector<int> y_begs1;
  Vector<int> y_lasts1;
  getYRange(y_begs1, y_lasts1);
  for (int i1 = 0; i1 < (int)y_begs1.size(); i1++) {
    const int& i = (i1 + offset1);
    y_begs[i] = y_begs1[i1];
    y_lasts[i] = y_lasts1[i1];
  }
  const int& offset2 = (sr.bx0 - rect.bx);
  Vector<int> y_begs2;
  Vector<int> y_lasts2;
  sr.getYRange(y_begs2, y_lasts2);
  for (int i2 = 0; i2 < (int)y_begs2.size(); i2++) {
    const int& i = (i2 + offset2);
    if (y_begs[i] < 0 || y_begs2[i2] < y_begs[i]) {
      y_begs[i] = y_begs2[i2];
    }
    if (y_lasts[i] < 0 || y_begs[i] < y_lasts2[i2]) {
      y_lasts[i] = y_lasts2[i2];
    }
  }
  Assert(find(y_begs.begin(), y_begs.end(), (int)(-1)) == y_begs.end());
  Assert(find(y_lasts.begin(), y_lasts.end(), (int)(-1))==y_lasts.end());

  for (int i = (int)y_begs.size() - 2; i >= 0; i--) {
    if (y_begs[i + 1] < y_begs[i]) {
      y_begs[i] = y_begs[i + 1];
    }
  }
  for (int i = 1; i < (int)y_lasts.size(); i++) {
    if (y_lasts[i] < y_lasts[i - 1]) {
      y_lasts[i] = y_lasts[i - 1];
    }
  }
  StripRegion sr1;
  sr1.setFromYRange(rect, y_begs, y_lasts);
  return sr1;
}

// margin methods
int
StripRegion::
getAreaMargin() const
{
  const int& leftMargin0 = 1;
  const int& rightMargin0 = 1;
  const int& topLeftCorner = 1;
  const int& bottomRightCorner = 1;
  int area = getArea();
  area += (leftMargin0 + rightMargin0) * (ex0 - bx0 + 1);
  for (int i = 0; i < (int)size(); i++) {
    const Trapezoid& t = get(i);
    if (i == 0) {
      area += (topLeftCorner + (t.bxey - t.by + 1) + rightMargin0);
    } else {
      const Trapezoid& t1 = get(i - 1);
      area += (t.bxey - t1.ey);
      area += (t.by - t1.exby);
    }
    if (i == (int) size() - 1) {
      area += (leftMargin0 + (t.ey - t.exby + 1) + bottomRightCorner);
    }
  }
  return area;
}
int
StripRegion::
getAreaMarginBefore(const Point& pt) const
{ //excludes (x, y)
  const int& leftMargin0 = 1;
  const int& rightMargin0 = 1;
  const int& topLeftCorner = 1;
  int area = getAreaBefore(pt);
  area += (leftMargin0 + rightMargin0) * (pt.x - bx0) + leftMargin0;
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    if (i == 0) {
      area += (topLeftCorner + (t.bxey - t.by + 1) + rightMargin0);
    } else {
      const Trapezoid& t1 = get(i - 1);
      area += (t.bxey - t1.ey);
      area += (t.by - t1.exby);
    }
    if (pt.x <= t.ex) break;

  }
  return area;
}
int
StripRegion::
getTopLeftMargin(const int& i) const
{
  const Trapezoid& t = get(i);
  int rightMargin = 1;
  int leftMargin = 1;
  if (i == 0) {
    rightMargin += (t.bxey - t.by + 1);
  } else {
    const Trapezoid& t1 = get(i - 1);
    rightMargin += (t.bxey - t1.ey);
    leftMargin += (t.by - t1.exby);
  }
  return (rightMargin + leftMargin) ;
}
int
StripRegion::
getBottomRightMargin(const int& i) const
{
  const Trapezoid& t = get(i);
  int rightMargin = 1;
  int leftMargin = 1;
  if (i == (int) size() - 1) {
    leftMargin += (t.ey - t.exby + 1);
  } else {
    const Trapezoid& t1 = get(i + 1);
    rightMargin += (t1.bxey - t.ey);
    leftMargin += (t1.by - t.exby);
  }
  return (rightMargin + leftMargin);
}

void
StripRegion::
testConsistency() const
{
#ifdef DEBUG
  Assert(0 <= bx0 && bx0 <= ex0);
  Assert(0 <= by0 && by0 <= ey0);
  if (!empty()) {
    const Trapezoid& t0 = get(0);
    const Trapezoid& t1 = get(size() - 1);
    Assert(t0.bx == bx0 && t0.by == by0);
    Assert(t1.ex == ex0 && t1.ey == ey0);
  }
  for (int i = 0; i < (int) size(); i++) {
    const Trapezoid& t = get(i);
    t.testConsistency();
    
    if (i > 0) {
      const Trapezoid& t1 = get(i - 1);
      Assert(t1.ex + 1 == t.bx);
      Assert(t1.exby <= t.by);
      Assert(t1.ey <= t.bxey);
    }
  }
#endif
}
