/*
 * Weight.cpp
 *
 */
#include "SeqData.hpp"
#include "Weight.hpp"
#include "TreeIterator.hpp"
#include "GuideTree.hpp"

const float
Weight::
MIN_VALUE = 0.1;

Weight::
Weight()
  : type(FLAT), seqIds(), data()
{

}
Weight::
~Weight()
{
}
Weight::
Weight(const Weight& weights)
  : type(weights.type),
    seqIds(weights.seqIds),
    data(weights.data)
{
}
Weight&
Weight::
operator=(const Weight& weights)
{
  type = weights.type;
  seqIds = weights.seqIds;
  data = weights.data;
  return *this;
}
string
Weight::
toString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    oss << setprecision(2) 
	<< "- [ "
	<< "seqId: " << seqIds[i] << ", "
	<< "value: " << data[i] << " ] \n";
  }
  return oss.str();
}
void 
Weight::
print() const {
  cout << toString() << flush;
}

bool
Weight::
empty() const
{
  return seqIds.empty();
}
int
Weight::
size() const
{
  return seqIds.size();
}
void
Weight::
clear()
{
  seqIds.clear();
  data.clear();
}
void
Weight::
fill(const float& val)
{
  data.fill(val);
}
const float&
Weight::
ref(const SeqId& seqId) const
{
  const int& i = seqIds.getIndex(seqId);
  Assert(i >= 0);
  return data[i];
}
float&
Weight::
ref(const SeqId& seqId)
{
  const int& i = seqIds.getIndex(seqId);
  Assert(i >= 0);
  return data[i];
}
void
Weight::
setType(const string& weightType)
{
  static const int NTYPE = 2;
  static const struct {string name; Type type;} table[NTYPE] = {
    {"FLAT", FLAT},
    {"CLUSTALW", CLUSTALW}
  };
  for (int i = 0; i < NTYPE; i++) {
    if (weightType == table[i].name) {
      type = table[i].type;
      return;
    }
  }
  Die("bad type %s", weightType.c_str());
}
void
Weight::
setSeqIds(const SeqIds& seqIds1)
{
  clear();
  seqIds = seqIds1;
  data.resize(seqIds.size(), 1.0);
}
void
Weight::
assign(const SeqIds& seqIds1, const Vector<float>& values1)
{
  Assert((int)seqIds1.size() == (int)values1.size());
  seqIds = seqIds1;
  data = values1;
}
void
Weight::
appendPairs(const SeqIds& seqIds1, const Vector<float>& values1)
{
  seqIds.append(seqIds);
  data.insert(data.end(), values1.begin(), values1.end());
}
void
Weight::
append(const Weight& weights1)
{
  seqIds.append(weights1.seqIds);
  data.insert(data.end(), weights1.data.begin(), weights1.data.end());
}
Vector<char>
Weight::
getPackedData() const
{
  ostringstream oss;
  oss << "--- \n";
  for (int i = 0; i < (int)size(); i++) {
    oss << "- [ " 
	<< seqIds[i] << ", "
	<< data[i] << " ] \n";
  }
  const string& str = oss.str();
  return Vector<char>(str.begin(), str.end());
}
void
Weight::
setFromPackedData(const Vector<char>& packedData)
{
  clear();
  Vector<int> ids;
  Vector<float> vals;
  string buf(&(packedData[0]), packedData.size());
  istringstream iss(buf);

  if (!getline(iss, buf)) {
    Die("bad format: no line");
  }
  Vector<string> tokens = tokenize<string>(buf);
  Check((tokens.size() == 1) && (tokens[0] == "---"),
	"bad format: %s", buf.c_str());
  while (getline(iss, buf)) {
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 5) && (tokens[0] == "-")
	  && (tokens[1] == "[") && (tokens[4] == "]"),
	  "bad format: %s", buf.c_str());
    ids.push_back(atoi(tokens[2].c_str()));
    vals.push_back(atof(tokens[3].c_str()));
  }
  assign(SeqIds(ids.begin(), ids.end()), vals);
}
// void
// Weight::
// setFromTupleCount(const SeqData& seqData,
// 		  const SeqIds& seqIds1)
// {
//   GuideTree guideTree;
//   guideTree.setFromTupleCount(seqData, seqIds1);
//   setFromGuideTree(guideTree);
// }
void
Weight::
setFromSeqIdentity(const MSA& msa)
{
  GuideTree guideTree;
  guideTree.setFromSeqIdentity(msa);
  setFromGuideTree(guideTree);
}
void
Weight::
setFromGuideTree(const GuideTree& guideTree)
{
  switch (type) {
  case FLAT: setFlatWeight(guideTree); break;
  case CLUSTALW: setClustalWWeight(guideTree); break;
  default: Die("bad weight type %d", type); break;
  }
  testConsistency();
}
void
Weight::
setFlatWeight(const SeqIds& seqIds1)
{
  setSeqIds(seqIds1);
  data.fill(1.0);
}
void
Weight::
setFlatWeight(const GuideTree& guideTree)
{
  setFlatWeight(guideTree.getSeqIds());
}
void
Weight::
setClustalWWeight(const GuideTree& guideTree0)
{
  GuideTree guideTree(guideTree0);
  guideTree.setTotalEdgeLen(guideTree.getNLeaf());
  setSeqIds(guideTree.getSeqIds());

  typedef pair<float, int> NodeData; // {strength, nLeaf}
  typedef TreeIterator<GuideTree, NodeData> Iterator1;
  Iterator1 it1(guideTree, Iterator1::POSTORDER);
  while (it1.hasNext()) {
    const Iterator1::Link& ilink = it1.next();
    const Iterator1::TLink& tlink = ilink->tlink;
    NodeData& nodeData = ilink->getNodeData();
    
    if (tlink->isRoot()) {
      continue;
    } else if (tlink->isLeaf()) {
      nodeData.second = 1;
    } else {
      const int& n1 = ilink->getChild(0)->getNodeData().second;
      const int& n2 = ilink->getChild(1)->getNodeData().second;
      nodeData.second = n1 + n2;
    }
    const float& edgeLen = tlink->getEdgeLen(tlink->getParent());
    nodeData.first =  edgeLen / (float) nodeData.second;
  }
  typedef TreeIterator<GuideTree> Iterator2;
  Iterator2 it2(guideTree, Iterator2::PREORDER);
  while (it2.hasNext()) {
    const Iterator2::Link& ilink = it2.next();
    const Iterator2::TLink& tlink = ilink->tlink;
    const Iterator1::Link& ilink1 = it1.getLink(tlink);
    NodeData& nodeData = ilink1->getNodeData();
    
    if (tlink->isRoot()) {
      continue;
    } else if (tlink->isLeaf()) {
      ref(tlink->getSeqId()) =
	max((float) MIN_VALUE, nodeData.first);
    } else {
      NodeData& nodeData1 = ilink1->getChild(0)->getNodeData();
      NodeData& nodeData2 = ilink1->getChild(1)->getNodeData();
      nodeData1.first += nodeData.first;
      nodeData2.first += nodeData.first;
    }
  }
}
float
Weight::
getMinWeight() const
{
  if (empty()) return IMPOSSIBLE;

  return *min_element(data.begin(), data.end());
}
void
Weight::
testConsistency() const
{
  Assert(MIN_VALUE <= getMinWeight());
}
