/*
 * Alifold.cpp
 *
 *   Extracted from RNAalifold program of Vienna package (version 1.5)
 *   the author of the original code: Ivo L Hofacker
 *   Reference: 
 *    Secondary structure prediction for aligned RNA sequences.
 *    Ivo L Hofacker and  Martin Fekete and  Peter F Stadler
 *    J Mol Biol. 2002 vol 319 (18) 3724-32
 *
 *   modified to the present form by Hisanori Kiryu
 */
#include "Alifold.hpp"
#include "MSA.hpp"

Alifold::
Alifold()
  : pScoreParamList(NULL),
    pScoreParam(NULL),
    pWeight(NULL),
    pMSA(NULL),
    profile(),
    score(IMPOSSIBLE),
    dpPath()
{
  // fold vars
  energy_set = 0;          /* 0 = BP; 1=any with GC; 2=any with AU parameters */
  noGU = 0;                /* GU not allowed at all */
  no_closingGU = 0;        /* GU allowed only inside stacks */
  tetra_loop = 1;          /* Fold with specially stable 4-loops */
  dangles = 1;	           /* use dangling end energies */
  nonstandards = (char *)0;/* contains allowed non standard bases */
  temperature = 37.0;
  james_rule = 1;          /* interior loops of size 2 get energy 0.8Kcal and
			    * no mismatches (no longer used) */
  pf_scale = (-1);         /* scaling factor to avoid floating point overflows */
  fold_constrained = 0;    /* fold with constraints */
  do_backtrack = 1;        /* calculate pair prob matrix in part_func() */
  noLonelyPairs = 0;       /* avoid helices of length 1 */
  backtrack_type = 'F';    /* 'C' require (1,N) to be bonded;
			    * 'M' seq is part of s multi loop */
  cv_fact = 1.0;
  nc_fact = 1.0;

  memset(&param, 0, sizeof(paramT));
  param_id = (-1);
  P = NULL;
  init_length = (-1);
}
Alifold::
~Alifold()
{
}
void
Alifold::
setScoreParam(ScoreParam& scoreParam)
{
  pScoreParam = &scoreParam;
}
void
Alifold::
setScoreParamList(Vector<ScoreParam>& scoreParamList)
{
  pScoreParamList = &scoreParamList;
  pScoreParam = &scoreParamList[0];
}
void
Alifold::
setWeight(Weight& weights)
{
  pWeight = &weights;
}
void
Alifold::
setMSA(const MSA& msa)
{
  Assert(pWeight);
  pMSA = &msa;
  profile.setFromMSA(*pMSA, *pWeight);
  int n_seq = pMSA->size();
  int length = profile.length();
  seqs.resize(n_seq);
  for (int i = 0; i < n_seq; i++) {
    seqs[i].resize(length);
    for (int j = 0; j < length; j++) {
      int c = profile.seqs.ref(i, j);
      char& c1 = seqs[i][j];
      switch (c) {
      case Alpha::A_CODE:
	c1 = 'A';
	break;
      case Alpha::C_CODE:
	c1 = 'C';
	break;
      case Alpha::G_CODE:
	c1 = 'G';
	break;
      case Alpha::U_CODE:
	c1 = 'U';
	break;
      case Alpha::N_CODE:
	c1 = 'X';
	break;
      case Alpha::GAP_CODE:
	c1 = '_';
	break;
      default:
	Die("bad code [%d, %d]: %d", i, j, c);
	break;
      }
    }
  }
  structure.resize(length);
  structure.fill('.');
}
void
Alifold::
getPairProb(PairProb& pairProb1) const
{
  pairProb1.setLength(profile.length());
  pairProb1.fill(IMPOSSIBLE);
  for (int i = 0; i < (int)profile.length(); i++) {
    for (int j = (i + 1); j < profile.length(); j++) {
      const double& prob = pr[iindx[i + 1] - (j + 1)];
      Check(-0.001 < prob && prob < 1.001,
	    "[ %d, %d ]: { prob: %lf }", i, j, prob);
      pairProb1.ref(i, j) = log(max(1.0e-10, prob)); 
    }
  }
}
void
Alifold::
printPairProb(ostream& fo) const
{
  PairProb pairProb;
  getPairProb(pairProb);
  static const int width = 6;
  fo << setprecision(2);
  for (int i = 0; i < pairProb.length(); i++) {
    for (int j = 0; j < i; j++) {
      const float& val = pairProb.ref(j, i);
      fo << setw(width) << val;
      fo << ((j == (i - 1)) ? "\n" : " ");
    }
  }
}
void
Alifold::
printPairProb(ostream& fo, const float& threshold) const
{
  PairProb pairProb;
  getPairProb(pairProb);
  fo << setprecision(2);
  for (int i = 0; i < pairProb.length(); i++) {
    for (int j = i + 1; j < pairProb.length(); j++) {
      const float& val = pairProb.ref(i, j);
      if (val > threshold) {
	fo << "- [ " << i << ", " << j << " , " << val << " ]\n";
      }
    }
  }
}
void
Alifold::
drawPairProb(ostream& fo, float threshold) const
{
  PairProb pairProb;
  getPairProb(pairProb);
  fo << " ";
  for (int i = 0; i < pairProb.length(); i++) {
    fo << (i % 10) << " ";
  }
  fo << " \n";
  for (int i = 0; i < pairProb.length(); i++) {
    fo << (i % 10);
    fo << string(i * 2, ' ') << "\\ ";
    for (int j = (i + 1); j < pairProb.length(); j++) {
      const float& val = pairProb.ref(i, j);
      fo << (val < threshold ? "  " : "x ");
    }
    fo << (i % 10) << "\n";
  }
  fo << " ";
  for (int i = 0; i < pairProb.length(); i++) {
    fo << (i % 10) << " ";
  }
  fo << " \n" << flush;
}
string
Alifold::
toString() const
{
  ostringstream oss;
  printPairProb(oss, 0.1);
  return oss.str();
}
void
Alifold::
print() const
{
  cout << toString() << endl;
}
void
Alifold::
run()
{
  const float RT_KCAL_MOL = 0.61633008;
  if (purpose == CYK) {
    const float energy = alifold(seqs, structure);
    score = (-RT_KCAL_MOL * energy);
    setDPPath();
  } else {
    const float length = profile.length();
    const float sfact = 1.07;
    const float min_en = alifold(seqs, structure);
    const float kT = (temperature+273.15)*1.98717/1000.; /* in Kcal */
    pf_scale = exp(-(sfact*min_en)/kT/length);
    if (length>2000) fprintf(stderr, "scaling factor %f\n", pf_scale);
    const float energy = alipf_fold(seqs, structure);
    score = (-RT_KCAL_MOL * energy);
  }
}
void
Alifold::
setDPPath()
{
  Assert(dpPath.getAlignLen() == (int)structure.size());
  dpPath.ct.setFromSSCons(string(structure.begin(), structure.end()));
}
void 
Alifold::
make_pair_matrix() 
{// set alias[], pair[][], rtype[]
  static const int BP_pair[NBASES][NBASES] = {
    /*_  A  C  G  U  X  K  I */
    { 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 0, 0, 0, 5, 0, 0, 5},
    { 0, 0, 0, 1, 0, 0, 0, 0},
    { 0, 0, 2, 0, 3, 0, 0, 0},
    { 0, 6, 0, 4, 0, 0, 0, 6},
    { 0, 0, 0, 0, 0, 0, 2, 0},
    { 0, 0, 0, 0, 0, 1, 0, 0},
    { 0, 6, 0, 0, 5, 0, 0, 0}
  };
  static const int rtype0[8] = {0, 2, 1, 4, 3, 6, 5, 7}; 
  for (int i = 0; i < (MAXALPHA + 1); i++) {
    rtype[i] = (i < 8 ? rtype0[i] : 7);
  }
  if (energy_set == 0) {
    for (int i = 0; i < 5; i++) {
      alias[i] = (short) i;
    }
    alias[5] = 3; /* X <-> G */
    alias[6] = 2; /* K <-> C */
    alias[7] = 0; /* I <-> default base '@' */
    for (int i = 0; i < NBASES; i++) {
      for (int j = 0; j < NBASES; j++) {
	pair[i][j] = BP_pair[i][j];
      }
    }      
    if (noGU) {
      pair[3][4] = 0;
      pair[4][3] = 0;
    }
    if (nonstandards != NULL) {  /* allow nonstandard bp's */ 
      for (int i = 0; i < (int)strlen(nonstandards); i += 2) {
    	pair[encode_char(nonstandards[i])]
    	  [encode_char(nonstandards[i+1])]=7;
      }
    }
    for (int i = 0; i < NBASES; i++) {
      for (int j = 0; j < NBASES; j++) {
	rtype[pair[i][j]] = pair[j][i];
      }
    }      
  } else {
    for (int i = 0; i <= MAXALPHA; i++) {
      for (int j = 0; j <= MAXALPHA; j++) {
	pair[i][j] = 0;
      }
    }
    if (energy_set == 1) {
      for (int i = 1; i < MAXALPHA;) {
	alias[i++] = 3;  /* A <-> G */
	alias[i++] = 2;  /* B <-> C */
      }
      for (int i = 1; i < MAXALPHA; i++) {
	pair[i][i + 1] = 2;    /* AB <-> GC */
	i++;
	pair[i][i - 1] = 1;    /* BA <-> CG */
      }
    } else if (energy_set == 2) {
      for (int i = 1; i < MAXALPHA;) {
	alias[i++] = 1;  /* A <-> A*/
	alias[i++] = 4;  /* B <-> U */
      }
      for (int i = 1; i < MAXALPHA; i++) {
	pair[i][i + 1] = 5;    /* AB <-> AU */
	i++;
	pair[i][i - 1] = 6;    /* BA <-> UA */
      }
    } else if (energy_set == 3) {
      for (int i = 1; i < (MAXALPHA - 2); ) {
	alias[i++] = 3;  /* A <-> G */
	alias[i++] = 2;  /* B <-> C */
	alias[i++] = 1;  /* C <-> A */
	alias[i++] = 4;  /* D <-> U */
      }
      for (int i = 1; i < (MAXALPHA - 2); i++) {
	pair[i][i + 1] = 2;    /* AB <-> GC */
	i++;
	pair[i][i - 1] = 1;    /* BA <-> CG */
	i++;
	pair[i][i + 1] = 5;    /* CD <-> AU */
	i++;
	pair[i][i - 1] = 6;    /* DC <-> UA */
      }
    } else {
      nrerror("What energy_set are YOU using??");
    }
    for (int i = 0; i <= MAXALPHA; i++) {
      for (int j = 0; j <= MAXALPHA; j++) {
	rtype[pair[i][j]] = pair[j][i];
      }
    }
  }
}
void 
Alifold::
update_fold_params()
{
  P = scale_parameters();
  make_pair_matrix();
  if (init_length < 0) init_length=0;
}
Alifold::paramT *
Alifold::
scale_parameters()
{
  memset(&param, 0, sizeof(paramT));
  paramT& p = param;
  int& id = param_id;

  unsigned int i = 0;
  unsigned int j = 0;
  unsigned int k = 0;
  unsigned int l = 0;
  double tempf = 0.0;
  if ((fabs(p.temperature - temperature)<1e-6)&&(id == p.id)) return &p;

  tempf = ((temperature+EP::K0)/EP::Tmeasure);
  for (i=0; i<31; i++) 
    p.hairpin[i] = (int) (EP::hairpin37[i]*(tempf));
  for (i=0; i<=MIN2(30u,(unsigned)EP::MAXLOOP); i++) {
    p.bulge[i] = (int) (EP::bulge37[i]*tempf);
    p.internal_loop[i]= (int) (EP::internal_loop37[i]*tempf);
  }
  p.lxc = EP::lxc37*tempf;
  for (; i<=EP::MAXLOOP; i++) {
    p.bulge[i] = p.bulge[30]+(int)(p.lxc*log((double)(i)/30.));
    p.internal_loop[i] = p.internal_loop[30]+(int)(p.lxc*log((double)(i)/30.));
  }
  for (i=0; i<5; i++)
    p.F_ninio[i] = (int) (EP::F_ninio37[i]*tempf);
   
  for (i=0; (i*7)<strlen(EP::Tetraloops); i++) 
    p.TETRA_ENERGY[i] = (int)(EP::TETRA_ENTH37 - (EP::TETRA_ENTH37-EP::TETRA_ENERGY37[i])*tempf);
  for (i=0; (i*5)<strlen(EP::Triloops); i++) 
    p.Triloop_E[i] = EP::Triloop_E37[i];
   
  p.MLbase = (int)(EP::ML_BASE37*tempf);
  for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
    p.MLintern[i] = (int)(EP::ML_intern37*tempf);
    p.MLintern[i] +=  (i>2)?EP::TerminalAU:0;
  }
  p.MLclosing = (int)(EP::ML_closing37*tempf);

  p.TerminalAU = EP::TerminalAU;

  /* stacks    G(T) = H - [H - G(T0)]*T/T0 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      p.stack[i][j] = (int)(EP::enthalpies[i][j] -
			    (EP::enthalpies[i][j] - EP::stack37[i][j])*tempf);

  /* mismatches */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
	p.mismatchI[i][j][k] = (int)(EP::mism_H[i][j][k] -
	  (EP::mism_H[i][j][k] - EP::mismatchI37[i][j][k])*tempf);
	p.mismatchH[i][j][k] = (int)(EP::mism_H[i][j][k] -
	  (EP::mism_H[i][j][k] - EP::mismatchH37[i][j][k])*tempf);
	p.mismatchM[i][j][k] = (int)(EP::mism_H[i][j][k] -
	  (EP::mism_H[i][j][k] - EP::mismatchM37[i][j][k])*tempf);
      }
   
  /* dangles */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++) {
      int dd = 0;
      dd = (int)(EP::dangle5_H[i][j] - (EP::dangle5_H[i][j] - EP::dangle5_37[i][j])*tempf); 
      p.dangle5[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
      dd = (int)(EP::dangle3_H[i][j] - (EP::dangle3_H[i][j] - EP::dangle3_37[i][j])*tempf);
      p.dangle3[i][j] = (dd>0) ? 0 : dd;  /* must be <= 0 */
    }
  /* interior 1x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) 
	  p.int11[i][j][k][l] = (int)(EP::int11_H[i][j][k][l] -
	    (EP::int11_H[i][j][k][l] - EP::int11_37[i][j][k][l])*tempf);

  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m = 0;
	  for (m=0; m<5; m++)
	    p.int21[i][j][k][l][m] = (int)(EP::int21_H[i][j][k][l][m] -
	      (EP::int21_H[i][j][k][l][m] - EP::int21_37[i][j][k][l][m])*tempf);
	}
  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m = 0;
	  int n = 0;
	  for (m=0; m<5; m++)
	    for (n=0; n<5; n++)	     
	      p.int22[i][j][k][l][m][n] = (int)(EP::int22_H[i][j][k][l][m][n] -
		(EP::int22_H[i][j][k][l][m][n]-EP::int22_37[i][j][k][l][m][n])*tempf);
	}

  strncpy(p.Tetraloops, EP::Tetraloops, 1400);
  strncpy(p.Triloops, EP::Triloops, 240);

  p.temperature = temperature;
  p.id = ++id;
  return &p;
}

////////////////////////////////////////////////////////////////////
/// alifold
float 
Alifold::
alifold(Array< Array<char> >& strings, Array<char>& structure)
{
  //#define MAXSECTORS      500     /* dimension for a backtrack array */
  //#define LOCALITY        0.      /* locality parameter for base-pairs */
  //#define MIN2(A, B)      ((A) < (B) ? (A) : (B))
  enum {
    TURN = EP::TURN,
    MAXLOOP = EP::MAXLOOP
  };
  const int& TerminalAU = EP::TerminalAU;

  struct sect {
    int  i;
    int  j;
    int ml;
  } 
  sector[MAXSECTORS] = {};   /* backtracking sectors */

  int i = 0;
  int j = 0;
  int k = 0;
  int p = 0;
  int q = 0;
  int length = 0;
  int energy = 0;
  int new_c = 0;
  int decomp = 0;
  int MLenergy = 0;
  int new_fML = 0;
  int s = 0;
  int b = 0;
  int mm = 0;/*, max_separation */;
  int n_seq = 0;
  int type_2 = 0;
  int tt = 0;
  //int   *type;
  //short **S;
  Array<int> type;
  Array< Array<short> > S;
  int cov_en = 0;

  //length = (int) strlen(strings[0]);
  length = strings[0].size();
  if (length>init_length) init_alifold(length);
  if ((P==NULL)||(fabs(P->temperature - temperature)>1e-6)) { 
  	update_fold_params();  P = scale_parameters();
  }
  //for (s=0; strings[s]!=NULL; s++); 
  //n_seq = s;
  n_seq = strings.size();
  //S = (short **) space(n_seq*sizeof(short *));
  //type = (int *) space(n_seq*sizeof(int));
  S.resize(n_seq);
  type.resize(n_seq);
  for (s=0; s<n_seq; s++) { 
    //if (strlen(strings[s]) != length) nrerror("uneqal seqence lengths");
    if ((int)strings[s].size() != length) nrerror("uneqal seqence lengths");
    S[s] = encode_seq(strings[s]);
  }
  //make_pscores((const short **) S, n_seq, structure);
  make_pscores_alifold(S, n_seq, structure);

  for (j=1; j<=length; j++) {
    Fmi[j]=DMLi[j]=DMLi1[j]=DMLi2[j]=INF;
  }
   
  for (j = 1; j<=length; j++)
    for (i=(j>TURN?(j-TURN):1); i<j; i++) {
      c[indx[j]+i] = fML[indx[j]+i] = INF;

    }       
  
  for (i = length-TURN-1; i >= 1; i--) { /* i,j in [1..length] */
      
    for (j = i+TURN+1; j <= length; j++) {
      int ij = 0;
      int psc = 0;
      ij = indx[j]+i;

      for (s=0; s<n_seq; s++) {
	type[s] = pair[S[s][i]][S[s][j]];
	if (type[s]==0) type[s]=7;
      }
 
      psc = pscore[indx[j]+i];
		 
      if (psc>=cv_fact*MINPSCORE) {   /* a pair to consider */
	int stackEnergy = INF;
	/* hairpin ----------------------------------------------*/
	
	
	for (new_c=s=0; s<n_seq; s++)
	  new_c += HairpinE(j-i-1,type[s],S[s][i+1],S[s][j-1],&strings[s][i-1]);
	   
	/*--------------------------------------------------------
	  check for elementary structures involving more than one
	  closing pair.
	  --------------------------------------------------------*/
	   
	for (p = i+1; p <= MIN2(j-2-TURN,i+MAXLOOP+1) ; p++) {
	  int minq = j-i+p-MAXLOOP-2;
	  if (minq<p+1+TURN) minq = p+1+TURN;
	  for (q = minq; q < j; q++) {
	    if (pscore[indx[q]+p]<MINPSCORE) continue;


	    for (energy = s=0; s<n_seq; s++) {
	      type_2 = pair[S[s][q]][S[s][p]]; /* q,p not p,q! */
	      if (type_2 == 0) type_2 = 7;
	      energy += LoopEnergy(p-i-1, j-q-1, type[s], type_2,
				   S[s][i+1], S[s][j-1], 
				   S[s][p-1], S[s][q+1]);
	    }
	    new_c = MIN2(energy+c[indx[q]+p], new_c);
	    if ((p==i+1)&&(j==q+1)) stackEnergy = energy; /* remember stack energy */
	       
	  } /* end q-loop */
	} /* end p-loop */
	   
	/* multi-loop decomposition ------------------------*/
	
	decomp = DMLi1[j-1];
	if (dangles) {
	  int d3 = 0;
	  int d5 = 0; 
	  for (s=0; s<n_seq; s++) {
	    tt = rtype[type[s]];
	    d3 = P->dangle3[tt][S[s][i+1]];
	    d5 = P->dangle5[tt][S[s][j-1]];
	    decomp += d5 + d3;
	  }
	}
	
	MLenergy = decomp + n_seq*P->MLclosing;
	for (s=0; s<n_seq; s++) 
	  MLenergy += P->MLintern[type[s]];
	
	new_c = MLenergy < new_c ? MLenergy : new_c;
	
	new_c = MIN2(new_c, cc1[j-1]+stackEnergy);
	cc[j] = new_c - psc; /* add covariance bonnus/penalty */
	if (noLonelyPairs)
	  c[ij] = cc1[j-1]+stackEnergy-psc;
	else
	  c[ij] = cc[j];
	   
      } /* end >> if (pair) << */
	 
      else c[ij] = INF;


      /* done with c[i,j], now compute fML[i,j] */
      /* free ends ? -----------------------------------------*/

      new_fML = fML[ij+1]+n_seq*P->MLbase;
      new_fML = MIN2(fML[indx[j-1]+i]+n_seq*P->MLbase, new_fML);
      energy = c[ij];
      for (s=0; s<n_seq; s++) {
	energy += P->MLintern[type[s]];
	if (dangles) {  /* double dangles */
	  if (i>1)      energy += P->dangle5[type[s]][S[s][i-1]];
	  if (j<length) energy += P->dangle3[type[s]][S[s][j+1]];
	}
      }
      new_fML = MIN2(energy, new_fML);

      
      /* modular decomposition -------------------------------*/
    
      for (decomp = INF, k = i+1+TURN; k <= j-2-TURN; k++)
	decomp = MIN2(decomp, Fmi[k]+fML[indx[j]+k+1]);
      
      DMLi[j] = decomp;               /* store for use in ML decompositon */
      new_fML = MIN2(new_fML,decomp);
      
      /* coaxial stacking deleted */

      fML[ij] = Fmi[j] = new_fML;     /* substring energy */

    }

    {
      //int *FF; /* rotate the auxilliary arrays */
      Array<int> FF; /* rotate the auxilliary arrays */
      FF = DMLi2; DMLi2 = DMLi1; DMLi1 = DMLi; DMLi = FF;
      //FF.swap(DMLi2); DMLi2.swap(DMLi1); DMLi1.swap(DMLi); DMLi.swap(FF);
      FF = cc1; cc1=cc; cc=FF;
      //FF.swap(cc1); cc1.swap(cc); cc.swap(FF);
      for (j=1; j<=length; j++) {cc[j]=Fmi[j]=DMLi[j]=INF; }
    }
  }
  /* calculate energies of 5' and 3' fragments */
    
  f5[TURN+1]=0;
  for (j=TURN+2; j<=length; j++) {
    f5[j] = f5[j-1];
    if (c[indx[j]+1]<INF) {
      energy = c[indx[j]+1];
      for (s=0; s<n_seq; s++) {
	int type = 0;
	type = pair[S[s][1]][S[s][j]]; if (type==0) type=7;
	if (type>2) energy += TerminalAU;
	if ((dangles)&&(j<length))  /* double dangles */
	  energy += P->dangle3[type][S[s][j+1]];
      }
      f5[j] = MIN2(f5[j], energy);
    }
    for (i=j-TURN-1; i>1; i--) {
      if (c[indx[j]+i]<INF) {
	energy = f5[i-1]+c[indx[j]+i];
	for (s=0; s<n_seq; s++) {
	  int type = 0;
	  type = pair[S[s][i]][S[s][j]]; if (type==0) type=7;
	  if (type>2) energy += TerminalAU;
	  if (dangles) {
	    energy += P->dangle5[type][S[s][i-1]];
	    if (j<length) energy += P->dangle3[type][S[s][j+1]];
	  }
	}
	f5[j] = MIN2(f5[j], energy);
      }
    }
  }
  
#if ALIDEBUG
  for (i = length; i >= 1; i--) { /* i,j in [1..length] */
    for (j = i+1; j <= length; j++) {
      printf("[%d,%d]: c:%d,fML:%d,psc:%d\n", i, j, 
	     c[indx[j]+i],fML[indx[j]+i],pscore[indx[j]+i]);
    }
  }
  for (i = 1; i <= length; i++) {
    printf("f5[%d]=%d\n", i, f5[i]);
  }
#endif
 
   
  /*------------------------------------------------------------------
    trace back through the "c", "f5" and "fML" arrays to get the
    base pairing list. No search for equivalent structures is done.
    This inverts the folding procedure, hence it's very fast.
    ------------------------------------------------------------------*/

  b = s = 0;
  sector[++s].i = 1;
  sector[s].j = length;
  sector[s].ml = (backtrack_type=='M') ? 1 : ((backtrack_type=='C')?2:0);
   
  while (s>0) {
    int ss = 0;
    int ml = 0;
    int fij = 0;
    int fi = 0;
    int cij = 0;
    int traced = 0;
    int i1 = 0;
    int j1 = 0;
    int d3 = 0;
    int d5 = 0;
    int jj = 0;
    int canonical = 1;     /* (i,j) closes a canonical structure */
    i  = sector[s].i;
    j  = sector[s].j;
    ml = sector[s--].ml;   /* ml is a flag indicating if backtracking is to 
			      occur in the fML- (1) or in the f-array (0) */
#if ALIDEBUG
    printf("i=%d,j=%d,ml=%d,s=%d\n", i, j, ml, s);
#endif

    if (ml==2) {
      base_pair[++b].i = i;
      base_pair[b].j   = j;
      cov_en += pscore[indx[j]+i];
      goto repeat1; 
    }

    if (j < i+TURN+1) continue; /* no more pairs in this interval */

    fij = (ml)? fML[indx[j]+i] : f5[j];
    fi  = (ml)?(fML[indx[j-1]+i]+n_seq*P->MLbase):f5[j-1];

    if (fij == fi) {  /* 3' end is unpaired */
      sector[++s].i = i;
      sector[s].j   = j-1;
      sector[s].ml  = ml;
      continue;
    }
     
    if (ml == 0) { /* backtrack in f5 */
      /* j or j-1 is paired. Find pairing partner */
      for (i=j-TURN-1,traced=0; i>=1; i--) {
	int cc = 0;
	int en = 0;
	jj = i-1; 
	if (c[indx[j]+i]<INF) {
	  cc = c[indx[j]+i];
	  for (ss=0; ss<n_seq; ss++) {
	    type[ss] = pair[S[ss][i]][S[ss][j]];
	    if (type[ss]==0) type[ss] = 7;
	    if (type[ss]>2) cc += TerminalAU;
	  }
	  en = cc + f5[i-1];
	  if (dangles) {
	    for (ss=0; ss<n_seq; ss++) {
	      if (i>1)      en += P->dangle5[type[ss]][S[ss][i-1]];
	      if (j<length) en += P->dangle3[type[ss]][S[ss][j+1]]; 
	    }
	  }
	  if (fij == en) traced=j;
	}
	if (traced) break;
      }

      if (!traced) nrerror("backtrack failed in f5");
      sector[++s].i = 1;
      sector[s].j   = jj;
      sector[s].ml  = ml;
       
      j=traced;
      base_pair[++b].i = i;
      base_pair[b].j   = j;
      cov_en += pscore[indx[j]+i];
      goto repeat1;
    }
    else { /* trace back in fML array */
      int cij1 = INF;
      int ci1j = INF;
      int ci1j1 = INF;
      if (fML[indx[j]+i+1]+n_seq*P->MLbase == fij) { /* 5' end is unpaired */
	sector[++s].i = i+1;
	sector[s].j   = j;
	sector[s].ml  = ml;
	continue;
      } 

      cij = c[indx[j]+i];
      for (ss=0; ss<n_seq; ss++) {
	tt  = pair[S[ss][i]][S[ss][j]];
	if (tt==0) tt=7;
	cij += P->MLintern[tt];      
	if (dangles) {       /* double dangles */
	  if (i>1)      cij += P->dangle5[tt][S[ss][i-1]];
	  if (j<length) cij += P->dangle3[tt][S[ss][j+1]];
	}
      }

      if ((fij==cij)||(fij==ci1j)||(fij==cij1)||(fij==ci1j1)) {
	/* found a pair */
	if (fij==ci1j) i++;
	else if (fij==cij1) j--;
	else if (fij==ci1j1) {i++; j--;}
	base_pair[++b].i = i;
	base_pair[b].j   = j;
	cov_en += pscore[indx[j]+i];
	goto repeat1;
      } 
       
      for (k = i+1+TURN; k <= j-2-TURN; k++) 
	if (fij == (fML[indx[k]+i]+fML[indx[j]+k+1])) 
	  break;
      
      sector[++s].i = i;
      sector[s].j   = k;
      sector[s].ml  = ml;
      sector[++s].i = k+1;
      sector[s].j   = j;
      sector[s].ml  = ml;
      
      if (k>j-2-TURN) nrerror("backtrack failed in fML");
      continue;
    }
     
  repeat1:
      
    /*----- begin of "repeat:" -----*/
    if (canonical)  cij = c[indx[j]+i];

    for (ss=0; ss<n_seq; ss++) {
      type[ss] = pair[S[ss][i]][S[ss][j]];
      if (type[ss]==0) type[ss] = 7;
    }
    
    if (noLonelyPairs) 
      if (cij == c[indx[j]+i]) {
	/* (i.j) closes canonical structures, thus
	   (i+1.j-1) must be a pair                */
	for (ss=0; ss<n_seq; ss++) {
	  type_2 = pair[S[ss][j-1]][S[ss][i+1]];  /* j,i not i,j */
	  if (type_2==0) type_2 = 7;
	  cij -= P->stack[type[ss]][type_2];
	}
	cij += pscore[indx[j]+i];
	base_pair[++b].i = i+1;
	base_pair[b].j   = j-1;
	cov_en += pscore[indx[j-1]+i+1];
	i++; j--; 
	canonical=0;
	goto repeat1;
      }
    canonical = 1;
    cij += pscore[indx[j]+i];

    {int cc=0;
    //for (ss=0; ss<n_seq; ss++) 
    //  cc += HairpinE(j-i-1, type[ss], S[ss][i+1], S[ss][j-1], strings[ss]+i-1);
    for (ss=0; ss<n_seq; ss++) 
      cc += HairpinE(j-i-1, type[ss], S[ss][i+1], S[ss][j-1], &strings[ss][i-1]);

    if (cij == cc) /* found hairpin */ 
      continue;
    }
    for (p = i+1; p <= MIN2(j-2-TURN,i+MAXLOOP+1); p++) {
      int minq = j-i+p-MAXLOOP-2;
      if (minq<p+1+TURN) minq = p+1+TURN;
      for (q = j-1; q >= minq; q--) {
	 
	if (c[indx[q]+p]>=INF) continue;

	for (ss=energy=0; ss<n_seq; ss++) {
	  type_2 = pair[S[ss][q]][S[ss][p]];  /* q,p not p,q */
	  if (type_2==0) type_2 = 7;
	  energy += LoopEnergy(p-i-1, j-q-1, type[ss], type_2,
			       S[ss][i+1], S[ss][j-1], 
			       S[ss][p-1], S[ss][q+1]);
	}
	traced = (cij == energy+c[indx[q]+p]);
	if (traced) {
	  base_pair[++b].i = p;
	  base_pair[b].j   = q;
	  cov_en += pscore[indx[q]+p];
	  i = p, j = q;
	  goto repeat1;
	}
      }
    }
     
    /* end of repeat: --------------------------------------------------*/

#if ALIDEBUG
    printf("end of repeat i=%d,j=%d,ml=%d,s=%d\n", i, j, ml, s);
#endif


    /* (i.j) must close a multi-loop */

    mm = n_seq*P->MLclosing;
    for (ss=d3=d5=0; ss<n_seq; ss++) {
      tt = rtype[type[ss]];
      mm += P->MLintern[tt];
      d5 += P->dangle5[tt][S[ss][j-1]];
      d3 += P->dangle3[tt][S[ss][i+1]];
#if ALIDEBUG
    printf("ss:%d,type:%d,tt:%d,mm:%d,d5:%d,d3:%d\n", ss, type[ss],tt, mm, d5, d3);
#endif
    }
    i1 = i+1; j1 = j-1;
    sector[s+1].ml  = sector[s+2].ml = 1; 

    for (k = i+2+TURN; k < j-2-TURN; k++) {
      int en = 0;
      en = fML[indx[k]+i+1]+fML[indx[j-1]+k+1]+mm;
      if (dangles) /* double dangles */
	en += d5+d3;
#if ALIDEBUG
    printf("decomp i=%d,j=%d,ml=%d,s=%d,i1=%d,j1=%d,k=%d\n", i,j,ml,s,i1,j1,k);
    printf("decomp cij=%d,mm=%d,d5=%d,d3=%d,en=%d,dangles=%d,fML[%d]=%d,fML[%d]=%d\n", 
	    cij,mm,d5,d3,en,dangles,indx[k]+i+1,fML[indx[k]+i+1],indx[j-1]+k+1,fML[indx[j-1]+k+1]);
#endif
      if (cij == en) 
	break;

    }
    if (k<=j-3-TURN) { /* found the decomposition */
#if ALIDEBUG
    printf("found decomp i=%d,j=%d,ml=%d,s=%d,i1=%d,j1=%d,k=%d\n", i,j,ml,s,i1,j1,k);
    printf("found decomp cij=%d,mm=%d,d5=%d,d3=%d,fML[%d]=%d,fML[%d]=%d\n", 
	   cij,mm,d5,d3,indx[k]+i+1,fML[indx[k]+i+1],indx[j-1]+k+1,fML[indx[j-1]+k+1]);
#endif
      sector[++s].i = i1;
      sector[s].j   = k;
      sector[++s].i = k+1;
      sector[s].j   = j1;
    } else {
      Break::Here();
	nrerror("backtracking failed in repeat");
    } 
    
  }
  
  base_pair[0].i = b;    /* save the total number of base pairs */

  parenthesis_structure(structure, length);

  //free(type);
  //for (s=0; s<n_seq; s++) free(S[s]); 
  //free(S);
  /* fprintf(stderr, "covariance energy %6.2f\n", cov_en/100.); */
  if (backtrack_type=='C')
    return (float) c[indx[length]+1]/(n_seq*100.);
  else if (backtrack_type=='M')
    return (float) fML[indx[length]+1]/(n_seq*100.);
  else
    return (float) f5[length]/(n_seq*100.);
}
int 
Alifold::
encode_char(char c) {
  /* return numerical representation of base used e.g. in pair[][] */
  static const char Law_and_Order[] = "_ACGUTXKI";
  int code = 0;
  if (energy_set > 0) {
    code = ((int) (c - 'A') + 1);
  } else {
    char* pos = strchr(Law_and_Order, c);
    if (pos == NULL) {
      code = 0;
    } else {
      code = (int) (pos - Law_and_Order);
    }
    if (code > 4) {
      code--; /* make T and U equivalent */
    }
  }
  return code;
}
Array<short>
Alifold::
encode_seq(const Array<char>& sequence) {
  unsigned int i = 0;
  unsigned int l = 0;
  //short *S;
  Array<short> S;
  //l = strlen(sequence);
  l = sequence.size();
  //S = (short *) space(sizeof(short)*(l+1));
  S.resize(l + 1);
  S[0] = (short) l;
  
  /* make numerical encoding of sequence */
  for (i=1; i<=l; i++) 
    S[i]= (short) encode_char(toupper(sequence[i-1]));
  
  return S;
}
void 
Alifold::
parenthesis_structure(Array<char>& structure, int length)
{
  int n = 0;
  int k = 0;
  
  for (n = 0; n <= length-1; structure[n++] = '.') ;
  //structure[length] = '\0';

  for (k = 1; k <= base_pair[0].i; k++) {
    structure[base_pair[k].i-1] = '<';
    structure[base_pair[k].j-1] = '>';
  }
}
void 
Alifold::
init_alifold(int length)
{
  unsigned int n = 0;
  if (length<1) nrerror("initialize_fold: argument must be greater 0");
  if (init_length>0) free_alifold_arrays();
  //get_arrays((unsigned) length);
  get_arrays_alifold((unsigned int) length);
  make_pair_matrix();
  init_length=length;

  for (n = 1; n <= (unsigned int) length; n++)
    indx[n] = (n*(n-1)) >> 1;        /* n(n-1)/2 */
  
  update_fold_params();
}
void 
Alifold::
get_arrays_alifold(unsigned int size)
{
  //indx =  (int *) space(sizeof(int)*(size+1));
  //c     = (int *) space(sizeof(int)*((size*(size+1))/2+2));
  //fML   = (int *) space(sizeof(int)*((size*(size+1))/2+2));
  //
  //pscore = (int *) space(sizeof(int)*((size*(size+1))/2+2));
  //f5    = (int *) space(sizeof(int)*(size+2));
  //cc    = (int *) space(sizeof(int)*(size+2));
  //cc1   = (int *) space(sizeof(int)*(size+2));
  //Fmi   = (int *) space(sizeof(int)*(size+1));
  //DMLi  = (int *) space(sizeof(int)*(size+1));
  //DMLi1  = (int *) space(sizeof(int)*(size+1));
  //DMLi2  = (int *) space(sizeof(int)*(size+1));
  //base_pair = (struct bond *) space(sizeof(struct bond)*(1+size/2));

  indx.resize(size + 1);
  c.resize((size* (size + 1)) / 2 + 2);
  fML.resize((size * (size + 1)) / 2 + 2);

  pscore.resize((size * (size + 1)) / 2 + 2);
  f5.resize(size + 2);
  cc.resize(size + 2);
  cc1.resize(size + 2);
  Fmi.resize(size + 1);
  DMLi.resize(size + 1);
  DMLi1.resize(size + 1);
  DMLi2.resize(size + 1);
  base_pair.resize(1 + size / 2);

  indx.fill(0);
  c.fill(0);
  fML.fill(0);
  pscore.fill(0);
  f5.fill(0);
  cc.fill(0);
  cc1.fill(0);
  Fmi.fill(0);
  DMLi.fill(0);
  DMLi1.fill(0);
  DMLi2.fill(0);
  bond null_bond = {};
  base_pair.fill(null_bond);
}
void 
Alifold::
free_alifold_arrays()
{
  //free(indx); free(c); free(fML); free(f5); free(cc); free(cc1); 
  //free(pscore);
  //free(base_pair); free(Fmi);
  //free(DMLi); free(DMLi1);free(DMLi2);
  init_length=0;
}
void 
Alifold::
make_pscores_alifold(const Array< Array<short> >& S, int n_seq,
		     const Array<char>& structure) 
{
  enum {
    TURN = EP::TURN
  };
  /* calculate co-variance bonus for each pair depending on  */
  /* compensatory/consistent mutations and incompatible seqs */
  /* should be 0 for conserved pairs, >0 for good pairs      */
#define NONE -10000 /* score for forbidden pairs */
  int n = 0;
  int i = 0;
  int j = 0;
  int k = 0;
  int l = 0;
  int s = 0;
  int score = 0;
  int dm[7][7]={{0,0,0,0,0,0,0}, /* hamming distance between pairs */
	       	{0,0,2,2,1,2,2} /* CG */,
		{0,2,0,1,2,2,2} /* GC */,
		{0,2,1,0,2,1,2} /* GU */,
		{0,1,2,2,0,2,1} /* UG */,
		{0,2,2,1,2,0,2} /* AU */,
		{0,2,2,2,1,2,0} /* UA */};
  n=S[0][0];  /* length of seqs */
  for (i=1; i<n; i++) {
    for (j=i+1; (j<i+TURN+1) && (j<=n); j++) 
      pscore[indx[j]+i] = NONE;
    for (j=i+TURN+1; j<=n; j++) {
      int pfreq[8]={0,0,0,0,0,0,0,0};
      for (s=0; s<n_seq; s++) {
	int type = 0;
	if (S[s][i]==0 && S[s][j]==0) type = 7; /* gap-gap  */	
	else type = pair[S[s][i]][S[s][j]];
	
	pfreq[type]++;
      }
      if (pfreq[0]*2>n_seq) { pscore[indx[j]+i] = NONE; continue;}
      for (k=1,score=0; k<=6; k++) /* ignore pairtype 7 (gap-gap) */
	for (l=k+1; l<=6; l++) 
	  /* scores for replacements between pairtypes    */
	  /* consistent or compensatory mutations score 1 or 2  */
	  score += pfreq[k]*pfreq[l]*dm[k][l];
      /* counter examples score -1, gap-gap scores -0.25   */
      pscore[indx[j]+i] = (int)(cv_fact *
	((UNIT*score)/n_seq - UNIT*pfreq[0]*nc_fact - UNIT*pfreq[7]*0.25));
    }
  }
  
  if (noLonelyPairs) /* remove unwanted pairs */
    for (k=1; k<n-TURN-1; k++) 
      for (l=1; l<=2; l++) {
	int type = 0;
	int ntype = 0;
	int otype = 0;
	i=k; j = i+TURN+l;
	type = pscore[indx[j]+i];
	while ((i>=1)&&(j<=n)) {
	  if ((i>1)&&(j<n)) ntype = pscore[indx[j+1]+i-1];
	  if ((otype<-4*UNIT)&&(ntype<-4*UNIT))  /* worse than 2 counterex */
	    pscore[indx[j]+i] = NONE; /* i.j can only form isolated pairs */
	  otype =  type;
	  type  = ntype;
	  i--; j++;
	}
      }
  

  //if (fold_constrained&&(structure!=NULL)) {
  if (fold_constrained && (!structure.empty())) {
    //int hx, *stack;
    int hx = 0;
    Array<int> stack;
    //stack = (int *) space(sizeof(int)*(n+1));
    stack.resize(n + 1);
    
    for(hx=0, j=1; j<=n; j++) {
      switch (structure[j-1]) {
      case 'x': /* can't pair */ 
        for (l=1; l<j-TURN; l++) pscore[indx[j]+l] = NONE;
        for (l=j+TURN+1; l<=n; l++) pscore[indx[l]+j] = NONE;
        break;
      case '(':
        stack[hx++]=j;
        /* fallthrough */
      case '<': /* pairs upstream */
        for (l=1; l<j-TURN; l++) pscore[indx[j]+l] = NONE;
        break;
      case ')':
        if (hx<=0) {
	  string structure1(structure.begin(), structure.end());
          fprintf(stderr, "%s\n", structure1.c_str());
          nrerror("unbalanced brackets in constraints");
        }
        i = stack[--hx];
	for (k=i+1; k<=n; k++) pscore[indx[i]+k] = NONE;
        for (l=i+1; l<=j; l++) 
	  for (k=j; k<=n; k++) pscore[indx[l]+k] = NONE;
	for (k=1; k<=i; k++) 
	  for (l=i; l<=j; l++) pscore[indx[k]+l] = NONE;
	for (k=1; k<j; k++) pscore[indx[k]+j] = NONE;
        if (pscore[indx[j]+i]==NONE) 
	  pscore[indx[j]+i] = 0;
        /* fallthrough */
      case '>': /* pairs downstream */
        for (l=j+TURN+1; l<=n; l++) pscore[indx[l]+j] = NONE;
        break;
      }
    }
    if (hx!=0) {
      string structure1(structure.begin(), structure.end());
      fprintf(stderr, "%s\n", structure1.c_str());
      nrerror("unbalanced brackets in constraint string");
    }
    //free(stack);
  }
#undef NONE
}
int 
Alifold::
HairpinE(int size, int type, int si1, int sj1, const char *str) 
{
  int energy;
  energy = (size <= 30) ? P->hairpin[size] :
    P->hairpin[30]+(int)(P->lxc*log((size)/30.));
  if (tetra_loop)
    if (size == 4) { /* check for tetraloop bonus */
      char tl[7]={};
      char *ts = NULL;
      strncpy(tl, str, 6);
      if ((ts=strstr(P->Tetraloops, tl))) 
	energy += P->TETRA_ENERGY[(ts - P->Tetraloops)/7];
    }
  if (size == 3) {
    char tl[6]={0,0,0,0,0,0};
    char *ts = NULL;
    strncpy(tl, str, 5);
    if ((ts=strstr(P->Triloops, tl))) 
      energy += P->Triloop_E[(ts - P->Triloops)/6];
    
    if (type>2)  /* neither CG nor GC */
      energy += P->TerminalAU; /* penalty for closing AU GU pair */
  }
  else  /* no mismatches for tri-loops */
    energy += P->mismatchH[type][si1][sj1];
  
  return energy;
}
int 
Alifold::
LoopEnergy(int n1, int n2, int type, int type_2,
	   int si1, int sj1, int sp1, int sq1) 
{
  enum {
    MAXLOOP = EP::MAXLOOP
  };
  const int& MAX_NINIO = EP::MAX_NINIO;
  /* compute energy of degree 2 loop (stack bulge or interior) */
  int nl = 0;
  int ns = 0;
  int energy = 0;
  
  if (n1>n2) { nl=n1; ns=n2;}
  else {nl=n2; ns=n1;}

  if (nl == 0)
    return P->stack[type][type_2];    /* stack */

  if (ns==0) {                       /* bulge */
    energy = (nl<=MAXLOOP)?P->bulge[nl]:
      (P->bulge[30]+(int)(P->lxc*log(nl/30.)));
    if (nl==1) energy += P->stack[type][type_2];
    else {
      if (type>2) energy += P->TerminalAU;
      if (type_2>2) energy += P->TerminalAU;
    }
    return energy;
  }
  else {                             /* interior loop */
    if (ns==1) { 
      if (nl==1)                     /* 1x1 loop */
	return P->int11[type][type_2][si1][sj1];
      if (nl==2) {                   /* 2x1 loop */
	if (n1==1)
	  energy = P->int21[type][type_2][si1][sq1][sj1];
	else
	  energy = P->int21[type_2][type][sq1][si1][sp1];
	return energy;
      }
    }
    else if (n1==2 && n2==2)         /* 2x2 loop */
      return P->int22[type][type_2][si1][sp1][sq1][sj1];
    { /* generic interior loop (no else here!)*/
      energy = (n1+n2<=MAXLOOP)?(P->internal_loop[n1+n2]):
	(P->internal_loop[30]+(int)(P->lxc*log((n1+n2)/30.)));
      
      energy += MIN2((int)MAX_NINIO, (nl-ns)*P->F_ninio[2]);
      
      energy += P->mismatchI[type][si1][sj1]+
	P->mismatchI[type_2][sq1][sp1];
    }
  }
  return energy;
}

//////////////////////////////////////////////////////////////////////////////////////
// alipf_fold
float 
Alifold::
alipf_fold(Array< Array<char> >& sequences, Array<char>& structure)
{
  enum {
    TURN = EP::TURN,
    MAXLOOP = EP::MAXLOOP
  };
#define Triloops EP::Triloops
#define Tetraloops EP::Tetraloops

  //short **S;
  //int *type;
  Array< Array<short> > S;
  Array<int> type;
  int s = 0;
  // int n = 0;
  int n_seq = 0;
  int i = 0;
  int j = 0;
  int k = 0;
  int l = 0;
  int ij = 0;
  int kl = 0;
  int u = 0;
  int u1 = 0;
  int d = 0;
  int ii = 0;
  int ll = 0;
  int type_2 = 0;
  int tt = 0;
  int ov = 0;
  double temp = 0.0;
  double Q = 0.0;
  double Qmax = 0.0;
  double prm_MLb = 0.0;
  double prmt = 0.0;
  double prmt1 = 0.0;
  //double qbt1, *tmp;
  double qbt1 = 0.0;
  Array<double> tmp;
   
  double free_energy = 0.0;
  double kTn = 0.0;

  //n = (int) strlen(sequences[0]);
  const int& n = sequences[0].size();
  //for (s=0; sequences[s]!=NULL; s++); 
  //n_seq = s;
  n_seq = sequences.size();
  init_alipf_fold(n, n_seq);  /* (re)allocate space */
  //kTn = (temperature+K0)*GASCONST*n_seq/10.;   /* kT in cal/mol  */
  kTn = (temperature+EP::K0)*EP::GASCONST*n_seq/10.;   /* kT in cal/mol  */
  
  //S = (short **) space(sizeof(short *)*(n_seq+1));
  S.resize(n_seq + 1);
  //type = (int *) space(n_seq*sizeof(int));
  type.resize(n_seq);
  for (s=0; s<n_seq; s++) { 
    //if (strlen(sequences[s]) != n) nrerror("uneqal seqence lengths");
    if ((int)sequences[s].size() != n) nrerror("uneqal seqence lengths");
    S[s] = encode_seq(sequences[s]);
  }
  //make_pscores((const short *const*)S, n_seq, structure);
  make_pscores_alipf_fold(S, n_seq, structure);
   
  /* array initialization ; qb,qm,q
     qb,qm,q (i,j) are stored as ((n+1-i)*(n-i) div 2 + n+1-j */

  for (d=0; d<=TURN; d++) 
    for (i=1; i<=n-d; i++) {
      j=i+d;
      ij = iindx[i]-j;
      q[ij]=1.0*scale[d+1];
      qb[ij]=qm[ij]=0.0;
    }

  for (i=1; i<=n; i++) 
    qq[i]=qq1[i]=qqm[i]=qqm1[i]=prm_l[i]=prm_l1[i]=prml[i]=0;

  for (j=TURN+2;j<=n; j++) {
    for (i=j-TURN-1; i>=1; i--) {
      int ij = 0;
      int psc = 0;
      /* construction of partition function for segment i,j */
      /* calculate pf given that i and j pair: qb(i,j)      */
      u = j-i-1; ij = iindx[i]-j;

      for (s=0; s<n_seq; s++) {
	type[s] = pair[S[s][i]][S[s][j]];
	if (type[s]==0) type[s]=7;
      }
      psc = pscore[ij];
      if (psc>=cv_fact*MINPSCORE) {   /* otherwise ignore this pair */
	
	/* hairpin contribution */
	for (qbt1=1,s=0; s<n_seq; s++) {
	  qbt1 *= exphairpin[u];
	  if ((tetra_loop)&&(u==4)) {
	    char tl[7]={};
	    char *ts = NULL;
	    strncpy(tl, &sequences[s][i-1], 6);
	    if ((ts=strstr(Tetraloops, tl)))
	      qbt1 *= exptetra[(ts-Tetraloops)/7];
	  } 
	  if (u==3) {
	    char tl[6]={0,0,0,0,0,0};
	    char *ts = NULL;
	    strncpy(tl, &sequences[s][i-1], 5);
	    if ((ts=strstr(Triloops, tl))) 
	      qbt1 *= expTriloop[(ts-Triloops)/6];
	    if (type[s]>2) 
	      qbt1 *= expTermAU;
	  }
	  else /* no mismatches for tri-loops */
	    qbt1 *= expmismatchH[type[s]][S[s][i+1]][S[s][j-1]];
	}
	qbt1 *= scale[u+2];

	/* interior loops with interior pair k,l */
	for (k=i+1; k<=MIN(i+MAXLOOP+1,j-TURN-2); k++) {
	  u1 = k-i-1;
	  for (l=MAX(k+TURN+1,j-1-MAXLOOP+u1); l<=j-1; l++) {
	    double qloop = 1;
	    if (qb[iindx[k]-l]==0) {qloop=0; continue;}
	    for (s=0; s<n_seq; s++) {
	      type_2 = pair[S[s][l]][S[s][k]]; if (type_2 == 0) type_2 = 7;
	      qloop *= expLoopEnergy(u1, j-l-1, type[s], type_2,
				     S[s][i+1], S[s][j-1],
				     S[s][k-1], S[s][l+1]);
	    }
	    qbt1 += qb[iindx[k]-l] * qloop * scale[u1+j-l-1+2];
	  }
	}
	/* multi-loop loop contribution */
	ii = iindx[i+1]; /* ii-k=[i+1,k-1] */
	temp = 0.0;
	for (k=i+2; k<=j-1; k++) temp += qm[ii-(k-1)]*qqm1[k];
	for (s=0; s<n_seq; s++) {
	  tt = rtype[type[s]];
	  temp *= expMLintern[tt]*expMLclosing*
	    expdangle3[tt][S[s][i+1]]*expdangle5[tt][S[s][j-1]];
	}
	qbt1 += temp*scale[2];
	qb[ij] = qbt1;
	qb[ij] *= exp(psc/kTn);
      } /* end if (type!=0) */
      else qb[ij] = 0.0;
       
      /* construction of qqm matrix containing final stem
	 contributions to multiple loop partition function
	 from segment i,j */
      qqm[i] = qqm1[i]*expMLbase[1];  /* expMLbase[1]^n_seq */
      for (qbt1=1, s=0; s<n_seq; s++) {
	qbt1 *= expMLintern[type[s]];
	if (i>1) qbt1 *= expdangle5[type[s]][S[s][i-1]];
	if (j<n) qbt1 *= expdangle3[type[s]][S[s][j+1]];
	else if (type[s]>2) qbt1 *= expTermAU;
      }
      qqm[i] += qb[ij]*qbt1;      
      /* construction of qm matrix containing multiple loop
	 partition function contributions from segment i,j */
      temp = 0.0;
      ii = iindx[i];  /* ii-k=[i,k-1] */
      for (k=i+1; k<=j; k++)
	temp += (qm[ii-(k-1)]+expMLbase[k-i])*qqm[k];
      qm[ij] = (temp + qqm[i]);
      
      /* auxiliary matrix qq for cubic order q calculation below */
      qbt1 = qb[ij];
      if (qbt1>0)
	for (s=0; s<n_seq; s++) {
	  if (i>1) qbt1 *= expdangle5[type[s]][S[s][i-1]];
	  if (j<n) qbt1 *= expdangle3[type[s]][S[s][j+1]];
	}
      qq[i] = qq1[i]*scale[1] + qbt1;
      
      /* construction of partition function for segment i,j */
      temp = 1.0*scale[1+j-i] + qq[i];
      for (k=i; k<=j-1; k++) temp += q[ii-k]*qq[k+1];
      q[ij] = temp;

      //#ifndef LARGE_PF
      if (temp>Qmax) {
	Qmax = temp;
	if (Qmax>FLT_MAX/10.)
	  fprintf(stderr, "%d %d %g\n", i,j,temp);
      }
      if (temp>FLT_MAX) {
	char msg[128] = {};
	sprintf(msg, "overflow in pf_fold while calculating q[%d,%d]\n"
		"use larger pf_scale", i,j);
	nrerror(msg);
      }
      //#endif
    }
    //tmp = qq1;  qq1 =qq;  qq =tmp;
    tmp.swap(qq1); qq1.swap(qq); qq.swap(tmp);
    //tmp = qqm1; qqm1=qqm; qqm=tmp;
    tmp.swap(qqm1); qqm1.swap(qqm); qqm.swap(tmp);
  }
  if (backtrack_type=='C')      Q = qb[iindx[1]-n];
  else if (backtrack_type=='M') Q = qm[iindx[1]-n];
  else Q = q[iindx[1]-n];

  /* ensemble free energy in Kcal/mol */
  if (Q<=FLT_MIN) fprintf(stderr, "pf_scale too large\n");
  //free_energy = (-log(Q)-n*log(pf_scale))*(temperature+K0)*GASCONST/1000.0;
  free_energy = (-log(Q)-n*log(pf_scale))*(temperature+EP::K0)*EP::GASCONST/1000.0;
  /* in case we abort because of floating point errors */ 
  if (n>1600) fprintf(stderr, "free energy = %8.2f\n", free_energy); 
      
  /* backtracking to construct binding probabilities of pairs*/
   
  if (do_backtrack) {
    Qmax=0;

    for (k = 1; k <= n; k++) {
      q1k[k] = q[iindx[1] - k];
      //qln[k] = q[iindx[k] - n];
      volatile double z = q[iindx[k] - n]; // to circumvent optimizer bug of gcc 4
      qln[k] = z;
    }
    q1k[0] = 1.0;
    qln[n+1] = 1.0;
      
    pr = q;     /* recycling */

    /* 1. exterior pair i,j and initialization of pr array */
    for (i=1; i<=n; i++) {
      for (j=i; j<=MIN(i+TURN,n); j++) pr[iindx[i]-j] = 0;
      for (j=i+TURN+1; j<=n; j++) {
	ij = iindx[i]-j;
	if (qb[ij]>0.) {
	  pr[ij] = q1k[i-1]*qln[j+1]/q1k[n] * exp(pscore[ij]/kTn);
	  for (s=0; s<n_seq; s++) {
	    int typ = 0;
	    typ = pair[S[s][i]][S[s][j]]; if (typ==0) typ=7;
	    if (i>1) pr[ij] *= expdangle5[typ][S[s][i-1]];
	    if (j<n) pr[ij] *= expdangle3[typ][S[s][j+1]];
	    else if (typ>2) pr[ij] *= expTermAU;
	  }
	} else
	  pr[ij] = 0;
      }
    }
      
    for (l=n; l>TURN+1; l--) {

      /* 2. bonding k,l as substem of 2:loop enclosed by i,j */
      for (k=1; k<l-TURN; k++) {
	double pp = 0;
	kl = iindx[k]-l;
	if (qb[kl]==0) continue;
	for (s=0; s<n_seq; s++) {
          type[s] = pair[S[s][l]][S[s][k]];
          if (type[s]==0) type[s]=7;
        }
	
	for (i=MAX(1,k-MAXLOOP-1); i<=k-1; i++) 
	  for (j=l+1; j<=MIN(l+ MAXLOOP -k+i+2,n); j++) {
	    ij = iindx[i] - j;
	    if ((pr[ij]>0)) {
	      double qloop=1;
	      for (s=0; s<n_seq; s++) {
		int typ = 0;
		typ = pair[S[s][i]][S[s][j]]; if (typ==0) typ=7;
		qloop *=  expLoopEnergy(k-i-1, j-l-1, typ, type[s],
				S[s][i+1], S[s][j-1], S[s][k-1], S[s][l+1]);
	      }
	      pp += pr[ij]*qloop*scale[k-i-1 + j-l-1 + 2];
	    } 
	  }
	pr[kl] += pp * exp(pscore[kl]/kTn);
      }
      /* 3. bonding k,l as substem of multi-loop enclosed by i,j */
      prm_MLb = 0.;
      if (l<n) for (k=2; k<l-TURN; k++) {
	i = k-1;
	prmt = prmt1 = 0.0;
	    
	ii = iindx[i];     /* ii-j=[i,j]     */
	ll = iindx[l+1];   /* ll-j=[l+1,j-1] */
	prmt1 = pr[ii-(l+1)];
	for (s=0; s<n_seq; s++) {
	  tt = pair[S[s][l+1]][S[s][i]]; if (tt==0) tt=7;
	  prmt1 *= expMLclosing*expMLintern[tt]*
	    expdangle3[tt][S[s][i+1]]*expdangle5[tt][S[s][l]];
	} 
	for (j=l+2; j<=n; j++) {
	  double pp=1;
	  if (pr[ii-j]==0) continue;
	  for (s=0; s<n_seq; s++) {
	    tt=pair[S[s][j]][S[s][i]]; if (tt==0) tt=7;
	    pp *=  expdangle3[tt][S[s][i+1]]*
	      expdangle5[tt][S[s][j-1]];
	  }
	  prmt +=  pr[ii-j]*pp*qm[ll-(j-1)];
	}
	kl = iindx[k]-l;
	for (s=0; s<n_seq; s++) {
	  int typ = 0;
	  typ=pair[S[s][k]][S[s][l]]; if (typ==0) typ=7;
	  prmt *= expMLclosing*expMLintern[typ];
	}
	prml[ i] = prmt;
	prm_l[i] = prm_l1[i]*expMLbase[1]+prmt1; /* expMLbase[1]^n_seq */
	
	prm_MLb = prm_MLb*expMLbase[1] + prml[i];
	/* same as:    prm_MLb = 0;
	   for (i=1; i<=k-1; i++) prm_MLb += prml[i]*expMLbase[k-i-1]; */
	
	prml[i] = prml[ i] + prm_l[i];
	
	if (qb[kl] == 0.) continue; 
	    
	temp = prm_MLb;
	
	for (i=1;i<=k-2; i++) 
	  temp += prml[i]*qm[iindx[i+1] - (k-1)];

	for (s=0; s<n_seq; s++) {
	  tt=pair[S[s][k]][S[s][l]]; if (tt==0) tt=7;
	  temp *= expMLintern[tt];
	  if (k>1) temp *= expdangle5[tt][S[s][k-1]];
	  if (l<n) temp *= expdangle3[tt][S[s][l+1]];
	}
	pr[kl] += temp * scale[2] * exp(pscore[kl]/kTn);
	
	//#ifndef LARGE_PF
	if (pr[kl]>Qmax) {
	  Qmax = pr[kl];
	  if (Qmax>FLT_MAX/10.)
	    fprintf(stderr, "%d %d %g %g\n", i,j,pr[kl],qb[kl]);
	}
	if (pr[kl]>FLT_MAX) {
	  ov++;
	  pr[kl]=FLT_MAX;
	}
	//#endif
      } /* end for (k=..) */
      tmp = prm_l1; prm_l1=prm_l; prm_l=tmp;

    }  /* end for (l=..)   */
    
    for (i=1; i<=n; i++)
      for (j=i+TURN+1; j<=n; j++) {
	ij = iindx[i]-j;
	pr[ij] *= qb[ij] *exp(-pscore[ij]/kTn);
      }

    //if (pi != NULL)
    //  *pi = make_pairinfo((const short **)S, n_seq);
  
    //if (structure!=NULL)
    //  sprintf_bppm(n, structure);
  }   /* end if (do_backtrack)*/

  //for (i=0; i<n_seq; i++) free(S[i]);
  //free(S);
  if (ov>0) fprintf(stderr, "%d overflows occurred while backtracking;\n"
		    "you might try a smaller pf_scale than %g\n",
		    ov, pf_scale);
  //free(type);
  free_alipf_arrays();
  return (free_energy); 

#undef TriLoops
#undef TetraLoops
}
void 
Alifold::
scale_pf_params(unsigned int length, int n_seq)
{
  /*------------------------------------------------------------------------*/
  /* dangling ends should never be destabilizing, i.e. expdangle>=1         */
  /* specific heat needs smooth function (2nd derivative)                   */
  /* we use a*(sin(x+b)+1)^2, with a=2/(3*sqrt(3)), b=Pi/6-sqrt(3)/2,       */
  /* in the interval b<x<sqrt(3)/2                                          */
#define SCALE 10
#define SMOOTH(X) ((X)/SCALE<-1.2283697)?0:(((X)/SCALE>0.8660254)?(X):\
          SCALE*0.38490018*(sin((X)/SCALE-0.34242663)+1)*(sin((X)/SCALE-0.34242663)+1))

  /* scale energy parameters and pre-calculate Boltzmann weights */
  unsigned int i = 0;
  unsigned int j = 0;
  unsigned int k = 0;
  unsigned int l = 0;
  double  kT = 0.0;
  double TT = 0.0;
  double kTn = 0.0;
  double  GT = 0.0;

  kT = (temperature+EP::K0)*EP::GASCONST;   /* kT in cal/mol  */
  kTn = kT*n_seq;
  TT = (temperature+EP::K0)/(EP::Tmeasure);
  
   /* scaling factors (to avoid overflows) */
  if (pf_scale==-1) { /* mean energy for random sequences: 184.3*length cal */
    pf_scale = exp(-(-185+(temperature-37.)*7.27)/kT);
    if (pf_scale<1) pf_scale=1;
  }
  scale[0] = 1.;
  for (i=1; i<=length; i++) {
    scale[i] = scale[i-1]/pf_scale;
  }

  /* loop energies: hairpins, bulges, interior, mulit-loops */
  for (i=0; i<=MIN(30u,length); i++) {
    GT =  EP::hairpin37[i]*TT;
    exphairpin[i] = exp( -GT*10./kTn);
  }
  for (i=0; i<=MIN(30u, (unsigned)EP::MAXLOOP); i++) {
    GT =  EP::bulge37[i]*TT;
    expbulge[i] = exp( -GT*10./kTn);
    GT =  EP::internal_loop37[i]*TT;
    expinternal[i] = exp( -GT*10./kTn);
  }
  /* special case of size 2 interior loops (single mismatch) */
  if (james_rule) expinternal[2] = exp ( -80*10/kTn);
   
  lxc = EP::lxc37*TT;
  for (i=31; i<length; i++) {
    GT = EP::hairpin37[30]*TT + (lxc*log( i/30.));
    exphairpin[i] = exp( -GT*10./kTn);
  }
  for (i=31; i<=MAXLOOP; i++) {
    GT = EP::bulge37[30]*TT + (lxc*log( i/30.));
    expbulge[i] = exp( -GT*10./kTn);
    GT = EP::internal_loop37[30]*TT + (lxc*log( i/30.));
    expinternal[i] = exp( -GT*10./kTn);
  }

  for (i=0; i<5; i++) {
    GT = EP::F_ninio37[i]*TT;
    for (j=0; j<=MAXLOOP; j++)
      expninio[i][j]=exp(-MIN((double)EP::MAX_NINIO,j*GT)*10/kTn);
  }
  for (i=0; (i*7)<strlen(Tetraloops); i++) {
    GT = EP::TETRA_ENTH37 - (EP::TETRA_ENTH37-EP::TETRA_ENERGY37[i])*TT;
    exptetra[i] = exp( -GT*10./kTn);
  }
  for (i=0; (i*5)<strlen(Triloops); i++) 
    expTriloop[i] = exp(-EP::Triloop_E37[i]*10/kTn);

  GT =  EP::ML_closing37*TT;
  expMLclosing = exp( -GT*10/kTn);

  for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
    GT =  EP::ML_intern37*TT;
    /* if (i>2) GT += TerminalAU; */
    expMLintern[i] = exp( -GT*10./kTn);
  }
  expTermAU = exp(-EP::TerminalAU*10/kTn);

  GT =  EP::ML_BASE37*TT;
  for (i=0; i<length; i++) {
    expMLbase[i] = exp( -10.*i*GT/kT)*scale[i];
  }

  /* if dangles==0 just set their energy to 0,
      don't let dangle energies become > 0 (at large temps),
      but make sure go smoothly to 0                        */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=4; j++) {
      GT = EP::dangle5_H[i][j] - (EP::dangle5_H[i][j] - EP::dangle5_37[i][j])*TT;
      expdangle5[i][j] = dangles?exp(SMOOTH(-GT)*10./kTn):1.;
      GT = EP::dangle3_H[i][j] - (EP::dangle3_H[i][j] - EP::dangle3_37[i][j])*TT;
      expdangle3[i][j] =  dangles?exp(SMOOTH(-GT)*10./kTn):1.;
      if (i>2) /* add TermAU penalty into dangle3 */
	expdangle3[i][j] *= expTermAU;
    }

  /* stacking energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++) {
      GT =  EP::enthalpies[i][j] - (EP::enthalpies[i][j] - EP::stack37[i][j])*TT;
      expstack[i][j] = exp( -GT*10/kTn);
    }

  /* mismatch energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
	GT = EP::mism_H[i][j][k] - (EP::mism_H[i][j][k] - EP::mismatchI37[i][j][k])*TT;
	expmismatchI[i][j][k] = exp(-GT*10.0/kTn);
	GT = EP::mism_H[i][j][k] - (EP::mism_H[i][j][k] - EP::mismatchH37[i][j][k])*TT;
	expmismatchH[i][j][k] = exp(-GT*10.0/kTn);
	GT = EP::mism_H[i][j][k] - (EP::mism_H[i][j][k] - EP::mismatchM37[i][j][k])*TT;
	expmismatchM[i][j][k] = exp(-GT*10.0/kTn);
      }

  /* interior lops of length 2 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  GT = EP::int11_H[i][j][k][l] -
	    (EP::int11_H[i][j][k][l] - EP::int11_37[i][j][k][l])*TT;
	  expint11[i][j][k][l] = exp(-GT*10./kTn);
	}
  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m = 0;
	  for (m=0; m<5; m++) {
	    GT = EP::int21_H[i][j][k][l][m] - 
	      (EP::int21_H[i][j][k][l][m] - EP::int21_37[i][j][k][l][m])*TT;
	    expint21[i][j][k][l][m] = exp(-GT*10./kTn);
	  }
	}
  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m = 0;
	  int n = 0;
	  for (m=0; m<5; m++)
	    for (n=0; n<5; n++) {            
	      GT = EP::int22_H[i][j][k][l][m][n] -
		(EP::int22_H[i][j][k][l][m][n]-EP::int22_37[i][j][k][l][m][n])*TT;
	      expint22[i][j][k][l][m][n] = exp(-GT*10./kTn);
	    }
	}  
#undef SCALE
#undef SMOOTH
}
double 
Alifold::
expLoopEnergy(int u1, int u2, int type, int type2,
	      short si1, short sj1, short sp1, short sq1) 
{
  double z=0;
  int no_close = 0;

  if ((no_closingGU) && ((type2==3)||(type2==4)||(type==2)||(type==4)))
    no_close = 1;

  if ((u1==0) && (u2==0)) /* stack */
    z = expstack[type][type2];
  else if (no_close==0) {
    if ((u1==0)||(u2==0)) { /* bulge */
      int u = 0;
      u = (u1==0)?u2:u1;
      z = expbulge[u];
      if (u2+u1==1) z *= expstack[type][type2];
      else {
	if (type>2) z *= expTermAU;
	if (type2>2) z *= expTermAU;
      }
    }
    else {     /* interior loop */
      if (u1+u2==2) /* size 2 is special */
	z = expint11[type][type2][si1][sj1];
      else if ((u1==1) && (u2==2)) 
	z = expint21[type][type2][si1][sq1][sj1];
      else if ((u1==2) && (u2==1))
	z = expint21[type2][type][sq1][si1][sp1];
      else if ((u1==2) && (u2==2))
	z = expint22[type][type2][si1][sp1][sq1][sj1];
      else {
	z = expinternal[u1+u2]*
	  expmismatchI[type][si1][sj1]*
	  expmismatchI[type2][sq1][sp1];
	z *= expninio[2][abs(u1-u2)];
      }
    }
  }
  return z;
}
void 
Alifold::
get_arrays_alipf_fold(unsigned int length)
{
  unsigned int size = 0;
  unsigned int i = 0;
   
  //size = sizeof(FLT_OR_DBL) * ((length+1)*(length+2)/2);
  //q   = (FLT_OR_DBL *) space(size);
  //qb  = (FLT_OR_DBL *) space(size);
  //qm  = (FLT_OR_DBL *) space(size);
  //pscore = (short *) space(sizeof(short)*((length+1)*(length+2)/2));
  //q1k = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  //qln = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //qq  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //qq1 = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //qqm  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //qqm1 = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //prm_l = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //prm_l1 =(FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //prml = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  //exphairpin = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  //expMLbase  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  //scale = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  //iindx = (int *) space(sizeof(int)*(length+1));

  size = ((length + 1) * (length + 2)) / 2;
  q.resize(size);
  qb.resize(size);
  qm.resize(size);
  pscore.resize(size);
  q1k.resize(length + 1);
  qln.resize(length + 2);
  qq.resize(length + 2);
  qq1.resize(length + 2);
  qqm.resize(length + 2);
  qqm1.resize(length + 2);
  prm_l.resize(length + 2);
  prm_l1.resize(length + 2);
  prml.resize(length + 2);
  exphairpin.resize(length + 1);
  expMLbase.resize(length + 1);
  scale.resize(length + 1);
  iindx.resize(length + 1);

  q.fill(0);
  qb.fill(0);
  qm.fill(0);
  pscore.fill(0);
  q1k.fill(0);
  qln.fill(0);
  qq.fill(0);
  qq1.fill(0);
  qqm.fill(0);
  qqm1.fill(0);
  prm_l.fill(0);
  prm_l1.fill(0);
  prml.fill(0);
  exphairpin.fill(0);
  expMLbase.fill(0);
  scale.fill(0);
  iindx.fill(0);
  
  for (i=1; i<=length; i++) {
    iindx[i] = ((length+1-i)*(length-i))/2 +length+1;
  }
}
void 
Alifold::
init_alipf_fold(int length, int n_seq)
{
  if (length<1) nrerror("init_pf_fold: length must be greater 0");
  if (init_length>0) free_alipf_arrays(); /* free previous allocation */
//#ifdef SUN4
//  nonstandard_arithmetic();
//#else
//#ifdef HP9
//  fpsetfastmode(1);
//#endif
//#endif
  make_pair_matrix();
  get_arrays_alipf_fold((unsigned) length);
  scale_pf_params((unsigned) length, n_seq);
  init_length=length;
}
void 
Alifold::
free_alipf_arrays()
{
  //  free(q);
  //  free(qb);
  //  free(qm);
  //  free(pscore);
  //  free(qq); free(qq1);
  //  free(qqm); free(qqm1);
  //  free(q1k); free(qln);
  //  free(prm_l); free(prm_l1); free(prml);
  //  free(exphairpin);
  //  free(expMLbase);
  //  free(scale);
  //  free(iindx);
  //#ifdef SUN4
  //  standard_arithmetic();
  //#else
  //#ifdef HP9
  //  fpsetfastmode(0);
  //#endif
  //#endif
  init_length=0;
}
//#define PMIN 0.0008
//int 
//Alifold::
//compare_pair_info(const void *pi1, const void *pi2) {
//  pair_info *p1, *p2;
//  int  i, nc1, nc2;
//  p1 = (pair_info *)pi1;  p2 = (pair_info *)pi2;
//  for (nc1=nc2=0, i=1; i<=6; i++) {
//    if (p1->bp[i]>0) nc1++;
//    if (p2->bp[i]>0) nc2++;
//  }
//  /* sort mostly by probability, add
//     epsilon * comp_mutations/(non-compatible+1) to break ties */
//  return (p1->p + 0.01*nc1/(p1->bp[0]+1.)) <
//         (p2->p + 0.01*nc2/(p2->bp[0]+1.)) ? 1 : -1;
//}
//Array<Alifold::pair_info>
//Alifold::
//make_pairinfo(const Array< Array<char> >& S, int n_seq) {
//  int i,j,n, num_p=0, max_p = 64;
//  //pair_info *pi;
//  //double *duck, p;
//  Array<pair_info> pi;
//  double p;
//  Array<double> duck;
//  n = S[0][0];
//  max_p = 64; 
//  //pi = space(max_p*sizeof(pair_info));
//  pi.resize(max_p);
//  pair_info null_pair_info = {};
//  pi.fill(null_pair_info);
//  //duck =  (double *) space((n+1)*sizeof(double));
//  duck.resize(n + 1);
//  duck.fill(0);
//  for (i=1; i<n; i++)
//    for (j=i+TURN+1; j<=n; j++) 
//      if ((p=pr[iindx[i]-j])>0) {
//	duck[i] -=  p * log(p);
//	duck[j] -=  p * log(p);
//      }
//
//  for (i=1; i<n; i++)
//    for (j=i+TURN+1; j<=n; j++) {
//      if ((p=pr[iindx[i]-j])>=PMIN) {
//	int type, s;
//	pi[num_p].i = i;
//	pi[num_p].j = j;
//	pi[num_p].p = p;
//	pi[num_p].ent =  duck[i]+duck[j]-p*log(p);
//	for (type=0; type<8; type++) pi[num_p].bp[type]=0;
//	for (s=0; s<n_seq; s++) {
//	  if (S[s][i]==0 && S[s][j]==0) type = 7; /* gap-gap  */  
//	  else type = pair[S[s][i]][S[s][j]];
//	  pi[num_p].bp[type]++;
//	}
//	num_p++;
//	if (num_p>=max_p) {
//	  max_p *= 2;
//	  //pi = realloc(pi, max_p * sizeof(pair_info));
//	  pi.resize(max_p);
//	  if (pi==NULL) nrerror("out of memory in alipf_fold");
//	}
//      }  
//    }
//  free(duck);
//  pi = realloc(pi, (num_p+1)*sizeof(pair_info));
//  pi[num_p].i=0;
//  qsort(pi, num_p, sizeof(pair_info), compare_pair_info );
//  return pi;
//}
//void 
//Alifold::
//sprintf_bppm(int length, char *structure)
//{
//  enum {L = 3};
//  extern char  bppm_symbol(float *x);
//  int    i,j;
//  float  P[L];   /* P[][0] unpaired, P[][1] upstream p, P[][2] downstream p */
//         
//  for( j=1; j<=length; j++ ) {
//    P[0] = 1.0;
//    P[1] = P[2] = 0.0;
//    for( i=1; i<j; i++) {
//      P[2] += pr[iindx[i]-j];    /* j is paired downstream */
//      P[0] -= pr[iindx[i]-j];    /* j is unpaired */
//    }
//    for( i=j+1; i<=length; i++ ) {
//      P[1] += pr[iindx[j]-i];    /* j is paired upstream */
//      P[0] -= pr[iindx[j]-i];    /* j is unpaired */
//    }
//    structure[j-1] = bppm_symbol(P);
//  }
//  structure[length] = '\0';
//}   
//short * 
//Alifold::
//encode_seq(const char *sequence) {
//  unsigned int i,l;
//  short *S;
//  l = strlen(sequence);
//  S = (short *) space(sizeof(short)*(l+1));
//  S[0] = (short) l;
//  
//  /* make numerical encoding of sequence */
//  for (i=1; i<=l; i++) 
//    S[i]= (short) encode_char(toupper(sequence[i-1]));
//  
//  return S;
//}
void 
Alifold::
make_pscores_alipf_fold(const Array< Array<short> >& S, int n_seq,
			const Array<char>& structure) 
{
  enum {
    TURN = EP::TURN
  };
  /* calculate co-variance bonus for each pair depending on  */
  /* compensatory/consistent mutations and incompatible seqs */
  /* should be 0 for conserved pairs, >0 for good pairs      */
#define NONE -10000 /* score for forbidden pairs */
  int n = 0;
  int i = 0;
  int j = 0;
  int k = 0;
  int l = 0;
  int s = 0;
  int score = 0;
  int dm[7][7]={{0,0,0,0,0,0,0}, /* hamming distance between pairs */
                {0,0,2,2,1,2,2} /* CG */,
                {0,2,0,1,2,2,2} /* GC */,
                {0,2,1,0,2,1,2} /* GU */,
                {0,1,2,2,0,2,1} /* UG */,
                {0,2,2,1,2,0,2} /* AU */,
                {0,2,2,2,1,2,0} /* UA */};

  n=S[0][0];  /* length of seqs */
  for (i=1; i<n; i++) {
    for (j=i+1; (j<i+TURN+1) && (j<=n); j++) 
      pscore[iindx[i]-j] = NONE;
    for (j=i+TURN+1; j<=n; j++) {
      int pfreq[8]={0,0,0,0,0,0,0,0};
      for (s=0; s<n_seq; s++) {
	int type = 0;
	if (S[s][i]==0 && S[s][j]==0) type = 7; /* gap-gap  */	
	else type = pair[S[s][i]][S[s][j]];
	
	pfreq[type]++;
      }
      if (pfreq[0]*2>n_seq) { pscore[iindx[i]-j] = NONE; continue;}
      for (k=1,score=0; k<=6; k++) /* ignore pairtype 7 (gap-gap) */
	for (l=k+1; l<=6; l++) 
	  /* scores for replacements between pairtypes    */
	  /* consistent mutations (if l==k+2) score 1     */
	  /* compensatory (all else) score 2              */
	  score += pfreq[k]*pfreq[l]*dm[k][l];
      /* counter examples score -1, gap-gap scores -0.25  */
      pscore[iindx[i]-j] = (int)(cv_fact *
	((UNIT*score)/n_seq - nc_fact*(UNIT*pfreq[0] + UNIT*pfreq[7]*0.25)));
    }
  }

  if (noLonelyPairs) /* remove unwanted pairs */
    for (k=1; k<=n-TURN-1; k++) 
      for (l=1; l<=2; l++) {
	int type = 0;
	int ntype = 0;
	int otype = 0;
	i=k; j = i+TURN+l;
	type = pscore[iindx[i]-j];
	while ((i>=1)&&(j<=n)) {
	  if ((i>1)&&(j<n)) ntype = pscore[iindx[i-1]-j-1];
	  if ((otype<cv_fact*MINPSCORE)&&(ntype<cv_fact*MINPSCORE))
	    /* too many counterexamples */
	    pscore[iindx[i]-j] = NONE; /* i.j can only form isolated pairs */
	  otype =  type;
	  type  = ntype;
	  i--; j++;
	}
      }
  

  //if (fold_constrained&&(structure!=NULL)) {
  if (fold_constrained && (!structure.empty())) {
    //int hx, *stack;
    int hx = 0;
    Array<int> stack;
    //stack = (int *) space(sizeof(int)*(n+1));
    stack.resize(n + 1);
    stack.fill(0);
    
    for(hx=0, j=1; j<=n; j++) {
      switch (structure[j-1]) {
      case 'x': /* j can't pair */ 
        for (l=1; l<j-TURN; l++) pscore[iindx[l]-j] = NONE;
        for (l=j+TURN+1; l<=n; l++) pscore[iindx[j]-j] = NONE;
        break;
      case '(':
        stack[hx++]=j;
        /* fallthrough */
      case '<': /* j pairs upstream */
        for (l=1; l<j-TURN; l++) pscore[iindx[l]-j] = NONE;
        break;
      case ')': /* j pairs with i */
        if (hx<=0) {
	  string structure1(structure.begin(), structure.end());
          fprintf(stderr, "%s\n", structure1.c_str());
          nrerror("unbalanced brackets in constraints");
        }
        i = stack[--hx];
	for (k=i+1; k<=n; k++) pscore[iindx[i]-k] = NONE;
        for (l=i+1; l<=j; l++) 
	  for (k=j; k<=n; k++) pscore[iindx[l]-k] = NONE;
	for (k=1; k<=i; k++) 
	  for (l=i; l<=j; l++) pscore[iindx[k]-l] = NONE;
	for (k=1; k<j; k++) pscore[iindx[k]-j] = NONE;
        if (pscore[iindx[i]-j]==NONE) 
	  pscore[iindx[i]-j] = 0;
        /* fallthrough */
      case '>': /* j pairs downstream */
        for (l=j+TURN+1; l<=n; l++) pscore[iindx[j]-l] = NONE;
        break;
      }
    }
    if (hx!=0) {
      string structure1(structure.begin(), structure.end());
      fprintf(stderr, "%s\n", structure1.c_str());
      nrerror("unbalanced brackets in constraint string");
    }
    //free(stack);
  }
#undef NONE
}
