/*
 * Alifold.hpp
 *
 *   Extracted from RNAalifold program of Vienna package (version 1.5)
 *   the author of the original code: Ivo L Hofacker
 *   Reference: 
 *    Secondary structure prediction for aligned RNA sequences.
 *    Ivo L Hofacker and  Martin Fekete and  Peter F Stadler
 *    J Mol Biol. 2002 vol 319 (18) 3724-32
 *
 *   modified to the present form by Hisanori Kiryu
 */
#ifndef ALIFOLD_HPP
#define ALIFOLD_HPP
#include "DPDeck.hpp"
#include "Profile.hpp"
#include "DPPath.hpp"
#include "ScoreParam.hpp"
class
Alifold
: public DPDeck
{
public:
  typedef UpTriMtx<float> PairProb;
  Alifold();
  ~Alifold();
  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setWeight(Weight& weights);
  void setGuideTree(GuideTree& guideTree) {}
  void setMSA(const MSA& msa);
  void setMSAs(const MSA& msaX, const MSA& msaY) {setMSA(msaX);}
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const PairProb& pairProbX,
	       const PairProb& pairProbY) {
    setMSA(msaX);
    pairProb = pairProbX;
  }
  void setDPRegion(const DPRegion& dpRegion) {}
  void setDPPath(const DPPath& dpPath0) {dpPath = dpPath0;}
  void run();
  float getScore() const {return score;}
  const DPPath& getDPPath() const {return dpPath;}

  void getPairProb(PairProb& pairProb) const;
  void getPairProb1(PairProb& pairProb) const;
  const char* getParamNameFromOffset(int offset) const {return "???";}
  string toString() const;
  void print() const;
  void printPairProb(ostream& fo) const;
  void printPairProb(ostream& fo, const float& threshold) const;
  void drawPairProb(ostream& fo, float threshold) const;

public:
  // private:
  const Vector<ScoreParam>* pScoreParamList;
  const ScoreParam* pScoreParam;
  const Weight* pWeight;
  const MSA* pMSA;
  Profile profile;
  float score;
  DPPath dpPath;
  Array< Array<char> > seqs;
  Array<char> structure;
  PairProb pairProb;

  void setDPPath();

  

  enum {
    MAXALPHA = 20,
    NBASES = 8,
    NBPAIRS = 7
  };
  enum {
    MAX_NUM_NAMES = 500,
    MAXSECTORS = 500, /* dimension for a backtrack array */

    LOCALITY = 0,     /* locality parameter for base-pairs */
    UNIT = 100,
    MINPSCORE = (-2 * UNIT),
    ISOLATED = 256,
    INF = 1000000,
    FORBIDDEN = 9999,
    BONUS = 10000,
    MAXLOOP = 30
  };
   struct paramT {
    int id;
    int stack[NBPAIRS+1][NBPAIRS+1];
    int hairpin[31];
    int bulge[MAXLOOP+1];
    int internal_loop[MAXLOOP+1];
    int mismatchI[NBPAIRS+1][5][5];
    int mismatchH[NBPAIRS+1][5][5];
    int mismatchM[NBPAIRS+1][5][5];
    int dangle5[NBPAIRS+1][5];
    int dangle3[NBPAIRS+1][5];
    int int11[NBPAIRS+1][NBPAIRS+1][5][5];
    int int21[NBPAIRS+1][NBPAIRS+1][5][5][5];
    int int22[NBPAIRS+1][NBPAIRS+1][5][5][5][5];
    int F_ninio[5];
    double lxc;
    int MLbase;
    int MLintern[NBPAIRS+1];
    int MLclosing;
    int TerminalAU;
    int TETRA_ENERGY[200];
    char Tetraloops[1401];
    int Triloop_E[40];
    char Triloops[241];
    double temperature;
  };
  struct bond {        /* base pair */
    int i;
    int j;
  };  
  typedef bond bondT;
  

  // fold_vars
  int  noGU;               /* GU not allowed at all */
  int  no_closingGU;       /* GU allowed only inside stacks */
  int  tetra_loop;         /* Fold with specially stable 4-loops */
  int  energy_set;         /* 0 = BP; 1=any mit GC; 2=any mit AU-parameter */
  int  dangles;	           /* use dangling end energies (not in part_func!) */
  char* nonstandards;      /* contains allowed non standard bases */
  double temperature;      /* rescale parameters to this temperature */
  int  james_rule;         /* interior loops of size 2 get energy 0.8Kcal and
			    * no mismatches, default 1 */
  int  logML;              /* use logarithmic multiloop energy function */
  int  cut_point;          /* first position of 2nd strand for co-folding */        
  int fold_constrained;    /* fold with constraints */
  int do_backtrack;        /* calculate pair prob matrix in part_func() */
  int noLonelyPairs;       /* avoid helices of length 1 */
  char backtrack_type;     /* usually 'F'; 'C' require (1,N) to be bonded;
			    * 'M' seq is part of a multi loop */
  double cv_fact /* =1 */;
  double nc_fact /* =1 */;

  void nrerror(const char* message) {fprintf(stderr, message); exit(EXIT_FAILURE);}


  short alias[MAXALPHA + 1];
  int pair[MAXALPHA + 1][MAXALPHA + 1];
  int rtype[MAXALPHA + 1];/* rtype[pair[i][j]]:=pair[j][i] */
  paramT param;
  int param_id;
  void make_pair_matrix();
  void update_fold_params();
  paramT *scale_parameters();

  // alifold.c
  paramT* P;
  int init_length;
  Array<int> indx;
  Array<int> c;       /* energy array, given that i-j pair */
  Array<int> cc;      /* linear array for calculating canonical structures */
  Array<int> cc1;     /*   "     "        */
  Array<int> f5;      /* energy of 5' end */
  Array<int> fML;     /* multi-loop auxiliary energy array */

  Array<int> Fmi;     /* holds row i of fML (avoids jumps in memory) */
  Array<int> DMLi;    /* DMLi[j] holds MIN(fML[i,k]+fML[k+1,j])  */
  Array<int> DMLi1;   /*             MIN(fML[i+1,k]+fML[k+1,j])  */
  Array<int> DMLi2;   /*             MIN(fML[i+2,k]+fML[k+1,j])  */
  Array<int> pscore;  /* precomputed array of pair types */ 
  Array<bond> base_pair; /* list of base pairs */
 

  template <typename T>
  inline static T MIN2(const T& A, const T& B) {return std::min(A, B);}
  //float  alifold(char **strings, char *structure);
  float  alifold(Array< Array<char> >& strings, Array<char>& structure);
  void   init_alifold(int length);
  //void   free_alifold_arrays(void);
  void   free_alifold_arrays();
  //void  parenthesis_structure(char *structure, int length);
  void  parenthesis_structure(Array<char>& structure, int length);
  void  get_arrays_alifold(unsigned int size);
  //void  make_pscores_alifold(const short *const *S, int n_seq,
  //			     const char *structure);
  void  make_pscores_alifold(const Array< Array<short> >& S, int n_seq,
			     const Array<char>& structure);
  int encode_char(char c);
  //short *encode_seq(const char *sequence);
  Array<short> encode_seq(const Array<char>& sequence);
  int LoopEnergy(int n1, int n2, int type, int type_2,
		 int si1, int sj1, int sp1, int sq1);
  int HairpinE(int size, int type, int si1, int sj1, const char *str);


  // alipf_fold
  double lxc;
  double expMLclosing;
  double expMLintern[NBPAIRS+1];
  Array<double> expMLbase;
  double expTermAU;
  double expTriloop[40];
  double exptetra[40];
  Array<double> exphairpin;
  double expbulge[MAXLOOP+1];
  double expinternal[MAXLOOP+1];
  double expninio[5][MAXLOOP+1];
  double expdangle5[NBPAIRS+1][5];
  double expdangle3[NBPAIRS+1][5];
  double expstack[NBPAIRS+1][NBPAIRS+1];
  double expmismatchI[NBPAIRS+1][5][5];
  double expmismatchH[NBPAIRS+1][5][5];
  double expmismatchM[NBPAIRS+1][5][5];
  double expint11[NBPAIRS+1][NBPAIRS+1][5][5];
  double expint21[NBPAIRS+1][NBPAIRS+1][5][5][5];
  double expint22[NBPAIRS+1][NBPAIRS+1][5][5][5][5];

  double pf_scale;         /* scaling factor to avoid float overflows*/
  Array<int> iindx;              /* pr[i,j] -> pr[iindx[i]-j] */
  Array<double> pr;              /* base pairing prob. matrix */
  Array<double> q;
  Array<double> qb;
  Array<double> qm;
  Array<double> qqm;
  Array<double> qqm1;
  Array<double> qq;
  Array<double> qq1;
  Array<double> prml;
  Array<double> prm_l;
  Array<double> prm_l1;
  Array<double> q1k;
  Array<double> qln;
  Array<double> scale;

  template <typename T>
  inline static T MIN(const T& A, const T& B) {return std::min(A, B);}
  template <typename T>
  inline static T MAX(const T& A, const T& B) {return std::max(A, B);}
  //float alipf_fold(char **sequences, char *structure, pair_info **pi);
  float alipf_fold(Array< Array<char> >& strings, Array<char>& structure);
  void  init_alipf_fold(int length, int n_seq);
  void  free_alipf_arrays(void);
  /* void  update_alipf_params(int length); */
  //void  sprintf_bppm(int length, char *structure);
  //void  sprintf_bppm(int length, Array<char>& structure);
  void  scale_pf_params(unsigned int length, int n_seq);
  void  get_arrays_alipf_fold(unsigned int length);
  double expLoopEnergy(int u1, int u2, int type, int type2,
		       short si1, short sj1, short sp1, short sq1);
  //void make_pscores_alipf_fold(const short *const *S, int n_seq,
  //			       const char *structure);
  void  make_pscores_alipf_fold(const Array< Array<short> >& S, int n_seq,
				const Array<char>& structure);
  //pair_info *make_pairinfo(const short *const* S, int n_seq);
  //Array<pair_info> make_pairinfo(const Array< Array<char> >& S, int n_seq);
  //short * encode_seq(const char *sequence);


  class EnergyParam;
  typedef EnergyParam EP;
};
class
Alifold::
EnergyParam
{
public:
  enum {
    TURN = 3,
    MAXLOOP = 30,
    STACK_BULGE = 1, /* stacking energies for bulges of size 1 */
    STACK_BULGE1 = 1,
    NEW_NINIO = 1,   /* new asymetry penalty */
    NST = 0,        /* Energy for nonstandard stacked pairs */
    DEF = -50,      /* Default terminal mismatch, used for I */
    NSM = 0         /* terminal mismatch for non standard pairs */
  };
  static const double GASCONST;
  static const double K0;
  static const double Tmeasure;   /* temperature of param measurements */
  static const double lxc37;      /* parameter for logarithmic loop
			     * energy extrapolation            
			     */
  /* constants for linearly destabilizing contributions for multi-loops
   * F = ML_closing + ML_intern*(k-1) + ML_BASE*u  
   */
  static const int ML_BASE37;
  static const int ML_closing37;
  static const int ML_intern37;
  /* penalty for helices terminated by AU (actually not GC) */
  static const int TerminalAU;
  /* Ninio-correction for asymmetric internal loops with branches n1 and n2 */
  /*    ninio_energy = min{max_ninio, |n1-n2|*F_ninio[min{4.0, n1, n2}] } */
  static const int MAX_NINIO;                   /* maximum correction */
  static const int F_ninio37[5];
  static const char Triloops[241];   /* str containing the special triloops */
  static const int Triloop_E37[40]; /* Bonus energy for special Triloops */  
  /* stabilizing contribution due to special hairpins of size 4 (tetraloops) */
  static const char Tetraloops[1400];  /* str containing the special tetraloops */
  static const int TETRA_ENERGY37[200];  /* Bonus energy for special tetraloops */
  static const int TETRA_ENTH37;
  static const int stack37[NBPAIRS+1][NBPAIRS+1];
  static const int enthalpies[NBPAIRS+1][NBPAIRS+1]; /* stack enthalpies */
  static const int entropies[NBPAIRS+1][NBPAIRS+1];  /* not used anymore */
  static const int hairpin37[31];
  static const int bulge37[31];
  static const int internal_loop37[31];
  static const int oldhairpin37[31];
  static const int oldbulge37[31];
  static const int oldinternal_loop37[31];
  //static int internal2_energy;
  static const int old_mismatch_37[NBPAIRS+1][5][5];
  static const int mismatchI37[NBPAIRS+1][5][5];  /* interior loop mismatches */
  static const int mismatchH37[NBPAIRS+1][5][5];  /* same for hairpins */
  static const int mismatchM37[NBPAIRS+1][5][5];  /* same for multiloops */
  static const int mism_H[NBPAIRS+1][5][5];       /* mismatch enthalpies */
  static const int dangle5_37[NBPAIRS+1][5];      /* 5' dangle exterior of pair */
  static const int dangle3_37[NBPAIRS+1][5];      /* 3' dangle */
  static const int dangle3_H[NBPAIRS+1][5];       /* corresponding enthalpies */
  static const int dangle5_H[NBPAIRS+1][5];
  static const int int11_37[NBPAIRS+1][NBPAIRS+1][5][5]; /* 1x1 interior loops */
  static const int int11_H[NBPAIRS+1][NBPAIRS+1][5][5]; 
  static const int int21_37[NBPAIRS+1][NBPAIRS+1][5][5][5]; /* 2x1 interior loops */
  static const int int21_H[NBPAIRS+1][NBPAIRS+1][5][5][5];
  static const int int22_37[NBPAIRS+1][NBPAIRS+1][5][5][5][5]; /* 2x2 interior loops */
  static const int int22_H[NBPAIRS+1][NBPAIRS+1][5][5][5][5];
  static string toString() {
    ostringstream os;
    os << "---\n"
       << "CHARS: \"_ACGUXKI\"\n"
       << "BP_pair:\n"
       << "- [ 0, 0, 0, 0, 0, 0, 0, 0 ]\n"// _
       << "- [ 0, 0, 0, 0, 5, 0, 0, 5 ]\n"// A
       << "- [ 0, 0, 0, 1, 0, 0, 0, 0 ]\n"// C
       << "- [ 0, 0, 2, 0, 3, 0, 0, 0 ]\n"// G
       << "- [ 0, 6, 0, 4, 0, 0, 0, 6 ]\n"// U
       << "- [ 0, 0, 0, 0, 0, 0, 2, 0 ]\n"// X
       << "- [ 0, 0, 0, 0, 0, 1, 0, 0 ]\n"// K
       << "- [ 0, 6, 0, 0, 5, 0, 0, 0 ]\n"// I
       << "BP_pair_1:\n"
       << "  CG: 1\n"
       << "  KX: 1\n"
       << "  GC: 2\n"
       << "  XK: 2\n"
       << "  GU: 3\n"
       << "  UG: 4\n"
       << "  AU: 5\n"
       << "  AI: 5\n"
       << "  IU: 5\n"
       << "  UA: 6\n"
       << "  UI: 6\n"
       << "  IA: 6\n"
       << "NBASES: " << NBASES << "\n"
       << "NBPAIRS: " << NBPAIRS << "\n"
       << "MAXLOOP: " << MAXLOOP << "\n"
       << "UNIT: " << UNIT << "\n"
       << "TURN: " << TURN << "\n"
       << "MAXLOOP: " << MAXLOOP << "\n"
       << "STACK_BULGE: " << STACK_BULGE << "\n"
       << "STACK_BULGE1: " << STACK_BULGE1 << "\n"
       << "NST: " << NST << "\n"
       << "DEF: " << DEF << "\n"
       << "NSM: " << NSM << "\n"
       << "GASCONST: " << GASCONST << "\n"
       << "K0: " << K0 << "\n"
       << "Tmeasure: " << Tmeasure << "\n"
       << "lxc37: " << lxc37 << "\n"
       << "ML_BASE37: " << ML_BASE37 << "\n"
       << "ML_closing37: " << ML_closing37 << "\n"
       << "ML_intern37: " << ML_intern37 << "\n"
       << "TerminalAU: " << TerminalAU << "\n"
       << "MAX_NINIO: " << MAX_NINIO << "\n";
    os << "F_ninio37:\n";
    for (int i = 0; i < 5; i++) {
      os << "- " << F_ninio37[i] << "\n";
    }
    os << "Triloops:\n";// no entry
    os << "Tetraloops:\n";
    Vector<string> tokens = tokenize<string>(string(Tetraloops));
    for (int i = 0; i < (int)tokens.size(); i++) {
      os << "- [ " << tokens[i] << ", " << TETRA_ENERGY37[i] << " ]\n";
    }
    os << "stack37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < (NBPAIRS+1); j++) {
	os << (j==0 ? "" : "  ") << "- " 
	   << stack37[i][j] << "\n";
      }
    }
    // os << "enthalpies:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < (NBPAIRS+1); j++) {
    // 	os << "  - " << enthalpies[i][j] << "\n";
    //   }
    // }
    // os << "entropies:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < (NBPAIRS+1); j++) {
    // 	os << "  - " << entropies[i][j] << "\n";
    //   }
    // }
    os << "hairpin37:\n";
    for (int i = 0; i < 31; i++) {
      os << "- " << hairpin37[i] << "\n";
    }
    os << "bulge37:\n";
    for (int i = 0; i < 31; i++) {
      os << "- " << bulge37[i] << "\n";
    }
    os << "internal_loop37:\n";
    for (int i = 0; i < 31; i++) {
      os << "- " << internal_loop37[i] << "\n";
    }
    os << "mismatchI37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < 5; j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- " 
	     << mismatchI37[i][j][k] << "\n";
	}
      }
    }
    os << "mismatchH37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < 5; j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- " 
	     << mismatchH37[i][j][k] << "\n";
	}
      }
    }
    os << "mismatchM37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < 5; j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- " 
	     << mismatchM37[i][j][k] << "\n";
	}
      }
    }
    // os << "mism_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < 5; j++) {
    // 	os << "  -\n";
    // 	for (int k = 0; k < 5; k++) {
    // 	  os << "    - " << mism_H[i][j][k] << "\n";
    // 	}
    //   }
    // }
    os << "dangle5_37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < 5; j++) {
	os << (j==0 ? "" : "  ") << "- " 
	   << dangle5_37[i][j] << "\n";
      }
    }
    os << "dangle3_37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < 5; j++) {
	os << (j==0 ? "" : "  ") << "- " 
	   << dangle3_37[i][j] << "\n";
      }
    }
    // os << "dangle3_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < 5; j++) {
    // 	os << "  - " << dangle3_H[i][j] << "\n";
    //   }
    // }
    // os << "dangle5_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < 5; j++) {
    // 	os << "  - " << dangle5_H[i][j] << "\n";
    //   }
    // }
    os << "int11_37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < (NBPAIRS+1); j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- ";
	  for (int l = 0; l < 5; l++) {
	    os << (l==0 ? "" : "      ") << "- " 
	       << int11_37[i][j][k][l] << "\n";
	    Check(int11_37[i][j][k][l] == int11_37[j][i][l][k]);
	  }
	}
      }
    }
    // os << "int11_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < (NBPAIRS+1); j++) {
    // 	os << "  -\n";
    // 	for (int k = 0; k < 5; k++) {
    // 	  os << "    -\n";
    // 	  for (int l = 0; l < 5; l++) {
    // 	    os << "      - " << int11_H[i][j][k][l] << "\n";
    // 	  }
    // 	}
    //   }
    // }
    os << "int21_37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < (NBPAIRS+1); j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- ";
	  for (int l = 0; l < 5; l++) {
	    os << (l==0 ? "" : "      ") << "- ";
	    for (int m = 0; m < 5; m++) {
	      os << (m==0 ? "" : "        ") << "- " 
		 << int21_37[i][j][k][l][m] << "\n";
	    }
	  }
	}
      }
    }
    // os << "int21_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < (NBPAIRS+1); j++) {
    // 	os << "  -\n";
    // 	for (int k = 0; k < 5; k++) {
    // 	  os << "    -\n";
    // 	  for (int l = 0; l < 5; l++) {
    // 	    os << "      -\n";
    // 	    for (int m = 0; m < 5; m++) {
    // 	      os << "        - " << int21_H[i][j][k][l][m] << "\n";
    // 	    }
    // 	  }
    // 	}
    //   }
    // }
    os << "int22_37:\n";
    for (int i = 0; i < (NBPAIRS+1); i++) {
      os << "- ";
      for (int j = 0; j < (NBPAIRS+1); j++) {
	os << (j==0 ? "" : "  ") << "- ";
	for (int k = 0; k < 5; k++) {
	  os << (k==0 ? "" : "    ") << "- ";
	  for (int l = 0; l < 5; l++) {
	    os << (l==0 ? "" : "      ") << "- ";
	    for (int m = 0; m < 5; m++) {
	      os << (m==0 ? "" : "        ") << "- ";
	      for (int n = 0; n < 5; n++) {
		os << (n==0 ? "" : "          ") << "- " 
		   << int22_37[i][j][k][l][m][n] << "\n";
	      }
	    }
	  }
	}
      }
    }
    // os << "int22_H:\n";
    // for (int i = 0; i < (NBPAIRS+1); i++) {
    //   os << "-\n";
    //   for (int j = 0; j < (NBPAIRS+1); j++) {
    // 	os << "  -\n";
    // 	for (int k = 0; k < 5; k++) {
    // 	  os << "    -\n";
    // 	  for (int l = 0; l < 5; l++) {
    // 	    os << "      -\n";
    // 	    for (int m = 0; m < 5; m++) {
    // 	      os << "        -\n";
    // 	      for (int n = 0; n < 5; n++) {
    // 		os << "          - " << int22_H[i][j][k][l][m][n] << "\n";
    // 	      }
    // 	    }
    // 	  }
    // 	}
    //   }
    // }
    return os.str();
  }
};
#endif
