/*
 * TBTree.cpp
 *
 */
#include "EScript.hpp"
#include "DPDeckPSCFGTBTree.hpp"
#include "TreeIterator.hpp"

DPDeckPSCFG::TBTree::
TBTree()
  : head(NULL),
    left(NULL),
    right(NULL),
    state_h(Model::X_st),
    state_l(Model::X_st),
    state_r(Model::X_st)
{
}
DPDeckPSCFG::TBTree::
TBTree(const TBTree& tbTree)
  : head(NULL),
    left(NULL),
    right(NULL),
    state_h(Model::X_st),
    state_l(Model::X_st),
    state_r(Model::X_st)
{
  *this = tbTree;
}
DPDeckPSCFG::TBTree::
TBTree(const Vector<Link>& treeNodes, const Vector<State>& states)
  : head(NULL),
    left(NULL),
    right(NULL),
    state_h(Model::X_st),
    state_l(Model::X_st),
    state_r(Model::X_st)
{
  reset(treeNodes, states);
}
DPDeckPSCFG::TBTree::
~TBTree()
{
  clear();
}
DPDeckPSCFG::TBTree&
DPDeckPSCFG::TBTree::
operator=(const TBTree& tbTree)
{
  clear();
  if (tbTree.empty()) return *this;

  state_h = tbTree.state_h;
  state_l = tbTree.state_l;
  state_r = tbTree.state_r;

 typedef TreeIterator<TBTree, TBTree::Link> Iterator;
  Iterator it(tbTree, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink tlink = ilink->tlink;
    TBTree::Link& data = ilink->getNodeData();
    data = new TBTree::Node();
    data->rect = tlink->rect;
    data->state = tlink->state;

    if (tbTree.isLeftEnd(tlink)) {
      left = data;
      data->leftChild = tlink->leftChild;
      data->rightChild = tlink->rightChild;
      it.prune();
    } else if (tbTree.isRightEnd(tlink)) {
      right = data;
      data->leftChild = tlink->leftChild;
      data->rightChild = tlink->rightChild;
      it.prune();
    } else {
      if (tlink->hasChild(0)) {
	const Iterator::Link child1 = ilink->getChild(0);
	TBTree::Link& data1 = child1->getNodeData();
	data->leftChild = data1;
	data1->parent = data;
      }
      if (tlink->hasChild(1)) {
	const Iterator::Link child2 = ilink->getChild(1);
	TBTree::Link& data2 = child2->getNodeData();
	data->rightChild = data2;
	data2->parent = data;
      }
    }
  }
  const Iterator::Link link = it.getRoot();
  const TBTree::Link data = link->getNodeData();
  head = data;
  return *this;
}
string
DPDeckPSCFG::TBTree::
toString() const
{
  ostringstream oss;
  if (empty()) return oss.str();

  Vector<Link> arr;
  Vector<int> vDepth;
  arr.push_back(head);
  vDepth.push_back(0);
  while (!arr.empty()) {
    Link link = arr.back(); arr.pop_back();
    const int& depth = vDepth.back(); vDepth.pop_back();
    while (link->getNChild() == 1) {
      oss << string(2 * depth, ' ');
      oss << link->toString() << "\n";
      link = link->leftChild;
    } 
    Assert(Model::bifurcates(link->state) || link->getNChild() == 0);
    oss << string(2 * depth, ' ');
    oss << link->toString() << "\n";
    if (link == left) {
      oss << string(2 * depth, ' ');
      oss << "left: " << Model::state2str(state_l) << "\n";
    } else if (link == right) {
      oss << string(2 * depth, ' ');
      oss << "right: " << Model::state2str(state_r) << "\n";
    }

    if (Model::bifurcates(link->state)) {
      arr.push_back(link->rightChild);
      arr.push_back(link->leftChild);
      const int& depth1 = (depth + 1);
      vDepth.push_back(depth1);
      vDepth.push_back(depth1);
    }
    oss << "\n";
  }
  return oss.str();
}
void
DPDeckPSCFG::TBTree::
print() const
{
  cout << toString() << flush;
}
void
DPDeckPSCFG::TBTree::
clear()
{
  if (head) {
    if (left) {
      left->leftChild = NULL;
      left->rightChild = NULL;
    }
    if (right) {
      right->leftChild = NULL;
      right->rightChild = NULL;
    }
    head->deleteR();
  }
  head = NULL;
  left = NULL;
  right = NULL;
  state_h = Model::X_st;
  state_l = Model::X_st;
  state_r = Model::X_st;
}
void
DPDeckPSCFG::TBTree::
reset(const Vector<Link>& treeNodes, const Vector<State>& states)
{
  Assert(treeNodes.size() == 3);
  Assert(states.size() == 3);
  Assert(Model::isValidState(states[0]));
  Assert(Model::isValidState(states[1]));
  Assert(Model::isValidState(states[2]));

  clear();
  head = treeNodes[0];
  left = treeNodes[1];
  right = treeNodes[2];

  state_h = states[0];
  state_l = states[1];
  state_r = states[2];
}

// currently DPPath::update method is not supported
//DPPath
//DPDeckPSCFG::TBTree::
//getDPPath() const
//{
//  DPPath dpPath;
//  if (empty()) return dpPath;
//
//  Rect rect0 = head->rect;
//  dpPath.setDiagonalPath(rect0.lengthX() - 1, rect0.lengthY() - 1);
//  dpPath = getDPPath(dpPath);
//  return dpPath;
//}
//DPPath
//DPDeckPSCFG::TBTree::
//getDPPath(const DPPath& dpPath0) const
//{
//  DPPath dpPath(dpPath0);
//  if (empty()) return dpPath;
//  
//  Vector<DPPath> paths;
//  Point prev_pt(Point::null());
//  AlnFlagments af = head->getAlnFlagmentsR();
//  for (AlnFlagments::const_iterator it = af.begin();
//       it != af.end(); it++) {
//    const AlnElem* curr = (*it);
//    const Point& curr_pt = curr->pt;
//    if (!(prev_pt == curr_pt)) {
//      paths.push_back(DPPath());
//    }
//    DPPath& dpPath1 = paths.back();
//    
//    prev_pt = curr_pt;
//    switch (curr->state) {
//    case M_st:
//      dpPath1.pushMState(curr_pt.x, curr_pt.y);
//      prev_pt.x++;
//      prev_pt.y++;
//      break;
//    case I_st:
//      dpPath1.pushIState(curr_pt.y);
//      prev_pt.y++;
//      break;
//    case D_st:
//      dpPath1.pushDState(curr_pt.x);
//      prev_pt.x++;
//      break;
//    default:
//      Die("bad state=%s", curr->state);
//      break;
//    }
//  }
//  for (int i = 0; i < (int)paths.size(); i++) {
//    dpPath.update(paths[i]);
//  }
//  for (AlnFlagments::const_iterator it = af.begin();
//       it != af.end(); it++) {
//    const AlnElem* curr = (*it);
//    if (curr->pair == NULL) continue;
//
//    const int& a1 = dpPath.getAlnPos(curr->pt);
//    const int& a2 = dpPath.getAlnPos(curr->pair->pt);
//    Assert(0 <= a1 && a1 < a2 && a2 < (int)dpPath.getAlignLen());
//    dpPath.ct.addPair(a1, a2);
//  }
//  dpPath.testConsistency();
//  for (AlnFlagments::iterator it = af.begin(); it != af.end(); it++) {
//    delete (*it);
//  }
//  return dpPath;
//}

DPPath
DPDeckPSCFG::TBTree::
getDPPath() const
{
  DPPath dpPath;
  if (empty()) return dpPath;
  
  AlnFlagments af = head->getAlnFlagmentsR();
  for (AlnFlagments::const_iterator it = af.begin();
       it != af.end(); it++) {
    const AlnElem* curr = (*it);
    const Point& curr_pt = curr->pt;    

    switch (curr->state) {
    case M_st:
      dpPath.pushMState(curr_pt.x, curr_pt.y);
      break;
    case I_st:
      dpPath.pushIState(curr_pt.y);
      break;
    case D_st:
      dpPath.pushDState(curr_pt.x);
      break;
    default:
      Die("bad state=%s", curr->state);
      break;
    }
  }
  for (AlnFlagments::const_iterator it = af.begin();
       it != af.end(); it++) {
    const AlnElem* curr = (*it);
    if (curr->pair == NULL) continue;

    const int& a1 = dpPath.getAlnPos(curr->pt);
    const int& a2 = dpPath.getAlnPos(curr->pair->pt);
    Assert(0 <= a1 && a1 < a2 && a2 < (int)dpPath.getAlignLen());
    dpPath.ct.addPair(a1, a2);
  }
  dpPath.testConsistency();
  for (AlnFlagments::iterator it = af.begin(); it != af.end(); it++) {
    delete (*it);
  }
  return dpPath;
}
bool
DPDeckPSCFG::TBTree::
isLeftEnd(const Node*const& link) const
{
  return (left == link);
}
bool
DPDeckPSCFG::TBTree::
isRightEnd(const Node*const& link) const
{
  return (right == link);
}
DPDeckPSCFG::State
DPDeckPSCFG::TBTree::
getStateLeft() const
{
  return state_l;
}
DPDeckPSCFG::State
DPDeckPSCFG::TBTree::
getStateRight() const
{
  return state_r;
}
void
DPDeckPSCFG::TBTree::
replaceNode(Link& link, const Vector<Link>& treeNodes)
{
  Assert(treeNodes.size() == 3);
  Assert(treeNodes[0] != NULL);
  treeNodes[0]->testConsistencyR();
  //testConsistency(); //not satisfied

  const Link& head1 = treeNodes[0];
  const Link& left1 = (treeNodes[1] ? treeNodes[1] : treeNodes[0]);
  const Link& right1 = (treeNodes[2] ? treeNodes[2] : treeNodes[0]);

  if (head == link) {
    head = head1;
  }
  if (left == link) {
    left = left1;
  }
  if (right == link) {
    right = right1;
  }
  // be careful about the differences between left1 and treeNodes[1] etc.
  //if (link->parent) {
  //  //treeNodes[0]->parent = link->parent; wrong
  //  treeNodes[0]->parent->replaceChild(link, treeNodes[0]);
  //}
  //testConsistency();
  //if (link->leftChild) {
  //  if (treeNodes[1]) {
  //    //treeNodes[1]->leftChild = link->leftChild;
  //    treeNodes[1]->leftChild->parent = treeNodes[1];
  //  } else {
  //    Assert(treeNodes[2]);
  //    Assert(link->rightChild == NULL);
  //    //treeNodes[2]->leftChild = link->leftChild;
  //    treeNodes[2]->leftChild->parent = treeNodes[2];
  //  }
  //}
  //testConsistency();
  //if (link->rightChild) {
  //  Assert(treeNodes[1]);
  //  Assert(treeNodes[2]);
  //  Assert(link->leftChild);
  //  //treeNodes[2]->leftChild = link->rightChild;
  //  treeNodes[2]->leftChild->parent = treeNodes[2];
  //}
  testConsistency();
}
void
DPDeckPSCFG::TBTree::
testConsistency() const
{
#ifdef DEBUG
  if (!empty()) {
    head->testConsistencyR();
  }
#endif
}

DPDeckPSCFG::TBTree::Node::
Node()
  : parent(NULL),
    leftChild(NULL),
    rightChild(NULL),
    rect(),
    state(Model::X_st)
{
}
DPDeckPSCFG::TBTree::Node::
Node(const Link& parent,
     const Link& leftChild,
     const Link& rightChild,
     const Rect& rect,
     const State& state)
  : parent(parent),
    leftChild(leftChild),
    rightChild(rightChild),
    rect(rect),
    state(state)
{
}
DPDeckPSCFG::TBTree::Node::
Node(const Node& node)
  : parent(node.parent),
    leftChild(node.leftChild),
    rightChild(node.rightChild),
    rect(node.rect),
    state(node.state)
{
}
DPDeckPSCFG::TBTree::Node::
~Node()
{
}
DPDeckPSCFG::TBTree::Node&
DPDeckPSCFG::TBTree::Node::
operator=(const Node& node)
{
  parent = node.parent;
  leftChild = node.leftChild;
  rightChild = node.rightChild;
  rect = node.rect;
  state = node.state;
  return *this;
}
string
DPDeckPSCFG::TBTree::Node::
toString() const
{
  ostringstream oss;
  oss << "{ rect: " << rect.toString() << ", ";
  oss << "state: " << Model::state2str(state) << " }";
  return oss.str();
}
void
DPDeckPSCFG::TBTree::Node::
print() const
{
  cout << toString() << flush;
}
bool 
DPDeckPSCFG::TBTree::Node::
hasParent() const
{
  return (parent != NULL);
}
bool
DPDeckPSCFG::TBTree::Node::
hasChild(int which) const
{
  switch (which) {
  case 0: return (leftChild != NULL);
  case 1: return (rightChild != NULL);
  default: Die("bad child num=%d", which); return false;
  }
}
DPDeckPSCFG::TBTree::Link
DPDeckPSCFG::TBTree::Node::
getParent() const
{
  return parent;
}
DPDeckPSCFG::TBTree::Link
DPDeckPSCFG::TBTree::Node::
getChild(int which) const
{
  switch (which) {
  case 0: return leftChild;
  case 1: return rightChild;
  default: Die("bad child num=%d", which); return NULL;
  }
}
int
DPDeckPSCFG::TBTree::Node::
getNChild() const
{
  int n = 0;
  if (leftChild) {
    n++;
  }
  if (rightChild) {
    n++;
  }
  return n;
}
void
DPDeckPSCFG::TBTree::Node::
replaceChild(const Link& curr, const Link& next)
{
  if (curr == leftChild) {
    leftChild = next;
  } else if (curr == rightChild) {
    rightChild = next;
  } else {
    Die("curr is not a child");
  }
}
bool
DPDeckPSCFG::TBTree::Node::
isLeftChild(const Link& link) const
{
  return (leftChild && link == leftChild);
}
bool
DPDeckPSCFG::TBTree::Node::
isRightChild(const Link& link) const
{
  return (rightChild && link == rightChild);
}
bool
DPDeckPSCFG::TBTree::Node::
isChild(const Link& link) const
{
  return (isLeftChild(link) || isRightChild(link));
}
void
DPDeckPSCFG::TBTree::Node::
deleteR()
{
  if (leftChild) {
    leftChild->deleteR();
  }
  if (rightChild) {
    rightChild->deleteR();
  }
  delete this;
}
DPDeckPSCFG::TBTree::AlnFlagments
DPDeckPSCFG::TBTree::Node::
getAlnFlagmentsR() const
{
  Point pt;
  AlnFlagments v;

  if (leftChild) {
    const AlnFlagments& afl = leftChild->getAlnFlagmentsR();
    v.insert(v.begin(), afl.begin(), afl.end());
  }
  if (rightChild) {
    const AlnFlagments& afr = rightChild->getAlnFlagmentsR();
    v.insert(v.end(), afr.begin(), afr.end());
  }
  switch (Model::getModel().toPSCFGBaseState(state)) {
  case EmitModel::MP_st:
    pt = rect.b;
    v.push_front(new AlnElem(pt, M_st));
    pt.setVal(rect.ex - 1, rect.ey - 1);
    v.push_back(new AlnElem(pt, M_st));
    v.front()->pair = v.back();
    //v.back()->pair = v.front();
    break;
  case EmitModel::MR_st:
    pt.setVal(rect.ex - 1, rect.ey - 1);
    v.push_back(new AlnElem(pt, M_st));
    break;
  case EmitModel::IR_st:
    pt.setVal(rect.ex, rect.ey - 1);
    v.push_back(new AlnElem(pt, I_st));
    break;
  case EmitModel::DR_st:
    pt.setVal(rect.ex - 1, rect.ey);
    v.push_back(new AlnElem(pt, D_st));
    break;
  case EmitModel::ML_st:
    pt = rect.b;
    v.push_front(new AlnElem(pt, M_st));
    break;
  case EmitModel::IL_st:
    pt = rect.b;
    v.push_front(new AlnElem(pt, I_st));
    break;
  case EmitModel::DL_st:
    pt = rect.b;
    v.push_front(new AlnElem(pt, D_st));
    break;
  default:
    break;
  }
  return v;
}

void
DPDeckPSCFG::TBTree::Node::
testConsistencyR()
{
#ifdef DEBUG
  if (parent) {
    Assert(parent->isChild(this));
  }
  if (leftChild) {
    Assert(leftChild->parent == this);
    leftChild->testConsistencyR();
  }
  if (rightChild) {
    Assert(rightChild->parent == this);
    rightChild->testConsistencyR();
  }

  Assert(Model::isValidState(state));
  if (state == Model::BF_st) {
    Assert(leftChild && rightChild);
    Assert(rect.b == leftChild->rect.b);
    Assert(rect.e == rightChild->rect.e);
    Assert(leftChild->rect.e == rightChild->rect.b);
#ifdef SLR_STATE
    Assert(leftChild->state == Model::SL_st 
	   || leftChild->state == Model::X_st);
    Assert(rightChild->state == Model::SR_st
	   || rightChild->state == Model::X_st);
#else
   Assert(leftChild->state == Model::ST_st 
	   || leftChild->state == Model::X_st);
    Assert(rightChild->state == Model::ST_st
	   || rightChild->state == Model::X_st);
#endif
  } else 
#ifdef SLR_STATE
    if (state == Model::SL_st || state == Model::SR_st)
#else
    if (state == Model::ST_st) 
#endif
      {
    Assert(!rightChild);
    if (leftChild) {
      Assert(rect == leftChild->rect);
      Assert(Model::getModel().isChild(state, leftChild->state)
	     || leftChild->state == Model::X_st);
    }
  } else if (state == Model::X_st) {
    if (leftChild) {
      Assert(rect.includes(leftChild->rect));
#ifdef SLR_STATE
      Assert(leftChild->state != Model::SL_st && leftChild->state != Model::SR_st);
#else
      Assert(leftChild->state != Model::ST_st);
#endif
    }
    if (rightChild) {
      Assert(rect.includes(rightChild->rect));
#ifdef SLR_STATE
      Assert(rightChild->state != Model::SL_st && rightChild->state != Model::SR_st);
#else
      Assert(rightChild->state != Model::ST_st);
#endif
    }
    // this seems wrong when nose shape
    //if (leftChild && rightChild) {
    //Assert(leftChild->rect.e == rightChild->rect.b);
    //}
  } else {
    const Rect& rect1 = Model::getModel().addNEmit(rect, state);
    Assert(!rightChild);
    if (leftChild) {
      Assert(rect1 == leftChild->rect);
    }
  }
#endif
}

