/*
 * StripRegionSkip.hpp
 *
 */
#include "StripRegionSkip.hpp"
#include "SkipSet.hpp"

StripRegionSkip::
StripRegionSkip(const StripRegion& sr, const SkipSet& skipSet)
  : sr(sr), skipSet(skipSet)
{
  Assert(skipSet.getSkipSize() > 0);
}
StripRegionSkip::
StripRegionSkip(const StripRegionSkip& srk)
  : sr(srk.sr), skipSet(srk.skipSet)
{
  Assert(skipSet.getSkipSize() > 0);
}
StripRegionSkip::
~StripRegionSkip()
{
}
StripRegionSkip&
StripRegionSkip::
operator=(const StripRegionSkip& srk)
{
  Assert(&skipSet == &srk.skipSet);
  sr = srk.sr;
  Assert(skipSet.getSkipSize() > 0);
  return *this;
}
string
StripRegionSkip::
toString() const
{
  ostringstream oss;
  oss << "--- !strip_region_skip\n";
  oss << "sr: " << sr.toString();
  oss << "skip_set: " << skipSet.toString();
  return oss.str();
}
void
StripRegionSkip::
print() const
{
  cout << toString() << flush;
}
bool
StripRegionSkip::
empty() const
{
  return (getArea() == 0);
}
int
StripRegionSkip::
getNSkipRow() const
{
  if (sr.empty()) return 0;

  return skipSet.getNSkipRow(sr.bx0, sr.ex0);
}
int
StripRegionSkip::
getSkipSize() const
{
  return skipSet.getSkipSize();
}
int
StripRegionSkip::
getArea() const
{
  if (sr.empty()) return 0;

  RowIterator<true> it(*this);
  int n = 0;
  while (it.getNext()) {
    n += it.ly;
  }

#ifdef DEBUG
  Iterator0<true> it0(*this);
  int n0 = 0;
  while (it0.getNext()) {
    n0++;
  }
  Assert(n0 == n);
#endif

  return n;
}
int
StripRegionSkip::
getAreaRight(const int& y) const
{// includes y
  if (sr.ey0 < y) return 0;

  RowIterator<true> it(*this);
  Rect rect = {{{sr.bx0, max(sr.by0, y)}}, {{sr.ex0, sr.ey0}}};
  it.reset(rect);
  int n = 0;
  while (it.getNext()) {
    n += it.ly;
  }
  return n;
}
int
StripRegionSkip::
getAreaBefore(const Point& pt) const
{// excludes pt
  Assert(skipSet.isSkipPoint(pt));
  Assert(sr.rect0.includes(pt));
  RowIterator<true> it(*this);
  int n = 0;
  while (it.getNext()) {
    if (pt.x == it.x) {
      n += (skipSet.getNSkipPointY(pt.x, it.y_beg, pt.y) - 1);
      return n;
    } else {
      n += it.ly;
    }
  }
  pt.print();
  print();
  Die("never reach here");
  return (-1);
}
int
StripRegionSkip::
getAreaTopLeft(const Point& pt) const
{// includes pt
  if (pt.x < sr.bx0 || pt.y < sr.by0) return 0;

  RowIterator<true> it(*this);
  Rect bbox = {{sr.b0}, {pt}};
  it.reset(bbox);
  int n = 0;
  while (it.getNext()) {
    n += it.ly;
  }
  return n;
}
int
StripRegionSkip::
getAreaBottomRight(const Point& pt) const
{// includes pt
  if (sr.ex0 < pt.x || sr.ey0 < pt.y) return 0;

  RowIterator<true> it(*this);
  Rect bbox = {{pt}, {sr.e0}};
  it.reset(bbox);
  int n = 0;
  while (it.getNext()) {
    n += it.ly;
  }
  return n;
}
Point
StripRegionSkip::
findFirstSkipPoint(const Point& pt) const
{
  Assert(sr.includes(pt));
  const int& y = skipSet.findFirstSkipYRange(pt.x, pt.y, sr.getEY(pt.x));
  if (y >= 0) {
    Point pt1 = {pt.x, y};
    return pt1;
  }

  RowIterator<true> it(*this);
  Rect bbox = sr.rect0;
  bbox.bx = (pt.x + 1);
  it.reset(bbox);
  while (it.getNext()) {
    if (it.ly > 0) {
      Point pt1 = {it.x, it.y_beg};
      return pt1;
    }
  }
  return Point::null();
}
Point
StripRegionSkip::
findLastSkipPoint(const Point& pt) const
{
  Assert(sr.includes(pt));

  const int& y = skipSet.findLastSkipYRange(pt.x, sr.getBY(pt.x), pt.y);
  if (y >= 0) {
    Point pt1 = {pt.x, y};
    return pt1;
  }

  RowIterator<false> it(*this);
  Rect bbox = sr.rect0;
  bbox.ex = (pt.x - 1);
  it.reset(bbox);
  while (it.getNext()) {
    if (it.ly > 0) {
      Point pt1 = {it.x, it.y_last};
      return pt1;
    }
  }
  return Point::null();
}
Rect
StripRegionSkip::
getRectK() const
{
  if (empty()) return Rect::null();

  Rect rectk = {
    {findFirstSkipPoint(sr.b0)},
    {findLastSkipPoint(sr.e0)}
  };
  return rectk;
}
Rect
StripRegionSkip::
getRectK(const Rect& rect) const
{
  Assert(sr.includes(rect.b) && sr.includes(rect.e));
  Rect rectk = {
    {findFirstSkipPoint(rect.b)},
    {findLastSkipPoint(rect.e)}
  };
  return rectk;
}
Point
StripRegionSkip::
advance(const Point& pt, const int& i) const
{
  if (empty()) return Point::null();

  const bool& fwd = (i >= 0);
  const int& period = skipSet.getHorizPeriod();
  const int& dy = (fwd ? period : (-period));
  Point pt1 = 
    (fwd ? findFirstSkipPoint(pt) : findLastSkipPoint(pt));
  if (pt1.isNull()) {
    pt.print();
    print();
    Die("no skipPoint i = %d", i);
    return Point::null();
  }
  const int& y_beg0 = (fwd ? pt1.y : sr.getBY(pt1.x));
  const int& y_last0 = (fwd ? sr.getEY(pt1.x) : pt1.y);
  const int& ly = skipSet.getNSkipPointY(pt1.x, y_beg0, y_last0);

  int r = (fwd ? i : (-i));
  if (r < ly) {
    pt1.setVal(pt1.x, pt1.y + r * dy);
    return pt1;
  }
  r -= ly;

  Rect bbox = sr.rect0;
  (fwd ? (bbox.bx = (pt1.x + 1)) : (bbox.ex = (pt1.x - 1)));
  if (fwd) {
    RowIterator<true> it(*this);
    it.reset(bbox);
    while (it.getNext()) {
      if (r < it.ly) {
	const int& y0 = (fwd ? it.y_beg : it.y_last);
	pt1.setVal(it.x, y0 + (r * dy));
	return pt1;
      }
      r -= it.ly;
    }
  } else {// the same code other than true <-> false
    RowIterator<false> it(*this);
    it.reset(bbox);
    while (it.getNext()) {
      if (r < it.ly) {
	const int& y0 = (fwd ? it.y_beg : it.y_last);
	pt1.setVal(it.x, y0 + (r * dy));
	return pt1;
      }
      r -= it.ly;
    }
  }
  pt.print();
  print();
  sr.draw();
  Die("i=%d out of Range", i);
  return Point::null();
}
StripRegionSkip
StripRegionSkip::
getBoundedRegion(const Rect& rect) const
{
  return StripRegionSkip(sr.getBoundedRegion(rect), skipSet);
}

int
StripRegionSkip::
getFirstSkipRow() const
{
  if (sr.empty()) return (-1);
  
  return skipSet.findFirstSkipRowXRange(sr.bx0, sr.ex0);
}
int
StripRegionSkip::
getLastSkipRow() const
{
  if (sr.empty()) return (-1);

  return skipSet.findLastSkipRowXRange(sr.bx0, sr.ex0);
}
int
StripRegionSkip::
getFirstSkipRow(const int& y) const
{
  if (sr.empty() || y < sr.by0 || sr.ey0 < y) return (-1);

  const int& x_beg = sr.getBX(y);
  const int& x_last = sr.getEX(y);
  return skipSet.findFirstSkipRowXRange(x_beg, x_last);
}
int
StripRegionSkip::
getLastSkipRow(const int& y) const
{
  if (sr.empty() || y < sr.by0 || sr.ey0 < y) return (-1);

  const int& x_beg = sr.getBX(y);
  const int& x_last = sr.getEX(y);
  return skipSet.findLastSkipRowXRange(x_beg, x_last);
}
int
StripRegionSkip::
getNSkipRow(const int& y) const
{
  if (sr.empty() || y < sr.by0 || sr.ey0 < y) return 0;

  const int& x_beg = sr.getBX(y);
  const int& x_last = sr.getEX(y);
  return skipSet.getNSkipRow(x_beg, x_last);
}


